\ Curves.f  2022 Jul 31
\ select/get a Curve or CurveSet file and send it to or receive from the unit

decimal

26 constant #CURVES

VARIABLE CurvePad[] 128 allot \ pad area to arrange the curve and its name
\ 32 bytes of curve , 16 bytes of curve name

VARIABLE OldBuffer 4096 allot \ old format curve set area - the curves come first, then the names
\ 26 x 32 bytes of curves ( = 832 bytes ) , 26 x 16 bytes of curve names
\ note that Curve 0 curve name is used as the unit's ID name at power up

variable stopping   0 stopping !      \ CLR has already been pressed
variable CurveDone  0 CurveDone !     \ Curve transfer has been completed

defer ShowCurves

0 value FilePtr \ address of the memory for file
6656 constant FileLen \ length of the whole file
VARIABLE file-buffer FileLen 100 +  allot
file-buffer to FilePtr

0 value fid \ the currently open file's ID

((
\ WinDC MyDC

: InitDC
\   get-dc
   0 call CreateCompatibleDC       PutHandle: fill-dc
\   150 screen-mheight CreateCompatibleBitmap: dc to fill-bitmap
\   fill-bitmap                  SelectObject: fill-dc drop
\                0 0 150 screen-mheight    DKGRAY FillArea: fill-dc
\   release-dc
;
))

: XAllocateFileBuffer  \ allocate the file buffer
   FilePtr ?dup if free drop then
   FileLen malloc to FilePtr
;

: XDeAllocateFileBuffer  \ release the file buffer
   FilePtr ?dup if free drop then
;


CREATE 0Curve  HEX
    0 C, A0 C, 01 C, A0 C,   A0 C, A0 C, A0 C, A0 C, 
   A0 C, A0 C, A0 C, A0 C,   A0 C, A0 C, A0 C, A0 C, 
   A0 C, A0 C, A0 C, A0 C,   A0 C, A0 C, A0 C, A0 C, 
   A0 C, A0 C, A0 C, A0 C,    0 C, A0 C, A0 C, 14 C, 
   41 C, 42 C, 43 C, 44 C,   31 C, 32 C, 33 C, 34 C, 
   20 C, 20 C, 20 C, 20 C,   20 C, 20 C, 20 C, 20 C, 
DECIMAL

((
: >Curve ( n - a )   dup [curve] !
   ?dup if
      256 * FilePtr +  >r
      ( name ) r@  CurvePad[] 32 + 16 move
      ( data ) r> 16 + CurvePad[] 32 move
   else
      FilePtr
      ( name )  CurvePad[] 32 + 16 move
      ( data )  0Curve CurvePad[] 32 move
   then
  CurvePad[]
;
))

: >Curve ( n -- a )   
   dup [curve] !
   ?dup if
      256 * FilePtr + 16 + 
   else
      0Curve
   then
;

: >CurveName ( n - a )
   dup [curve] !
   256 * FilePtr + 
;

HEX
\ swap the 2 bytes in the 16 bit word w
: >< ( w - w')   dup  100 /  0ff and swap  100 * 0ff00 and  or ;
DECIMAL

\ ( Checksum calculations)
\ variable 'C0HIGH
\ : SAVE-C0HIGH   PAD 27 + >R  I C@  'C0HIGH !  160 R> C! ;
\ : RESTORE-C0HIGH   'C0HIGH @ PAD 27 + C! ;

: >OCS ( a n - n)   0 -rot  OVER + SWAP DO  I W@ ><  XOR  2 +LOOP ;

HEX : @OCS ( c -- n )   >Curve  20 >OCS  5FEC +  0ffff and ;  DECIMAL

25 constant #SPECIES            \ the number of species calibration curves
#SPECIES 1+ constant #SPECIES+1  \ plus Code0

: @FULLCSyyyy ( -- n )
   $0000
   #SPECIES+1 0 DO
      i @OCS xor  
   LOOP   
   dup CHECKSUM !
;

decimal

: CalculateLocalChecksum ( -- )   0 CHECKSUM !
   39 0 DO
      16 0 DO
         J 32 * I 2* + OldBuffer + W@ ><
         I 15 = IF  255 AND  THEN
         CHECKSUM +!
      LOOP
   LOOP
   CHECKSUM @  0xFFFF and  CHECKSUM !  
;

: (.HEX) ( n - a n)   BASE @ >R  hex  0 <#  32 HOLD # # # # #>  R> BASE ! ;

: (1.R) ( n - a n)   BASE @ >R  DECIMAL  0 <#  32 HOLD # #>  R> BASE ! ;


\ read the current curve into the file buffer
: @Curve ( n)   >r
   CurrentFilename count r/o open-file abort" Could not open file!" to fid
   r@ 256 * fid reposition-file abort" Could not reposition the file!"
   r@ >CurveName 256 fid read-file  abort" Could not read from the file!"
   fid close-file
   r> drop
;
100 Buffer: MyPad3

: GetFile ( $ -- ) 	\ get the file into the file buffer
    r/o open-file 
    if 
         s" Information" s" Can't find file " MyPad3 place
         CurrentFilename count MyPad3 +place
         s"  - please restart SinarF4..." MyPad3 +place

         0 CurrentFilename c!    \ no file selected
         CurrentFilename count s" Filename" SetSetting \ update the registry

         MyPad3 count Monologue 

         cr ." Could not open the file!"  abort
    then
    to fid
\   0 fid reposition-file abort" Could not reposition the file!"
   0 >CurveName FileLen fid read-file  abort" Could not read the file!"
   2drop
   fid close-file
;

: File>OldBuffer
   \ fill the old format buffer with the file data
   #CURVES 0 do
      ( data ) i >Curve     OldBuffer i 32 * + 32 move
      ( name ) i >CurveName OldBuffer i 16 * + 832 + 16 move
   loop
   0 OldBuffer c! \ THE FIRST TWO BYTES MUST BE THESE!
   160 OldBuffer 1+ c!
   CalculateLocalChecksum
;

:  PutFile
   \ clear the fileset name 
   FilePtr 66 + 62 blank
   s" NewUpload.cvs" r/w OPEN-FILE if  
\      s" Information" s" Could not find file NewUpload.cvs!" Monologue
\      cr ." Could not find file NewUpload.cvs!"  abort  
      s" NewUpload.cvs" r/w CREATE-FILE if 
         s" Information" s" File NewUpload.cvs is being used by another program!" Monologue  
         cr ." Could not find file NewUpload.cvs!"  abort
      then
   then  
   to fid
\   s" NewUpload.cvs" r/w open-file abort" Could not open NewUpload.cvs!" to fid
\   0 fid reposition-file abort" Could not reposition the file!"
   FilePtr FileLen fid write-file  abort" Could not write to the file!"
   fid close-file
;

: OldBuffer>File
   \ fill the file buffer from the old format buffer, and write it to the file
   #CURVES 0 do
      ( data )  OldBuffer i 32 * +  i >Curve  32 move
      ( name )  OldBuffer i 16 * + 832 +  i >CurveName 16 move
   loop 	
   PutFile
;

FileOpenDialog FileSelectDownloadWindow "Download File" "MNET Files (*.CVS)|*.CVS|All Files (*.*)|*.*|"
FileOpenDialog FileSelectUploadWindow "Upload File" "MNET Files (*.CVS)|*.CVS|All Files (*.*)|*.*|"
       
: SelectFile     ( -- )
   FilePtr if  
      FilePtr FileLen erase	\ clear the file buffer
   then

   UploadOrDownload @ if	\ Upload
       c" NewUpload.cvs"  -1 
\      0 Start: FileSelectUploadWindow dup c@  
   else
      0 Start: FileSelectDownloadWindow dup c@
   then
   if
      count CurrentFilename place
      UploadOrDownload @ 0= if 	\ Download
         CurrentFilename count GetFile \ from file to FilePtr RAM
         File>OldBuffer  \ in case we do the old format download	
      then
   else    
      drop 
      0 CurrentFilename ! \ mark as null string
      \ abort" No file selected"
   then    
;

: OpenCurrentFile ( -- )
   CurrentFilename c@ 0= if  exit  then
      UploadOrDownload @ 0= if 	\ Download
         CurrentFilename count GetFile \ from file to FilePtr RAM
         File>OldBuffer  \ in case we do the old format download	
      then
;

: GetCurrentFile
   CurrentFilename c@ 0= if	\ not selected yet
     SelectFile			\ open file select window
   else
     OpenCurrentFile
   then
;


VARIABLE Displayed #CURVES allot \ curves to display ( when they are uploaded )
VARIABLE Highlighted #CURVES allot \ curves to highlight

: DisplayAll   Displayed #CURVES 255 fill  Highlighted #CURVES erase ;
: DisplayNone  Displayed #CURVES   0 fill  Highlighted #CURVES erase ;

: .curve ( n -- )   >r

  r@ if  0  r@ 3 + c>p  r@ 2P.R  Htype  then

  Displayed r@ + c@ if
     5  r@ 3 + c>p r@ >CurveName 16 Htype
     25 r@ 3 + c>p r@ @OCS  (.HEX) Htype
  then
  r> drop
;

158 constant MAX_LINE
VARIABLE SpacePad MAX_LINE allot  spacepad MAX_LINE 32 fill \ a pad of spaces

\ gives the position for the message line, clear the line
: MSG< ( x y -- c l )   0 0 c>p SpacePad MAX_LINE Htype  0 0 c>p ;

: \curve ( n -- )   
   0  swap 3 + c>p  SpacePad MAX_LINE Htype
;

: \page
   29 0 do  i 2 - \curve  loop
   MSG< 2drop
;

: /msg   -3  \curve ;  \ clear the top line


: "tick"  ( 117 pad1 c! pad1 1 )  s" <<< ok       " ;

: .Curve+ ( n -- )   
   >r 32 r> 3 + c>p "tick" Htype
;

: .Curve- ( n -- )   
   >r 42 r> 3 + c>p s" K" Htype
;

: .Curves
   #CURVES 0 do
      i .Curve
      Highlighted i + c@ if  i .Curve+  ( else  i .Curve- )  then
   loop
;

defer .Timeout
variable Timeout#  0 Timeout# ! 

: .Timeout() ( -- )
   Timeout# @ 1+ Timeout# !
   Timeout# @ 15 > if  15 Timeout# !  then
   32 lastLine @ 3 + c>p  s"  Timeout " Htype
   40 lastLine @ 3 + c>p  
   Timeout# @ 15 = if  
      s" Failed!" bell 
      -1 CurveFailed !
   else
      Timeout# @ 3 / 1+ (1.R)
   then  Htype
;

' .Timeout() is .Timeout

\ Host/Target comms.

06 CONSTANT ACK   21 CONSTANT NAK

variable 'FAIL   variable KEY-EXIT   variable TME  0 TME !

: serialPurge+   
  PAD 48 ERASE  5 0 DO  I PAD I + !  LOOP  PAD 12 serialType
  serialPurge 
;
   
\ 2variable C#

: TABB   swap  gotoxy ; 

: TALLY ( a)   DUP @ 1+ SWAP ! ;

variable v_Timeout 

: set_Timeout ( n -- )   2 * v_Timeout ! ;
6 set_Timeout

: WAIT-serialKey ( c)   \ Wait for an ACK for v_Timeout ms 
   WhichKey !
   BEGIN  
      KEY? IF  KEY DROP  0 ELSE  -1  THEN  UNTIL 
      0 KEY-EXIT !  COUNTER v_Timeout @ +
   BEGIN  ( PAUSE )  
      DUP COUNTER - 0< 
      KEY? IF  KEY 27 = IF  0 TME !  -1 KEY-EXIT !  2DROP  EXIT  THEN  THEN
      serialKey? IF  serialKey() WhichKey @ = IF  2DROP  EXIT  THEN  THEN
    UNTIL  DROP  'FAIL TALLY  serialPurge+  .Timeout ;

: WAIT-ACK  ACK WAIT-serialKey ;
: WAIT-03   03 WAIT-serialKey ;

\ Comms. Error Checking
HEX 0F1 CONSTANT POLYNOMIAL

: ,TABLE  100 0 DO  I
   8 0 DO  DUP 1 AND NEGATE  POLYNOMIAL AND  SWAP 2/  XOR  LOOP  C,  LOOP ;

CREATE TABCRC  ,TABLE

TABCRC 0FF + C@ CONSTANT #VALID

variable CRC

\ CREATE +CRC   CRC 0 XOR B   0 HI 0 HI XOR B  0 W MOV
\   TABCRC W) 0 MOV  B  CRC STA B   RET
\ CODE >CRC8 ( a n - c)   1 POP   I 2 MOV   I POP   0FF # CRC MOV
\    BEGIN  LODS B   +CRC CALL  LOOP
\    2 I MOV   0 HI 0 HI XOR B   0 PUSH   NEXT

: +CRC ( c)  CRC @ XOR  TABCRC + C@  CRC ! ;
: >CRC8 ( a n - b)   0FF CRC !  OVER + SWAP DO  I C@ +CRC  LOOP  CRC @ ;

: >CS8 ( a n - b)   FF ROT ROT OVER + SWAP DO  I C@ XOR  LOOP ;

DECIMAL

( Protected data transfer)
variable Pad2  128 ALLOT

\  : 3MS   180 0 DO  ( PAUSE )  LOOP ;
: +serialType ( a n -- )   
   >R  Pad2 R@ MOVE  Pad2 R@ >CS8  Pad2 R@ + C!  Pad2 R@ 1+ >CRC8  
   Pad2 R@ 1+ + C!  Pad2 R> 2 + serialType ;

: +straight ( a n -- f )   >R
   R@ 2 + Pad2 + Pad2 DO  serialKey I C!  LOOP
   Pad2 R@ >CS8  Pad2 R@ + C@ = 1+  
   Pad2 R@ 1+ >CRC8  Pad2 R@ 1+ + C@ = 1+
   2* OR
   DUP 0= IF  Pad2  ROT  R@  MOVE  ACK  ELSE  NAK  THEN
   ( serialEmit ) DROP  R> DROP ;

\ Host/Target comms.

variable 'CMND
variable CBUF  4 ALLOT
variable DTA 80 ALLOT
variable 'RETRY

variable v#Retries

5 CONSTANT #RETRIES
2variable S&D
variable CurveAddress
variable CurveNumber

\ : SserialType ( a n -- )  OVER + SWAP DO  10 MS  I C@ serialEmit  LOOP ; \ Slow serialType
: SserialType ( a n -- )   serialType ;

variable Success
variable RetryTime

: RETRY ( xt -- f )  \ true if ok, else timedout
   'RETRY !  
   counter RetryTime !
   0 v#Retries !
   0 Success !
   0  
   #RETRIES 0 DO  
      1 v#Retries +!
      'RETRY @ EXECUTE
      IF  1-  LEAVE  THEN  
   LOOP  
\   v#Retries @ #RETRIES = 0= Success !
   counter RetryTime @ - dup RetryTime !
   9 (  * 56 ms ) > 0= Success !
;

: CMND() ( n -- )   
   [char] X CBUF c!  'CMND @ CBUF 1+ c!  CBUF 2 + !
   CBUF 1+ 3 >CS8 CBUF 4 + c!  
   CBUF 1+ 4 >CRC8  CBUF 5 + c! 
;

: CMND ( n -- )
   serialPurge
   0 'FAIL !  CMND()   CBUF 6 slowSerialType  
   WAIT-ACK  WAIT-ACK  	\ the remote unit returns two of these!
;

: CMND+ ( n -- ) 
   serialPurge  
   0 'FAIL !  
   CMND()  CBUF 6 slowSerialType  
   WAIT-ACK  
   WAIT-ACK  	\ the remote unit returns two of these!
   serialPurge
   04 serialEmit  
   DTA 16 +straight DROP  
   0 'FAIL ! 
;			

: PPP   BEGIN  serialKey? IF  serialKey() 3 U.R   0 ELSE  -1  THEN  UNTIL ;

: (RMSG)   S&D @ CMND  'FAIL @ 0= ;

: RFUNC()   S&D @ CMND+  'FAIL @ 0= ;

: REMOTE-IN()   CurveNumber @ CMND()
   CBUF 6 slowSerialType  WAIT-ACK  03 SerialEmit  CurveAddress @ 48 +straight ;

: REMOTE-IN ( -- f )   0 'FAIL !  3 set_Timeout  REMOTE-IN()   'FAIL @ IF  0 'FAIL !  0  ELSE  -1  THEN ;


( Execute vector at hex 3E )  HEX
: RCHIPFORTH  05 serialEmit  3E serialEmit  0 serialEmit  serialKey DROP ;
DECIMAL

: .RESULT ( f)  24 72 TABB IF ." done." ELSE ." Failed!" THEN ;

: RMSG ( n )   500 set_Timeout  0 'CMND !  S&D !  ['] (RMSG) RETRY  .RESULT ;

: RCTRL ( n )   500 set_Timeout  1 'CMND !  S&D !  ['] (RMSG) RETRY  DROP ;

: RFUNC ( n -- )
   \ cr 10 10 at-xy  ." RFUNC = " dup .  
   100 ms 
   500 set_Timeout  1 'CMND !  S&D !  ['] RFUNC() RETRY  drop ;

: SET+GET-CS ( -- )   DTA 16 ERASE  4 RFUNC  DTA 14 + W@  RemoteChecksum ! ;	

: RemoteCurveIn ( a curve# -- f )   
	0 CurveDone !  0 Timeout# !  
	CurveNumber !  
	CurveAddress !   
    2 'CMND !  
    ['] REMOTE-IN RETRY   
    -1 CurveDone !  
    swap drop  2ms_delay
    
\ *****************************************************************************
;

\ *****************************************************************************

: REMOTE-OUT()   3 'CMND !  CurveNumber @ CMND()
   CBUF 6 slowSerialType  WAIT-ACK
   WAIT-03
   CurveAddress @ 48 +serialType  WAIT-ACK  ;

: REMOTE-OUT ( -- f)   0 'FAIL !  3 set_Timeout  REMOTE-OUT()   'FAIL @ IF  0 'FAIL !  0  ELSE  -1  THEN ;

: RemoteCurveOut ( a curve# -- f )   0 CurveDone !  0 Timeout# !  CurveNumber !  CurveAddress !  \ was RCURVE-OUT
   3 'CMND ! ['] REMOTE-OUT RETRY  -1 CurveDone !  2ms_delay
;

: RDATA-IN ( c a -- f )   S&D 2! 4 'CMND ! ['] REMOTE-IN RETRY  2ms_delay ;  

: RDATA-OUT ( a c -- f )   S&D 2! 5 'CMND ! ['] REMOTE-OUT RETRY  2ms_delay ;

: TELEPHONE   BEGIN  2 RMSG DROP  bell  serialKey? IF  EXIT  THEN
   800 MS  7 RMSG  DROP  bell  800 MS  serialKey? UNTIL ;

: ZZZ  InitCom . ;

: ZZ  0Curve 0 RemoteCurveOut ;

: ?HELLO ( - f)   0 CMND  2 set_Timeout  WAIT-ACK  'FAIL @ 0= ; \ returns true if ack'ed OK

\ only forth also definitions hidden also forth

: hhh ( - f )
   norm
   s" C:\SINARF4\CSC.CVS"  CurrentFilename place
   CurrentFilename count GetFile
   File>OldBuffer
;


\ **************** Curve Set Transfer ***************************

: StopTransfer ( -- )
   \page   bell
\   500 ms

((
   stopping @ 0= if
      -1 stopping !
      20 0 do  100 ms  CurveDone @ if  leave  then  loop
   then
))

\   0 CurrentFilename c!  \ force a new file selection
   -1 [LeaveEarly] !    \ set the abort marker flag
   0 CurveSetDone !  \ not all done yet
   -1 DisplayOnly ! \ display only, no transfer of curves yet
   0 Transferring !
\   DisplayNone \ don't show any curves yet

   0 RemoteChecksum ! \ no longer valid - a new unit may be connected
   pad 48 erase  pad 48 serialType
   ComClose
\   -98 throw
;

: InitTransfer ( -- )
   StopTransfer
\   DisplayAll
\   \page
   0 stopping !      \ allow curve to finish transferring if false
   0 [LeaveEarly] !  \ clear the flag
;


: >CurvePad ( curve)   dup  >Curve  CurvePad[] 32 move  
   >CurveName  CurvePad[] 32 + 16 move ;

: <CurvePad ( curve)   dup  CurvePad[] swap >Curve 32 move  
   CurvePad[] 32 + swap >CurveName  16 move ;

variable #CurveSuccess

: NewDownload \ download the file...

   ShowCurves

   0 [LeaveEarly] !  \ don't leave early

   0 #CurveSuccess !
   -1 CurveFailed !  \ set by failed transfers

   CheckComPort

   #CURVES 0 do  
      i LastLine !
      i >CurvePad \ copy the curve data to the pad area
      CurvePad[] i RemoteCurveOut if  
\         i .curve+ 
         1 #CurveSuccess +!
         -1 Highlighted i + c! \ mark the curve as highlighted
      then
      LeaveEarlyKey? 
   loop

   #CurveSuccess @ #CURVES = 0= CurveFailed !  \ set by failed transfers

   0 LastLine !  \ put timeout message on CurveSet name line

   SET+GET-CS   \ update and fetch the checksum  ToDo: fix and put back

   ComClose

   0 [LeaveEarly] ! \ mark as done

\   bell
;

: OldDownload \ download the file... using the old format

   File>OldBuffer \ load the old buffer from the file

   0 #CurveSuccess !
   -1 CurveFailed !  \ set by failed transfers

   CheckComPort

   ShowCurves

   Begin  serialKey0 1 = until

   20 ms 

   #CURVES 0 do
      i LastLine !
      OldBuffer i 64 * + 64 slowSerialType  
      -1 Highlighted i + c! \ mark the curve as highlighted
      1 #CurveSuccess +!
      LeaveEarlyKey? 
   loop

   #CurveSuccess @ #CURVES = 0= CurveFailed !  \ set by failed transfers
 
   0 serialEmit  

   ComClose

   bell
;

: Download
   OldFormat @ if
      ['] OldDownload Catch
   else
      ['] NewDownload Catch
   then
   ?dup if  StopTransfer  then
\   ?dup if  bell  ." caught!!! "  StopTransfer  then
;

: NewUpload \ Upload the file...

   0 #CurveSuccess !
   -1 CurveFailed !  \ set by failed transfers

    CheckComPort
   #CURVES 0 do
\      bell
      i LastLine !
      CurvePad[] i RemoteCurveIn if
         i <CurvePad \ copy the pad area to the curve
         ( data )  i >Curve  OldBuffer i 32 * +  32 move
         ( name )  i >CurveName  OldBuffer i 16 * + 832 +  16 move
\         i !Curve \ store the curve to the file
          PutFile
         -1 Displayed i + c!   -1 Highlighted i + c!
         1 #CurveSuccess +!
      then
      LeaveEarlyKey?
      ShowCurves
   loop

   #CurveSuccess @ #CURVES = 0=
    CurveFailed !  \ set by failed transfers

   0 LastLine !  \ put timeout message on CurveSet name line
   SET+GET-CS   \ update and fetch the checksum

   ComClose

   0 [LeaveEarly] ! \ mark as done

   bell
;

: OldUpload	\ Upload the file... using the old format

   OldBuffer 1667 erase 	\ clean it all out

   0 #CurveSuccess !
   -1 CurveFailed !  \ set by failed transfers

   CheckComPort

   1 serialEmit  \ start the transfer
   #CURVES 0 do
      i LastLine !
      OldBuffer i 64 * + 64 serialStraight
\      i .curve+
      LeaveEarlyKey?
      -1 Displayed i + c!   -1 Highlighted i + c!
      1 #CurveSuccess +!
      ShowCurves
   loop

   #CurveSuccess @ #CURVES = 0=
    CurveFailed !  \ set by failed transfers

   ComClose

   OldBuffer>File

   bell
;

: Upload ( -- ) 
\   0 CurveFailed ! \ clear the fail flag - set by failed transfers

\   0 [LeaveEarly] !  \ don't leave early

   OldFormat @ if
      ['] OldUpload Catch
   else
      ['] NewUpload Catch
   then

   CalculateLocalChecksum \ update the checksum

  \ ?dup if  bell  ." caught!!! "  StopTransfer  then
;

: StartTransfer ( -- )

   \page   
   
   \ bell

   0 CurveSetDone !  \ not all done yet
   -1 DisplayOnly ! \ display only, no transfer of curves yet
   0 Transferring !

   0 RemoteChecksum ! \ no longer valid - a new unit may be connected
   pad 48 erase  pad 48 serialType

   0 [LeaveEarly] !    \ set the abort marker flag

   GetCurrentFile

   -1 Transferring !
   0 DisplayOnly !		\ transfer of curves now
   0 CurveSetDone ! 		\ not all done yet

   UploadOrDownload @ if   \ upload
      DisplayNone
      ShowCurves
      Upload
   else                    \ download
      DisplayAll
      ShowCurves
      Download
   then

   1 CurveSetDone ! 		\ all done

   UploadOrDownload @ if   \ upload
      DisplayNone
      ShowCurves
   else                    \ download
      DisplayAll
      ShowCurves
   then
\   ShowCurves

   -1 DisplayOnly !		\ display only, no transfer of curves yet

   0 Transferring !
;


: .DownloadStatus ( -- )
   DisplayOnly @ if       	\ wait for a click to start the transfer
      OldFormat @ if
         MSG< 
         S" Click Start button to start Downloading from file to Unit, Old Format..." Htype
      else
         MSG< 
         S" Click Start button to start Downloading from file to Unit..." Htype
      then
   else				\ show status of transfer
      OldFormat @ if
         MSG< 
         S" Press 'In' on unit to download file 'old' format..." Htype
      else
         MSG< 
         S" Downloading from file to Unit..." Htype
      then        
   then		 
;

: .UploadStatus
   DisplayOnly @ if       	\ wait for a Click to start the transfer
      OldFormat @ if
         MSG< 
         S" Click Start button to start Uploading from Unit to File NewUpload.cvs Old Format..." Htype
      else
         MSG< 
         S" Click Start button to start Uploading from Unit to File NewUpload.cvs ..." Htype
      then
   else				\ show status of transfer
      OldFormat @ if
        MSG< 
        S" Uploading from Unit to File NewUpload.cvs 'old' format..." Htype
      else
        MSG< 
        S" Uploading from Unit to File NewUpload.cvs ..." Htype
      then        
   then		 
;
    	       
: .FileTop
    0 1 c>p FilePtr 66 + 62 Htype 	\ the curve set name

    5 2 c>p s" CurveSet checksum = " Htype
   25 2 c>p  CHECKSUM @ (.HEX) Htype 		\ the curve set checksum
;

29 constant RESULT_LINE#

: .UploadResult
   CurveFailed @ if
      0 RESULT_LINE# c>p s"  ***  Not all curves uploaded! ***     " Htype
   else
      .FileTop
      0 RESULT_LINE# c>p s"  ***  Curves Uploaded successfully *** " Htype

   then
;

: xxx.DownloadResult
   CurveFailed @ if
      0 29 c>p s"  ***  Not all curves downloaded! ***     " Htype
   else
      0 29 c>p s"  ***  Curves Downloaded successfully *** " Htype
   then
;

: .DownloadResult
   CHECKSUM @ RemoteChecksum @ = 
   CHECKSUM @ RemoteChecksum @ - 0x3CA3 = or 
   if 
      CurveFailed @ if
         0 RESULT_LINE# c>p s"  ***  Not all curves downloaded! ***     " Htype
      else
   \      32 23 c>p s" KKKKKKKKKK" HType
         0 RESULT_LINE# c>p s"  ***  Curves Downloaded successfully *** " Htype
      then
   else
     0 RESULT_LINE# c>p s"  ***  Curve checksum error - please try again *** " Htype
   then
;



: DisplayCurves
\   /msg

   CurrentFilename c@ 0= if
      MSG< s" No file selected - Click Start button..." Htype  exit
   then

   CurveSetDone @ if
\      MSG< s" Close this window to return to the main menu screen..." Htype
   else
      UploadOrDownload @ if   \ upload
         .UploadStatus
      else
         .DownloadStatus
      then
   then

   UploadOrDownload @ 0= if \ only for download
      .FileTop
   then

   .Curves

   RemoteChecksum @ if
      32 2 c>p s" Remote CS = " Htype 
      44 2 c>p RemoteChecksum @ (.HEX) Htype
   then

\   1 AllowTime +!
\   AllowTime @ 7 and 0= if 

   CurveSetDone @ if
      UploadOrDownload @ if     \ upload
           .UploadResult
      else                      \ download 
           .DownloadResult
\           25 .Curve+ 
      then
   then
\   then

   [LeaveEarly] @ if
      MSG< 
      s" Transfer aborted - click on Start to start a new transfer..." Htype
   then
;



:object TransferWindow <super window
	     
Font fdFont		\ font data structure

int screen-cols
int screen-rows
int SlowCounter         \ increments when the window is updated

int cur-first-line \ current first line position
0 value first-line#    \ first line number
200 value last-line#   \ last line number
last-line# 20 - value last-top-line#
16 value bytes/line
  

: H.N.str  ( n1 n2 -- adr len )    \ display n1 as a hex number of n2 digits
         BASE @ >R HEX >R
         0 <# R> 0 ?DO # LOOP #>
         R> BASE ! ;
		
:m home: ( -- )
        first-line# to cur-first-line paint: self ;m


: SetupFont
\   InitDC
   8  Width: fdFont
   14 Height: fdFont
   s" Courier" SetFaceName: fdFont
   Create: fdFont
   Handle: fdFont SetFont: dc   \ set the font to be used
;

:M On_Init:     ( -- )          \ things to do at the start of window creation
   On_Init: super             \ do anything superclass needs

   SetupFont

\   0 AllowTime !

   GetCurrentFile

   CurrentFilename count dup if  settitle: self  else  2drop  then

   CurrentFilename count GetFile		\ from file to Buffer
   File>OldBuffer	\ in case we do the old format download	 

   -1 DisplayOnly !	\ display only, no transfer of curves yet

   0 to SlowCounter               \ initialize counter to zero
\   0 500 1 hWnd Call SetTimer drop 	\ kick off a 2Hz timer

   0 200 1 hWnd Call SetTimer drop \ init timer to a 200 ms rate

   -1 TransferWindowOpen !
;M

:m on_size: ( w -- )
        width char-width / to screen-cols
        height char-height  / to screen-rows
        last-line# screen-rows - 0max to last-top-line#
        \ set the vertical scroll limits
\        false last-top-line# first-line# SB_VERT
\        GetHandle: self Call SetScrollRange drop  
\   On_Paint: self
;m

:m on_move: ( w -- )
   On_Paint: self
;m

:m startpos: ( StartPos: parent 20 20 v+ ) 20 100 ;m

:m startsize: 64 char-width *  30 char-height * ;m

variable Painting  0 Painting !

:m on_paint: ( -- )
\   SaveDC: dc                   \ save device context
\   .ComData 			\ update the status bar

\   DisplayCurves
\   20 50 s" Windooooooow!!!"  TEXTOUT: dc
\   20 70 SlowCounter (.) textout: dc
 
\   RestoreDC: dc	
    on_paint: super
   1 +to SlowCounter		\ increment the counter
   SlowCounter @ (.)  settitle: self
;m						     

:M WM_TIMER	( h m w l - res )
   1 +to SlowCounter		\ increment the counter
   Paint: self
   0
;m

((
:m WM_PAINT ( -- )
   SaveDC: dc                   \ save device context
   SetupFont
\   .ComData 			\ update the status bar
   DisplayCurves
   RestoreDC: dc	
\   WM_PAINT super
;m
))


:m on_transfer: ( -- )
   bell
((
\   SaveDC: dc                  \ save device context
   SetupFont
\   Handle: fdFont SetFont: dc   \ set the font to be used
   -1 Transferring !
   0 DisplayOnly !		\ transfer of curves now
\   .ComData 			\ update the status bar
   UploadOrDownload @ if
      Upload
   else
      Download
   then
   MSG< s" Close this window to return to the main menu screen..." Htype
   -1 DisplayOnly !		\ display only, no transfer of curves yet
   0 Transferring !
\   RestoreDC: dc	
))
;m


:m on_display: ( -- )
   SaveDC: dc                  \ save device context
   SetupFont
   DisplayCurves
   RestoreDC: dc	
;m

:M WM_LBUTTONDOWN     ( -- res )
   Transferring @ 0= if
      bell
\      on_transfer: self
      StartTransfer 
   else
      Paint: self
   then  0
;M

:M WM_RBUTTONDOWN     ( -- res )
   Transferring @ 0= if
      bell
\     on_transfer: self
      StartTransfer 
   else
      Paint: self
   then  0
;M
   
:M SetClickFunc: ( cfa -- )
  SetClickFunc: self
;M

:M On_Done:    ( h m w l -- res )
   0 Transferring !
   0 TransferWindowOpen !
   1 hWnd Call KillTimer  drop		\ destroy the timer
   -1 [LeaveEarly] ! 			\ abort the download
   ComClose				\ release the serial port
   Delete: fdFont			\ remove the font
   On_Done: Super	\ the class's close function
   DestroyWindow: self
\   bye 			\ quit everything
   0 ;M

:M WindowStyle: ( -- style )            \ return the window style
                WindowStyle: super
\                WS_VSCROLL -1 xor and           \ don't add vertical scroll bar
\                WS_THICKFRAME -1 xor and           \ don't have a border
		WS_BORDER  or
                ;M
		   
((
:M ExWindowStyle: ( -- extended_style )
                ExWindowStyle: super
                WS_EX_TOPMOST or        \ if so, lock on top
;M
))
((
:m vposition: ( n -- ) \ move to position n
        0max last-top-line# min
        to cur-first-line paint: self ;m

:m vscroll: ( n -- ) \ move n lines up or down
        cur-first-line + vposition: self ;m

:m end: ( -- ) \ move to end, in this case it's 100 bytes down to pad
        last-top-line# to cur-first-line paint: self ;m

:m vpage: ( n -- )  \ down or up n pages
        screen-rows 1- * vscroll: self ;m

:M WM_VSCROLL   ( h m w l -- res )
                swap word-split >r
        CASE
                SB_BOTTOM        of          End: self   endof
                SB_TOP           of         Home: self   endof
                SB_LINEDOWN      of    1 VScroll: self   endof
                SB_LINEUP        of   -1 VScroll: self   endof
                SB_PAGEDOWN      of    1   VPage: self   endof
                SB_PAGEUP        of   -1   VPage: self   endof
                SB_THUMBPOSITION of r@ VPosition: self   endof
                SB_THUMBTRACK    of r@ VPosition: self   endof
        ENDCASE r>drop
        \ position the vertical button in the scroll bar
        TRUE cur-first-line SB_VERT
        GetHandle: self Call SetScrollPos drop

                0 ;M

))

;object


: DisplayFile ( -- )
   DisplayCurves
\        start: TransferWindow
\        on_display: TransferWindow
;

: ShowCurves() ( -- )
   CurrentFilename c@ if 
     DisplayCurves
   else
     MSG< s" No file selected - click on Start button..." Htype
   then
\   on_display: TransferWindow
;
      
' ShowCurves() is ShowCurves


: Wtest ( -- )
\       ['] bell  rel>abs SetClickFunc: TransferWindow 
;

\ ***** program end *****
