\ CURVES.F  2022 Jul 31
\ select/get a Curve or CurveSet file and send it to or receive from the unit

decimal

: MakeBeep ( -- )   1320 ( Hz ) 400 ( ms ) Beep drop ;
: MakeClick ( -- )   exit  4444 ( Hz ) 10 ( ms ) Beep drop ;

: ShowMessage ( $ line# attribute -- )
   Attribute
   >r
   Font12x16x1 FontNew   
   0 r> 7 and 2 + 2dup 
   At-XY #80 spaces
   At-XY type
   RESTORE.
   FontRestore
;

3 CONSTANT #RETRIES

26 constant #CURVES 

#128 cell+ Buffer: CurvePad[]  \ pad area to arrange the curve and its name
\ 32 bytes of curve , 16 bytes of curve name

#4096 cell+ Buffer: OldBuffer  \ old format curve set area - the curves come first, then the names
\ 26 x 32 bytes of curves ( = 832 bytes ) , 26 x 16 bytes of curve names
\ note that Curve 0 curve name is used as the unit's ID name at power up

variable stopping   0 stopping !      \ CLR has already been pressed
variable CurveDone  0 CurveDone !     \ Curve transfer has been completed
variable RemoteChecksum       0 RemoteChecksum !
variable RemoteChecksumValid  0 RemoteChecksumValid !
variable LocalChecksumValid  -1 LocalChecksumValid !  \ not valid when starting an upload, because we do not have all the curves...

variable Displayed[] #CURVES allot \ curves to display ( when they are uploaded )
variable HighlightedCurves[] #CURVES allot \ curves to highlight
variable TransferStatus[] #CURVES allot \ curves to highlight

: DisplayAll   Displayed[] #CURVES 255 fill  HighlightedCurves[] #CURVES erase ;
: DisplayNone  Displayed[] #CURVES   0 fill  HighlightedCurves[] #CURVES erase ;
DisplayAll

#6656 constant |old_FileLen| \ length of the whole file

CREATE 0Curve  HEX
   00 C, A0 C, 01 C, A0 C,   A0 C, A0 C, A0 C, A0 C, 
   A0 C, A0 C, A0 C, A0 C,   A0 C, A0 C, A0 C, A0 C, 
   A0 C, A0 C, A0 C, A0 C,   A0 C, A0 C, A0 C, A0 C, 
   A0 C, A0 C, A0 C, A0 C,   00 C, A0 C, A0 C, 14 C, 
   41 C, 42 C, 43 C, 44 C,   31 C, 32 C, 33 C, 34 C, 
   20 C, 20 C, 20 C, 20 C,   20 C, 20 C, 20 C, 20 C, 
DECIMAL

variable vCurveFileHandle

#256 constant |curve|
|curve| constant |OneCurveBuffer|
|OneCurveBuffer| Buffer: OneCurveBuffer[]
|OneCurveBuffer| Buffer: SavedCurveFile[]
|OneCurveBuffer| Buffer: SavedCurveFile2[]

MAX_PATH Buffer: LastCurrentCurveName[]
MAX_PATH Buffer: LastCurrentCurveSetName[]
MAX_PATH Buffer: SavedCurrentCurveSetName[]  \ saved here while Uploading

\ *****************************************************************************
\ *****************************************************************************
variable CurveSetFileHandle
\ variable CurveSetSize

$1A00 constant |CurveSetSize|

|CurveSetSize| Buffer: CurveSet[]
|CurveSetSize| Buffer: SavedCurveSet[]

: SaveCurveSetCopy ( -- )
   CurveSet[] SavedCurveSet[] |CurveSetSize| cmove  \ save the file
;

MAX_PATH Buffer: LocalCurrentCurveSetName

\ read the current curve into the file buffer
: GetCurveSetFile ( $name -- )

   2dup LocalCurrentCurveSetName place

\   CurveSet_ChangeDirectory

   ( $name -- ) r/o open-file abort" GetCurveSetFile could not open file!" CurveSetFileHandle !
   CurveSet[] |CurveSetSize| CurveSetFileHandle @ read-file  if 
      cr ." GetCurveSetFile could not read from the file = "  LocalCurrentCurveSetName count type  space  #9012 throw  
   then
   |CurveSetSize| = not if  RED. ." GetCurveSetFile wrong read length!"  RESTORE.  #9105 throw  then    
   CurveSetFileHandle @ close-file drop

   SaveCurveSetCopy

\   RestoreDirectory
;

{ 
: >Curve ( n - a )   dup [curve] !
   ?dup if
      |curve| * CurveSet[] +  >r
      ( name ) r@  CurvePad[] 32 + 16 move
      ( data ) r> 16 + CurvePad[] 32 move
   else
      CurveSet[]
      ( name )  CurvePad[] 32 + 16 move
      ( data )  0Curve CurvePad[] 32 move
   then
  CurvePad[]
;
} 

: >Curve ( curve -- a )
   0 max 25 min
   dup [curve] !
   ?dup if
      |curve| * CurveSet[] +  #16 + \ skip the #16 byte curve name
   else
      0Curve
   then
;

: >CurveName ( curve -- a )
   dup [curve] !
   |curve| * CurveSet[] +
;

: >CURVE-OFFSET ( - a)   >CURVE $1E + ;                          

\ Used to be : up to 32 pairs of 16 bit values stored in a curve file, in the top 128 of 256 bytes
\ : >SAMPLE-SET ( -- a )   >CURVE #16 - #128 +  + ;                           
\ : >SAMPLE ( n -- a )   2* 2*  >SAMPLE-SET + ;                     

\ ( Checksum calculations)
\ variable 'C0HIGH
\ : SAVE-C0HIGH   PAD 27 + >r  I c@  'C0HIGH !  160 r> c! ;
\ : RESTORE-C0HIGH   'C0HIGH @ PAD 27 + c! ;

: >OCS ( a n - n)   0 -rot  OVER + SWAP DO  I W@ ><  XOR  2 +LOOP ;

: @OCS ( c -- n )   >Curve  $20 >OCS  $5FEC +  $FFFF and ;  

25 constant #SPECIES            \ the number of species calibration curves
#SPECIES 1+ constant #SPECIES+1  \ plus Code0

: @FULLCSyyyy ( -- n )
   $0000
   #SPECIES+1 0 DO
      i @OCS xor  
   LOOP   
   dup CHECKSUM !
;

: CalculateLocalChecksum ( -- )   0 CHECKSUM !
   #39 0 DO
      #16 0 DO
         J #32 * I 2* + OldBuffer + W@ ><
         I #15 = IF  #255 AND  THEN
         CHECKSUM +!
      LOOP
   LOOP
   CHECKSUM @  $FFFF and  CHECKSUM !  
;

: File>OldBuffer ( -- ) 
   \ fill the old format buffer with the file data
   #CURVES 0 do
      ( data ) i >Curve     OldBuffer i #32 * + #32 move
      ( name ) i >CurveName OldBuffer i #16 * + #832 + #16 move
   loop
   0 OldBuffer c! \ THE FIRST TWO BYTES MUST BE THESE!
   #160 OldBuffer 1+ c!
   CalculateLocalChecksum
;

\ ******************************************************************************
\ Curve data access
\ ******************************************************************************

: >CurveValue ( i -- )
   OneCurveBuffer[] #16 + + 
;

: GetCurveValue ( i -- )
   >CurveValue c@
;

: IncCurveValue ( i -- )
   >CurveValue >r  r@ c@ 1+ 255 min r> c! 
;

: DecCurveValue ( i -- )
   >CurveValue >r r@ c@ 1- 0 max r> c! 
;

: ClearCurveInBuffer ( -- )
   #31 0 do  160 i >CurveValue c!  loop
   #20 #31 >CurveValue c!
;

( Clear curve)  
create CODE0Name ( -- a )   $12 C, $1E C,  ," Sinar Technolo"
   CODE0Name 5 + CODE0Name 2+ $0E MOVE

create DefaultCurve ( -- a )   \ the default empty curve 
   16c,  A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 
   16c,  A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 00 00 A0 14 

\   16c, 00 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 
\   16c, A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 A0 00 A0 A0 14  

\   0 C,  A0 C,  01 C, A0 C, A0 C, A0 C, A0 C, A0 C,
\   A0A0 , A0A0 , A0A0 , A0A0 , A0A0 , A0A0 , A0A0 , A0A0 ,
\   A0A0 , A0A0 ,  0 C,  A0 C,  A0 C, 14 C,


\ *****************************************************************************
\ Curve file copying
\ *****************************************************************************

: SaveCurveCopy ( -- )
   OneCurveBuffer[] SavedCurveFile[] |curve| cmove  \ save the file
;

: SaveCurveCopy2 ( -- )
   OneCurveBuffer[] SavedCurveFile2[] |curve| cmove  \ save the file
;

: RestoreCurveCopy2 ( -- )
   SavedCurveFile2[] OneCurveBuffer[] |curve| cmove  \ save the file
;

: CurveHasChanged? ( -- f )
   OneCurveBuffer[] |curve| SavedCurveFile[] |curve| compare 0= not
;

\ *****************************************************************************
\ Save curve file
\ *****************************************************************************

variable ForceSelectCurve
variable ForceSelectCurveSet

variable MyCurveFileHandle

$100 Buffer: MyPad3
MAX_PATH Buffer: MyPadSaveCurve[]

: SaveCurveFile_F2 ( $filename -- )
   
   MyPadSaveCurve[] place

   MyPadSaveCurve[] count R/W CREATE-FILE if  
      cr RED. ." SaveCurveFile_F2 could not open curve file for writing " RESTORE.  
      BLUE.
      MyPadSaveCurve[] count swap drop 0= if
         ." <no file selected>"
      else
         MyPadSaveCurve[] count type  
      then
      RESTORE.  
      #10004 throw  
   then  
   MyCurveFileHandle !
   OneCurveBuffer[] |curve| MyCurveFileHandle @ WRITE-FILE  if  
      MyCurveFileHandle @ CLOSE-FILE drop  
      cr RED. ." SaveCurveFile_F2 could not write file "  RESTORE.  
      #10005 throw  
   then
   MyCurveFileHandle @ CLOSE-FILE drop

\    MSG3< BLUE.  ."     File saved"  RESTORE.  >MSG
  
   SaveCurveCopy  \ update the copy
;

: SaveCurveIfEdited? ( -- )
   CurveHasChanged? 
   if
      z" Press Ok to save the Curve" z" Curve has been edited" Dialogue
      1 = if 
         CurrentCurveName count SaveCurveFile_F2 
         SaveCurveCopy
      then
   then
;

\ *****************************************************************************
\ Curve file read
\ *****************************************************************************

MAX_PATH Buffer: FilenamePad2[]    
MAX_PATH Buffer: JustFilename[]    
MAX_PATH Buffer: JustPath[]    

MAX_PATH Buffer: JustCurveSetFilename[]    
MAX_PATH Buffer: JustCurveSetPath[]    

\ MAX_PATH Buffer: CurrentPointsName
MAX_PATH Buffer: LastCurvePointsName[]

: UpdateJustFilenamesFromCurveName ( -- )
   CurrentCurveName count -Ext 
   2dup -PATH  JustFilename[] place
        -NAME  JustPath[] place
;

: UpdateJustFilenamesFromCurveSetName ( -- )
   CurrentCurveSetName count -Ext 
   2dup -PATH JustCurveSetFilename[] place
        -NAME JustCurveSetPath[] place
;

: CopyCurveToPointsFilename ( -- )
   CurrentCurveName count swap drop 0= if  exit  then

   CurrentCurveName count -Ext 
   2dup -PATH  JustFilename[] place
        -NAME  JustPath[] place

   CurrentCurveName count -Ext  CurrentPointsName place
   s" .SinarPoints" CurrentPointsName append

   SetRegValues
;

: CopyPointsToCurveFilename ( -- )

   CurrentPointsName count swap drop 0= if  exit  then

   CurrentPointsName count -Ext 
   2dup -PATH  JustFilename[] place
        -NAME  JustPath[] place

   CurrentPointsName count -Ext  CurrentCurveName place
   s" .SinarCurve" CurrentCurveName append

   SetRegValues

\   UpdateJustFilenamesFromCurveName
;


{ 
: xxxGetPointsFile ( -- )

   CurrentCurveName count -Ext CurrentPointsName place
   s" .SinarPoints" FilenamePad2[] append
   
   FilenamePad2[] count  r/o open-file abort" GetPointsFile could not open file!" vCurveFileHandle !
   OneCurveBuffer[] |curve| vCurveFileHandle @ read-file  if 
      cr ." GetPointsFile could not read from the file = "  CurrentCurveName count type  space  #9013 throw  
   then
   drop \ |curve| = not abort" GetCurveFile wrong read length!" 
   vCurveFileHandle @ close-file drop

   SaveCurveCopy
;
}


{ 
: XXXGetCurvexxxSetFile ( a n -- ) 	\ get the file into the file buffer
   2dup MyFileNameSaved[] place
   R/O OPEN-FILE 
   if 
      s" Information" s" Can't find file " MyPad3 place
      CurrentCurveName count MyPad3 append
      \         s"  - please restart Mnet8..." MyPad3 append

      0 CurrentCurveName c!    \ no file selected

      MyPad3 count Monologue 

      cr ." Could not open the file!"  #9001 throw
   then
   CurveSetFileHandle !
   \   0 CurveSetFileHandle @ reposition-file abort" Could not reposition the file!"
      CurveSet[] |CurveSet| CurveSetFileHandle @  READ-FILE  if  
      CurveSetFileHandle @ CLOSE-FILE drop  cr RED. ." Could not read file "  RESTORE.  #10002 throw  
   then
   CurveSetSize ! 

   CurveSetFileHandle @ CLOSE-FILE drop

   CurveSet[] SavedCurveSetFile[] |CurveSetSize| 0 max  |MyFileBuffer| min cmove  \ save the file

   File>OldBuffer
;
}

: (SaveUploadedCurveSetFile) ( -- ) 
   \ clear the fileset name 
   CurveSet[] 66 + 62 blank
   s" .\CurveSets\NewUpload.SinarCurveSet" r/w OPEN-FILE if  
\      s" Information" s" Could not find file NewUpload.cvs!" Monologue
\      cr ." Could not find file NewUpload.cvs!"  abort  
      s" .\CurveSets\NewUpload.SinarCurveSet" r/w CREATE-FILE if 
         s" Information" s" File NewUpload.SinarCurveSet is being used by another program!" Monologue  
         cr ." Could not find file NewUpload.SinarCurveSet!"  abort
      then
   then  
   CurveSetFileHandle !
\   s" NewUpload.cvs" r/w open-file abort" Could not open NewUpload.cvs!" CurveSetFileHandle !
\   0 CurveSetFileHandle @ reposition-file abort" Could not reposition the file!"
   CurveSet[] |CurveSetSize| CurveSetFileHandle @ write-file  abort" Could not write to the file NewUpload.SinarCurveSet!"
   CurveSetFileHandle @ close-file
;

variable vFolderNotFound

: ?ShowFolderNotFound ( -- )
   vFolderNotFound @ if
      3 yMSG<  RED. ." Could not save the uploaded data !!! "  RESTORE. >MSG  1000 ms
      3 yMSG<  RED. ." Could not find folder "  RESTORE.
      BLUE. ." .\CurveSets"  RESTORE. RED. ."  Try 'Save As' " RESTORE.  >MSG  exit \ #9010 throw       
   then
   0 vFolderNotFound !
;

: SaveUploadedCurveSetFile ( -- ) 
   ['] (SaveUploadedCurveSetFile) catch if  
      -1 vFolderNotFound !
   then
;

: OldBuffer>File
   \ fill the file buffer from the old format buffer, and write it to the file
   #CURVES 0 do
      ( data )  OldBuffer i 32 * +  i >Curve  32 move
      ( name )  OldBuffer i 16 * + 832 +  i >CurveName 16 move
   loop 	
   SaveUploadedCurveSetFile
;

\ FileOpenDialog FileSelectDownloadWindow "Download File" "MNET Files (*.CVS)|*.CVS|All Files (*.*)|*.*|"
\ FileOpenDialog FileSelectUploadWindow "Upload File" "MNET Files (*.CVS)|*.CVS|All Files (*.*)|*.*|"
CREATE MNET8-Curve
   ,Z" Mnet8 Curve        (*.SinarCurve)"  ,Z" *.SinarCurve"  
\   ,Z" Mnet8 CurveSet ( *.SinarCurveSet)"  ,Z" *.SinarCurveSet"
\   ,Z" Mnet legacy CurveSet    ( *.CVS)"   ,Z" *.CVS"
   ,Z" Text                      (*.txt)"  ,Z" *.txt"
   ,Z" All                         (*.*)"  ,Z" *.*"
   0 ,

CREATE MNET8-Points
   ,Z" Mnet8 Points      (*.SinarPoints)"  ,Z" *.SinarPoints"  
\   ,Z" Mnet8 CurveSet ( *.SinarCurveSet)"  ,Z" *.SinarCurveSet"
\   ,Z" Mnet legacy CurveSet    ( *.CVS)"   ,Z" *.CVS"
   ,Z" Text                      (*.txt)"  ,Z" *.txt"
   ,Z" All                         (*.*)"  ,Z" *.*"
   0 ,

CREATE MNET8-CurveSet
   ,Z" Mnet8 CurveSet ( *.SinarCurveSet)"  ,Z" *.SinarCurveSet"
   ,Z" Mnet legacy CurveSet    ( *.CVS)"   ,Z" *.CVS"
\   ,Z" Mnet8 Curve        (*.SinarCurve)"  ,Z" *.SinarCurve"  
   ,Z" Text                      (*.txt)"  ,Z" *.txt"
   ,Z" All                         (*.*)"  ,Z" *.*"
   0 ,
      
variable UploadOrDownload
variable CurveOrCurveSet
0 UploadOrDownload !

MAX_PATH Buffer: SavedDirectory[]
MAX_PATH Buffer: NewDirectory[]

: SaveCurrentDirectory ( -- )   
   MAX_PATH SavedDirectory[] GetCurrentDirectory  0= if  cr ." GetCurrentDirectory failed!"  #9001 throw  then ;

: RestoreCurrentDirectory ( -- )   
   SavedDirectory[] SetCurrentDirectory 0= if  cr ." SetCurrentDirectory failed!"  #9002 throw  then ;

: ChangeCurrentDirectory ( $ -- ) 
\   cr ." SavedDirectory = " SavedDirectory[] zcount type
   SavedDirectory[] zcount NewDirectory[] zplace 
   ( $ -- ) NewDirectory[] zappend
\   cr ." NewDirectory[] = " NewDirectory[] zcount type
\   NewDirectory[] SetCurrentDirectory 0= if  cr ." SetCurrentDirectory failed!"  #9003 throw  then    
;


\ *****************************************************************************
\ DataPointPair buffer
\ *****************************************************************************

\ *****************************************************************************
\ Layout of data point pairs on page in F1
\ *****************************************************************************
#24 constant NUMBER_OF_LINES_F1
#04 constant NUMBER_OF_COLUMNS_F1
NUMBER_OF_LINES_F1  NUMBER_OF_COLUMNS_F1 * constant NUMBER_OF_DATAPOINTS

\ *****************************************************************************
\ Data points as entered by user ( not in order )
\ *****************************************************************************
2 cells constant |SORT_ENTRY|

NUMBER_OF_DATAPOINTS |SORT_ENTRY| * constant |DataPointPairs|
|DataPointPairs| Buffer: DataPointPairs[]

variable vNumberDataPointPairs

: >DataPointPair ( n -- a )
   0 max  NUMBER_OF_DATAPOINTS 1- min
   cells 2* DataPointPairs[] + 
;

: GetDataPointPair ( n -- Code0 Moisture% )
   >DataPointPair 2@   
;

: PutDataPointPair ( Code0 Moisture% -- )
   vNumberDataPointPairs @ >DataPointPair 2!   
   1 vNumberDataPointPairs +!
;

\ *****************************************************************************
\ Data points in numerical order
\ *****************************************************************************
12 2** constant |DataPointPairsOrdered|   \ this is BIG to cover lots of midpoints
|DataPointPairsOrdered| |SORT_ENTRY| * Buffer: DataPointPairsOrdered[]

variable vNumberDataPointPairsOrdered

: >DataPointPairOrdered ( n -- a )
   0 max  |DataPointPairsOrdered| 1- min
   |SORT_ENTRY| * DataPointPairsOrdered[] + 
;

: GetDataPointPairOrdered ( n -- Code0 Moisture% )
   >DataPointPairOrdered 2@   
;

: PutDataPointPairOrdered ( Code0 Moisture% -- )
   vNumberDataPointPairsOrdered @ >DataPointPairOrdered 2!   
   1 vNumberDataPointPairsOrdered +!
;

\ *****************************************************************************
\ BubbleSort
\ Arrange the entries in alphabetical order
\ *****************************************************************************

: SortEntryExchange ( n1 n2 -- )  \ swap entries n1 and n2 
   2>r
   2r@ drop      >DataPointPairOrdered 2@          
   2r@ swap drop >DataPointPairOrdered 2@  
   2r@ drop      >DataPointPairOrdered 2!          
   2r@ swap drop >DataPointPairOrdered 2!           
   2r> 2drop

   exit

   2dup                          \ a' a a' a --
   |SORT_ENTRY| R-ALLOC dup >r   \ a' a a' a r --  
   |SORT_ENTRY| move             \ a' a a'  --    \ a copied to R-stack buffer
   >r  |SORT_ENTRY| move  r>     \ a'             \ a' copied to a
   r> swap                       \ r a'
   |SORT_ENTRY| move                              \ a copied to a' (from R-stack)
;

: SortEntryCompare ( c c' -- f )  \ compare entries c and c' 
   >DataPointPairOrdered 2@ drop       \ compare Code0 values ( swap drop to compare Moisture% values )      
   swap >DataPointPairOrdered 2@ drop 
   - 0< not
;

\ Sort the entries in numerical order
\ Original C source from : 
\ http://knol.google.com/k/vivek-bhadra/bubble-sort-algorithm-sorting-and/3c84lj4klzp0d/2# 
\ thanks!
: BubbleSort
   1 vNumberDataPointPairsOrdered @ 0 max ?do
\   cr ." Outer i = " i .   i 10 > if  key drop  then 
      i 1 ?do
\         cr ."     Inner i = " i .  i 10 > if  key drop  then 
         i i 1- SortEntryCompare if  i i 1- SortEntryExchange  then
      loop
   -1 +loop ;
;


: CopyDataPointsToOrderedArray ( -- )
   0 vNumberDataPointPairsOrdered !
   DataPointPairsOrdered[] |DataPointPairs| erase
   NUMBER_OF_DATAPOINTS 0 do
      i GetDataPointPair 2dup 0= swap 0= or if 
         2drop    \ ignore point pairs with either value zero
      else
         PutDataPointPairOrdered
      then
   loop
   BubbleSort
;

: ShowDataPointsOrdered ( -- )
   cr ."  #  Code0  Moisture% "
   vNumberDataPointPairsOrdered @ 0 ?do
      cr i 3 u.r  2 spaces  i GetDataPointPairOrdered  swap .P1 2 spaces  .P1 
   loop
;

\ *****************************************************************************

variable vNoDataPointFileLoaded

: DataPointArrayInit ( -- )
   DataPointPairs[] |DataPointPairs| erase   \ clean out earlier values
   0 vNumberDataPointPairs !
   -1 vNoDataPointFileLoaded !
;

: GetCurveFile() ( -- )
   
   UpdateJustFilenamesFromCurveName

   CurrentCurveName count r/o open-file abort" GetCurveFile could not open file!"  vCurveFileHandle !
   OneCurveBuffer[] |curve| vCurveFileHandle @ read-file  if 
      cr ." GetCurveFile could not read from the file = "  CurrentCurveName count type  space  #9011 throw  
   then
   |curve| = not abort" GetCurveFile wrong read length!" 
   vCurveFileHandle @ close-file drop

   SaveCurveCopy

   DataPointArrayInit   \ mark the data point array as empty, so we load a new file later
;

: (GetCurveFile-overlay) ( -- )
   
   UpdateJustFilenamesFromCurveName

   CurrentCurveName count r/o open-file abort" GetCurveFile could not open file!"  vCurveFileHandle !
   OneCurveBuffer[] |curve| vCurveFileHandle @ read-file  if 
      cr ." GetCurveFile could not read from the file = "  CurrentCurveName count type  space  #9011 throw  
   then
   |curve| = not abort" GetCurveFile wrong read length!" 
   vCurveFileHandle @ close-file drop

\   SaveCurveCopy

\   DataPointArrayInit   \ mark the data point array as empty, so we load a new file later
;

: SelectCurveFile ( -- )
   MNET8-Curve DEFAULT-OPEN-FLAGS  s" .\Curves" ChooseFile+ dup if
      CurrentCurveName place
      CopyCurveToPointsFilename
\      GetCurveFile()   \ done later
      UpdateJustFilenamesFromCurveName
      CopyCurveToPointsFilename
      SetRegValues
      0 RemoteChecksumValid ! \ no longer valid - a new unit may be connected
   else
      2drop
   then
;

\ *****************************************************************************
\ Overlay curves
\ *****************************************************************************

2 2** constant cNumberOfOverlays    \ must be a power of 2
variable vOverlayIndex

MAX_PATH cNumberOfOverlays * Buffer: OverlayFilenames[]

: >OverlayFilename ( index -- a )
   cNumberOfOverlays 1- and MAX_PATH * OverlayFilenames[] + 
;

: InitOverlays ( -- )
   0 vOverlayIndex !
   OverlayFilenames[]  MAX_PATH cNumberOfOverlays *  erase
;

\ just get the curve data points - used for Overlays
: SelectCurveFile- ( -- )
   MNET8-Curve DEFAULT-OPEN-FLAGS  s" .\Curves" ChooseFile+ dup if
      vOverlayIndex @ >OverlayFilename place
      1 vOverlayIndex +!
   else
      2drop
   then
;

\ read the current curve into the file buffer
: GetCurveFile ( -- )

   CurrentCurveName count FILE-STATUS swap drop 0= 
   if     \ the file exists

   else  \ the file does not exist
      SelectCurveFile
   then
   GetCurveFile()
;

: NumberOfCurvesHighlighted ( -- n )
   0 
   HighlightedCurves[] #CURVES over + swap do
      i c@ if  1+  then
   loop
;


\ *****************************************************************************
\ 
\ *****************************************************************************

: SelectCurveSetFile ( -- )
   SaveCurrentDirectory
   
   MNET8-CurveSet DEFAULT-OPEN-FLAGS  s" .\CurveSets" ChooseFile+ dup if
      CurrentCurveSetName place
      SetRegValues

      CurveSet[] @ if  
         CurveSet[] |CurveSetSize| erase	\ clear the file buffer
      then

      CurrentCurveSetName count GetCurveSetFile      \ from file to CurveSet[]

      CurrentCurveSetName count LastCurrentCurveSetName[] place

      HighlightedCurves[] #CURVES erase

      File>OldBuffer  \ for the checksum calculation	
      0 RemoteChecksumValid ! \ no longer valid - a new unit may be connected

      CHECKSUM @ SavedCurveSetChecksum !    \ so we can see if we need to save the curveSet

   else
\      0 CurrentCurveSetName c! \ mark as null string
      2drop
   then
 ;


: ttSelectFile ( -- )
   SelectCurveFile     cr  ." CurrentCurveName    = "  CurrentCurveName count type
   SelectCurveSetFile  cr  ." CurrentCurveSetName = "  CurrentCurveSetName count type
;

{  
: OpenCurrentFile
   CurrentCurveName c@ 0= if  exit  then
\   UploadOrDownload @ 0= if 	\ Download
      CurrentCurveName count GetCurveFile \ from file to CurveSet[] RAM
\   then
;

: GetCurrentFile
   CurrentCurveName c@ 0= if	\ not selected yet
     SelectCurve			      \ open file select window
   else
     OpenCurrentFile
   then
   File>OldBuffer  \ for the checksum calculation	
;
}

\ variable Checksum 0 Checksum !

\ What we want to do in the Transfer Window :
variable dummy  1024 allot
\ variable UploadOrDownload  	0 UploadOrDownload  !	\ true for upload of curves
variable OldFormat  0 OldFormat !		\ true for Old Format mode
variable DisplayOnly   -1 DisplayOnly !	\ true for display only, no transfer of curves

\ Abort the transfer if a C , c or Esc key is pressed
: LeaveEarlyKey? ( -- )
   vLeaveEarly @ if
      #10098 throw
   then
   ekey? if  
      ekey >r
      r@ [char] C = 
      r@ [char] c = or
      r> #27 = or
      if
         MakeBeep
         -1 vLeaveEarly !
         #10099 throw
      then
      0  vLeaveEarly !
   then     
; 

\ Example: Simple terminal program


\ : RPURGE   BEGIN  SerialKey? WHILE SerialBufferGet drop  REPEAT ;
: RPURGE.   CR  BEGIN  SerialKey? WHILE SerialBufferGet .  REPEAT ;


: RKEY0       ( -- c1 )       \ must return a key
   begin   SerialKey?  until  \ loop till we get one
   SerialBufferGet ;

: X?ESC   KEY? IF  27 = ABORT" Escape..."  THEN ;

\ returns a 0 if it times out
: RKEY ( -- c )   
   #1000 COUNTER +  
   BEGIN
      dup COUNTER - 0<
      SerialKey? IF  2drop  SerialBufferGet  EXIT  THEN
      LeaveEarlyKey?
   UNTIL  
   drop  
   0 
\   cr ." >>>> RKEY timed out!!!"  Bell
   #8009 throw
;

\ : SerialEmit      { char -- }     \ write a character to the serial port
\                &LOCAL char 1 ComHandle write-file
\                abort" Failed to write serial port" ;

\ Receives n characters from the serial port and puts them straight into address a
: serialStraight ( addr n -- )
   dup 0= if  2drop  exit  then
   OVER + SWAP DO  RKEY0 I c!  LOOP ;

{  variable len-written
: SerialType ( a n )
   0                                 \ lpOverlapped
   len-written rel>abs               \ lpNumberOfBytesWritten
   rot >r
   rot r> swap  	             \ nNumberOfBytesToWrite(slen)
    rel>abs                          \ *lpBuffer(adr)
   fileid                            \ hFile
   call WriteFile
   ;                                 \ ior = 0 = success
}


\ : SerialType ( a n )   over + swap do  i c@ SerialEmit 5 ms  loop ;	

variable rres

variable 1pad	
variable charpad	

;
\   over + swap do  i c@ SerialEmit  loop ;	

4 constant TOP_OF_CURVESET_DISPLAY

variable CurrentlyDisplayedCurve

: GetCurveLine ( -- y )   
   CurrentlyDisplayedCurve @  TOP_OF_CURVESET_DISPLAY 1 + +  
;

variable TransferStatusLine#
variable TransferCurve#

: >TransferStatus ( -- a )
   TransferStatus[] TransferCurve# @ +
;

: ShowTransferStatus ( c -- )
   3 and
   #44 TransferStatusLine# @ At-XY
   case
   0 of  10 spaces  endof
   1 of  GREEN. ."    <<<<   " RESTORE.  endof
   2 of  GREEN. ."    >>>>   " RESTORE.  endof
   3 of  RED.   ." Failed!!! " RESTORE.  endof
   endcase
;

: ShowRemotePage ( -- )
   CYAN.
   RemotePage 3 and
   case
   0 of  ."   "  endof
   1 of  ." B "  endof
   2 of  ." C "  endof
   3 of  ." D "  endof
   endcase
   RESTORE.
;

: .curve ( n -- )   
   dup CurrentlyDisplayedCurve ! 
   >r
   
   0 GetCurveLine At-XY  80 spaces
   r@ if  3 GetCurveLine At-XY r@  ShowRemotePage CYAN. 2 .dml RESTORE. then
   

   #40 GetCurveLine At-XY
   
   Displayed[] r@ + c@ if
      BLUE. #12 GetCurveLine At-XY   r@ >CurveName #16 type RESTORE.
      MAGENTA. #32 GetCurveLine dup TransferStatusLine# !  At-XY   r@ @OCS  4.hex RESTORE.
      HighlightedCurves[] r@ + c@ if          
         #40 GetCurveLine 2dup At-XY  #20 spaces At-XY GREEN. ."  <<<    " RESTORE.        
      then
      TransferCurve# @ r@ = if
         TransferStatus[] r@ + c@ ShowTransferStatus
      else
         0 ShowTransferStatus    \ clear the status area of all the other curves
      then
   else
      HighlightedCurves[] r@ + c@ if          
         #40 GetCurveLine 2dup At-XY  #20 spaces At-XY GREEN. ." <<<    " RESTORE.        
      then
   then
   r> drop
;

158 constant MAX_LINE
variable SpacePad MAX_LINE allot  spacepad MAX_LINE 32 fill \ a pad of spaces

\ gives the position for the message line, clear the line
: xxxMSG< ( x y -- )   0 0 At-XY SpacePad MAX_LINE type  0 0 At-XY ;

: \curve2 ( n -- )   
   TOP_OF_CURVESET_DISPLAY + CLEAN
;

: /msg ( -- )   0 CLEAN ;  \ clear the top line

{ 
: .Curve+ ( n -- )   
   >r #32 r> 3 + At-XY "tick" 
;

: .Curve- ( n -- )   
   >r #42 r> 3 + At-XY s" K" type
;
}

variable Timeout#  0 Timeout# ! 

: .Timeout ( -- ) 
   Timeout# @ 1+ Timeout# !
   Timeout# @ #RETRIES > if  #RETRIES Timeout# !  then
   Get-XY 2>r
   2r@ At-XY  #10 spaces 2r@ At-XY
   RED. ."   Timeout " RESTORE.
   Timeout# @ #RETRIES = if  
      RED. ."  " RESTORE.  MakeBeep 
      -1 CurveFailed !
      0 Timeout# !
   else
      Timeout# @ 1+ 1 u.r
   then 
   2r> At-XY
;

\ Host/Target comms.

\ 06 CONSTANT <ACK>   21 CONSTANT <NAK>

variable 'FAIL   
variable KEY-EXIT   
variable TME  0 TME !

: TABB ( y x -- )   swap  AT-XY ; 

: TALLY ( a -- )   dup @ 1+ swap ! ;

variable vTimeout  5000 vTimeout !

: SetTimeOut ( ms -- )   vTimeout ! ;  \ 55 ms per old 18.2Hz tick

variable ExitReason

: WAIT-RKEY ( c -- )   \ Wait for character c for vTimeout ms 
   WhichKey !
\   SerialPurge 
   0 ExitReason !   
   0 KEY-EXIT !  
   COUNTER vTimeout @ +
   BEGIN  
      2 ms ( PAUSE )  
      dup COUNTER - 0< if  1 ExitReason !  then
      Key? IF  Key $1B = IF  0 TME !  -1 KEY-EXIT !  2 ExitReason !  THEN  THEN
      SerialKey? IF  SerialBufferGet WhichKey @ = IF  -1 ExitReason !  THEN  THEN
   ExitReason @ UNTIL  
   drop
   ExitReason @ -1 = not if
      'FAIL TALLY  
\      PRGE  
\      MakeBeep  
\      cr ." >>> " ExitReason @ .  ."    char = " WhichKey @ emit+
      .Timeout  
   then
;

: WAIT-ACK ( -- ) 
\   cr ." WAIT-ACK" 
   <ACK> WAIT-RKEY 
;
: WAIT-NAK ( -- )  
\   cr ." WAIT-NAK" 
   <NAK> WAIT-RKEY 
;
: WAIT-ETX ( -- ) 
\   cr ." WAIT-ETX" 
   03 WAIT-RKEY 
;

\ Comms. Error Checking
$00F1 constant Polynomial

{ 
: ,TABLE ( -- )
   $100 0 do  I
      8 0 do  
         dup 1 and negate  Polynomial and  swap 2/  xor  
      loop  
      c,  
   loop 
;
create CRC8table[]  ,table
}

create CRC8table[] 
   16c,  00 E5 29 CC 52 B7 7B 9E A4 41 8D 68 F6 13 DF 3A ..).R.{..A.h...:  \   0
   16c,  AB 4E 82 67 F9 1C D0 35 0F EA 26 C3 5D B8 74 91 .N.g...5..&.].t.  \  10
   16c,  B5 50 9C 79 E7 02 CE 2B 11 F4 38 DD 43 A6 6A 8F .P.y...+..8.C.j.  \  20
   16c,  1E FB 37 D2 4C A9 65 80 BA 5F 93 76 E8 0D C1 24 ..7.L.e.._.v...$  \  30
   16c,  89 6C A0 45 DB 3E F2 17 2D C8 04 E1 7F 9A 56 B3 .l.E.>..-.....V.  \  40
   16c,  22 C7 0B EE 70 95 59 BC 86 63 AF 4A D4 31 FD 18 "...p.Y..c.J.1..  \  50
   16c,  3C D9 15 F0 6E 8B 47 A2 98 7D B1 54 CA 2F E3 06 <...n.G..}.T./..  \  60
   16c,  97 72 BE 5B C5 20 EC 09 33 D6 1A FF 61 84 48 AD .r.[. ..3...a.H.  \  70
   16c,  F1 14 D8 3D A3 46 8A 6F 55 B0 7C 99 07 E2 2E CB ...=.F.oU.|.....  \  80
   16c,  5A BF 73 96 08 ED 21 C4 FE 1B D7 32 AC 49 85 60 Z.s...!....2.I.`  \  90
   16c,  44 A1 6D 88 16 F3 3F DA E0 05 C9 2C B2 57 9B 7E D.m...?....,.W.~  \  A0
   16c,  EF 0A C6 23 BD 58 94 71 4B AE 62 87 19 FC 30 D5 ...#.X.qK.b...0.  \  B0
   16c,  78 9D 51 B4 2A CF 03 E6 DC 39 F5 10 8E 6B A7 42 x.Q.*....9...k.B  \  C0
   16c,  D3 36 FA 1F 81 64 A8 4D 77 92 5E BB 25 C0 0C E9 .6...d.Mw.^.%...  \  D0
   16c,  CD 28 E4 01 9F 7A B6 53 69 8C 40 A5 3B DE 12 F7 .(...z.Si.@.;...  \  E0 ()
   16c,  66 83 4F AA 34 D1 1D F8 C2 27 EB 0E 90 75 B9 5C f.O.4....'...u.\  \  F0   

CRC8table[] $FF + c@ constant #valid

variable CRC8

\ CREATE +CRC8   CRC8 0 XOR B   0 HI 0 HI XOR B  0 W MOV
\   CRC8table[] W) 0 MOV  B  CRC8 STA B   RET
\ CODE >CRC8 ( a n - c)   1 POP   I 2 MOV   I POP   0FF # CRC8 MOV
\    BEGIN  LODS B   +CRC8 CALL  LOOP
\    2 I MOV   0 HI 0 HI XOR B   0 PUSH   NEXT

: +CRC8 ( c -- )  CRC8 @ XOR  CRC8table[] + c@  CRC8 ! ;

: >CRC8 ( a n -- b )   $FF CRC8 !  OVER + SWAP DO  I c@ +CRC8  LOOP  CRC8 @ ;

: >CS8 ( a n -- b )   $FF ROT ROT OVER + SWAP DO  I c@ XOR  LOOP ;

DECIMAL

\ : SRTYPE ( a n -- )  OVER + SWAP DO  10 MS  I c@ SerialEmit  LOOP ; \ Slow SerialType
\ : SRTYPE ( a n -- )   SerialType ;

( Protected data transfer)
variable Pad2  128 ALLOT

\  : 3MS   180 0 DO  ( PAUSE )  LOOP ;
: +SerialType ( a n -- )   
   >r  
   ( a -- )          Pad2 r@ MOVE  
   Pad2 r@ >CS8      Pad2 r@ + c!  
   Pad2 r@ 1+ >CRC8  Pad2 r@ 1+ + c!  
   Pad2 r> 2 + serialType ;

: +straight ( a n -- f )   
   >r
   r@ 2 + Pad2 + Pad2 DO  RKEY I c!  LOOP
   Pad2 r@     >CS8  Pad2 r@ + c@ = 1+  
   Pad2 r@ 1+ >CRC8  Pad2 r@ 1+ + c@ = 1+ 2* 
   OR
   dup 0= IF  Pad2  ROT  r@  MOVE  <ACK>  ELSE  <NAK>  THEN
   ( SerialEmit ) drop  r> drop ;

\ *****************************************************************************
\ Host/Target comms.
\ *****************************************************************************

variable 'CMND
variable CBUF  4 ALLOT

#80 Buffer: DTA 

variable 'RETRY

variable v#Retries

2variable S&D
variable CurveAddress
variable v_CurveNumber

variable Success

: (Retry) ( --  )   \ true if ok, else timedout
   #RETRIES 0 do  
      1 v#Retries +!
      'RETRY @execute 
      Success @ if  leave  then 
      10 ms 
   loop  
;

: RETRY ( xt -- f )  \ true if ok, else timedout
   'RETRY ! 
   0 v#Retries !
   0 Success !
   ['] (Retry) catch if  0 Success !  then 
   Success @
;

: CMND() ( n -- )   
   [char] X CBUF c!  'CMND @ CBUF 1+ c!  CBUF 2 + !
   CBUF 1+ 3 >CS8   CBUF 4 + c!  
   CBUF 1+ 4 >CRC8  CBUF 5 + c! 
;

: slowSerialType ( a n )
   dup 0= if  2drop  exit  then	
   over + swap do  2 ms  i c@ SerialEmit  loop 
;	

: CMND ( n -- )
   SerialPurge 
   0 'FAIL !  CMND()  CBUF 6 slowSerialType  
   WAIT-ACK  WAIT-ACK  	\ the remote unit returns two of these!
   'Fail @ 0= Success !
;

: CMND+ ( n -- )
   0 Success !
   DTA #16 erase
\   04 SerialEmit  100 ms  RED. SerialShow  RESTORE.  \ SerialPurge 
   CMND()  CBUF 6 SlowSerialType  
   WAIT-ACK     \ the remote unit ACKs the command
   WAIT-ACK  	\ the remote unit ACKs that it is ready to return the data
   04 SerialEmit  
\   1000 ms 
   DTA #16 +straight drop  
   -1 Success !
;			

\ : PPP   BEGIN  SerialKey? IF  SerialBufferGet 3 U.R   0 ELSE  -1  THEN  UNTIL ;

: (RMSG)   S&D @ CMND ;

: (RFUNC) ( -- )   S&D @ CMND+  ;

: RFUNC ( n -- f )  
   #5000 SetTimeOut  1 'CMND !  S&D !  ['] (RFUNC) RETRY 
;

\ *****************************************************************************

: CurveNumber@ ( n -- )   
   v_CurveNumber @  
   RemotePage 0= not if  
      RemotePage 100 * + 
   then
;

: CurveNumber! ( n -- )   v_CurveNumber ! ;

: REMOTE-IN() ( -- )
   CurveNumber@ (CMND) CBUF 6 slowSerialType  WAIT-ACK  <ETX> SerialEmit  CurveAddress @ #48 +straight drop  ;

: REMOTE-IN ( -- )   
   0 'FAIL !  #500 SetTimeOut  ['] REMOTE-IN() catch 
   if  
      -1 'FAIL !  
   else
      -1 Success !
   then ;

: RemoteCurveIn ( c a -- f)   
   0 CurveDone !  0 Timeout# !  
   CurveAddress !  
   CurveNumber!
   2 'CMND !  
   ['] REMOTE-IN RETRY
  -1 CurveDone !
;

\ *****************************************************************************

: REMOTE-OUT()   3 'CMND !  CurveNumber@ (CMND)
   CBUF 6 slowSerialType  
   WAIT-ACK
   WAIT-ETX
   SerialPurge
   CurveAddress @ #48 +SerialType  
   \ Note : we are expecting an ACK here, but the FPA V7.9 does not seem to give one, 
   \ so we just timeout here without giving an error...
   2 ms 
   SerialKey <ACK> = if  
      -1 Success !
   then
;

: REMOTE-OUT ( -- )   
\   20 ms 
   0 'FAIL !  #2000 SetTimeOut  REMOTE-OUT()  
   50 ms 
;

: RemoteCurveOut ( a c -- f )   0 CurveDone !  0 Timeout# !  CurveNumber!  CurveAddress !
   3 'CMND ! 
   ['] REMOTE-OUT RETRY  
   -1 CurveDone !
;

\ *****************************************************************************

( Execute vector at hex 3E )  
: RCHIPFORTH ( -- )   $05 SerialEmit  $3E SerialEmit  0 SerialEmit  RKEY drop ;

: .RESULT ( f -- )   if  GREEN. ."  done."  else  RED. ."  Failed!"  then  RESTORE. ;

: RMSG ( n -- )   #5000 SetTimeOut  0 'CMND !  S&D !  ['] (RMSG) RETRY  .RESULT ;

: RCTRL ( n -- )   #5000 SetTimeOut  1 'CMND !  S&D !  ['] (RMSG) RETRY  drop ;

: Get_RemoteChecksum ( -- )   
   4 RFUNC if  
      DTA #14 + W@  RemoteChecksum !  -1 RemoteChecksumValid !  
   then 
;	

: RDATA-IN ( c a -- f )   S&D 2! 4 'CMND ! ['] REMOTE-IN RETRY ;  

: RDATA-OUT ( a c -- f )   S&D 2! 5 'CMND ! ['] REMOTE-OUT RETRY ;

: TELEPHONE ( -- )
   BEGIN  
      2 RMSG drop  
      MakeBeep  
      SerialKey? if  exit  then
      #800 ms  7 RMSG  
      drop  MakeBeep  #800 MS  
   SerialKey? UNTIL 
;

: zzz   counter >r  0Curve 0 RemoteCurveOut  cr ." ms = " r> timer ;

: zzzz   counter >r  0 pad  RemoteCurveIn  cr ." ms = " r> timer ;

: ?HELLO ( -- f )   
   0 Success !
   #1500 SetTimeOut  0 'CMND !  0 CMND  Success @ ; \ returns true if ack'ed OK

\ only forth also definitions hidden also forth

\ **************** Curve Set Transfer ***************************

: StopTransfer ( -- )
 \  MakeBeep
\   500 ms

\   0 CurrentCurveName c!  \ force a new file selection
   -1 vLeaveEarly !    \ set the abort marker flag
   0 CurveSetDone !  \ not all done yet
   -1 DisplayOnly ! \ display only, no transfer of curves yet
   0 Transferring !
\   DisplayNone \ don't show any curves yet

   0 RemoteChecksumValid ! \ no longer valid - a new unit may be connected
\   pad 48 erase  pad 48 SerialType
   SerialClose

   0 vLeaveEarly !    \ clear the abort marker flag
;

: FileBufferToCurvePad ( curve -- )   
   dup  >Curve  CurvePad[] #32 move      \ Note : >Curve uses the 0Curve fixed data for reads if n == 0 
   >CurveName  CurvePad[] #32 + #16 move 
;

: CurvePadToFileBuffer ( curve -- )  
   >r
   r@ if 
      CurvePad[] r@ >Curve #32 move            \ write to the File buffer 
   else
      CurvePad[] r@ >CurveName #16 + #32 move  \ even curve 0 
   then
   CurvePad[] #32 + r@ >CurveName #16 move 
   r> drop
;


variable #CurveSuccess

: ttsm ( -- )
   4 0 do  i (.) i i ShowMessage  loop
;

: ShowFilenameCamel- ( $ -- )
   Font24x32x1 FontNew
   2dup 0 max $10 min type
   dup $10 > if
\      Get-XY 1+ swap 2* swap At-XY
      #24 Attribute+  \ cyan on white
      $27 min  \ maximum number of 24x32 characters on one line  
      $10 /string type  \ remove the first 16 characters
      RESTORE.
   else
      2drop
   then
\   20 spaces 
\   cr ." 0123456789012345678901234567890123456789"
   
   FontRestore
;

: ShowFilenameCamel ( $ -- )
   Font8x16x1 FontNew
   0 8 PixelAt-XY  ." Name : " 
   7 8 *  4 PixelAt-XY 
   BLUE.
   FontRestore

   ShowFilenameCamel-
   
   Font24x32x1 FontNew
   20 spaces
   FontRestore
   RESTORE. 
;

: ?SerialOpen ( -- )
\   serialClose
\   GetRegValues
   ['] SerialOpen catch if  
      realSerialClose
      MSG3<  RED. ." No serial port!" RESTORE.  >MSG  
      ['] StopTransfer catch drop  0 vLeaveEarly ! 
      #1000 ms  
      #10101 throw  
   then
;

: ShowCurveSetChecksum ( -- )
   4 yMSG< 
   LocalChecksumValid @ if
      BLUE. ." CurveSet Checksum = " RESTORE.
      MAGENTA. CHECKSUM @  4.hex RESTORE.
   else
      #23 spaces 
   then
   RemoteChecksumValid @ if
      BLUE.  ."    Remote checksum = "  RESTORE.
      MAGENTA. RemoteChecksum @ 4.hex 	RESTORE.	\ the curve set checksum
   else
      #24 spaces 
   then

   LocalChecksumValid @ RemoteChecksumValid @ and if
      RemoteChecksum @ CHECKSUM @ = if
         GREEN. ."  ok "  RESTORE.
      else
         RED.  ."  !!!"  RESTORE. 
      then
   else
      4 spaces
   then

   >MSG
;

: Download ( -- )   \ download the file...

   0 vLeaveEarly !  \ don't leave early

   0 #CurveSuccess !
   -1 CurveFailed !  \ also set by failed transfers

   ?SerialOpen

{ 
   ['] SerialOpen catch if
      0 TransferCurve# !
      3 >TransferStatus c!    \ mark as Failed!!!
      .curve
      1000 ms 
      exit
   then
}

   #CURVES 0 do  
      i LastLine !
      i TransferCurve# !
      1 >TransferStatus c!    \ mark as downloading
      i .curve
      i FileBufferToCurvePad \ copy the curve data to the pad area
      CurvePad[] i RemoteCurveOut if  
\         i .curve+ 
         1 #CurveSuccess +!
         -1 HighlightedCurves[] i + c! \ mark the curve as Highlighted
      then
      0 >TransferStatus c!    \ clear the transfer status area
      i .curve
      LeaveEarlyKey? 
   loop

   #CurveSuccess @ #CURVES = 0= CurveFailed !  \ also set by failed transfers

   0 LastLine !  \ put timeout message on CurveSet name line

   serialOpen 
   Get_RemoteChecksum   \ update and fetch the checksum  ToDo: fix and put back

   ShowCurveSetChecksum

   SerialClose

   0 vLeaveEarly ! \ mark as done

   MakeBeep
;

: OldDownload \ download the file... using the old format

   File>OldBuffer \ load the old buffer from the file

   0 #CurveSuccess !
   -1 CurveFailed !  \ also set by failed transfers

   ?SerialOpen

   Begin  RKEY0 1 = until

   20 ms 

   #CURVES 0 do
      i LastLine !
      OldBuffer i 64 * + 64 slowSerialType  
      -1 HighlightedCurves[] i + c! \ mark the curve as highlighted
      1 #CurveSuccess +!
      LeaveEarlyKey? 
   loop

   #CurveSuccess @ #CURVES = 0= CurveFailed !  \ also set by failed transfers
 
   0 SerialEmit  

   SerialClose

   MakeBeep
;

: Upload ( -- )   \ Upload the file...

   0 vFolderNotFound !
   0 #CurveSuccess !
   -1 CurveFailed !  \ also set by failed transfers

   ?SerialOpen

   DisplayNone    \ do not display the curve names yet, as we don't know what they will be...
   
   #CURVES 0 do
      i LastLine !
      i TransferCurve# !
      2 >TransferStatus c!    \ mark as uploading
      i .curve
      i  CurvePad[]  RemoteCurveIn if
         i CurvePadToFileBuffer \ copy the pad area to the curve
\        ( data )  i >Curve  OldBuffer i #32 * +  #32 move
\         ( name )  i >CurveName  OldBuffer i #16 * + #832 +  #16 move
\         i !Curve \ store the curve to the file
         -1 Displayed[] i + c!   
         1 HighlightedCurves[] i + c!
         1 #CurveSuccess +!
      then
      0 >TransferStatus c!    \ clear the transfer status area
      i .curve
      LeaveEarlyKey?
   loop

   0
   #CURVES 0 do
      Displayed[] i + c@ $FF = not if  -1 or  leave  then   
   loop
   0= if

      CurrentCurveSetName count SavedCurrentCurveSetName[] place

      SaveUploadedCurveSetFile

      File>OldBuffer  \ also updates the checksum	

      #CurveSuccess @ #CURVES = 0= CurveFailed !  \ also set by failed transfers

      0 LastLine !  \ put timeout message on CurveSet name line

      500 ms Get_RemoteChecksum   \ update and fetch the checksum   
      ShowCurveSetChecksum

      SavedCurrentCurveSetName[] count CurrentCurveSetName place

   then

   SerialClose

   0 vLeaveEarly ! \ mark as done

   MakeBeep
;

: OldUpload	\ Upload the file... using the old format

   OldBuffer 1667 erase 	\ clean it all out

   0 #CurveSuccess !
   -1 CurveFailed !  \ also set by failed transfers

   ?SerialOpen

   1 SerialEmit  \ start the transfer
   #CURVES 0 do
      i LastLine !
      OldBuffer i 64 * + 64 serialStraight
\      i .curve+
      LeaveEarlyKey?
      -1 Displayed[] i + c!   -1 HighlightedCurves[] i + c!
      1 #CurveSuccess +!
\      ShowCurves
   loop

   #CurveSuccess @ #CURVES = 0=
    CurveFailed !  \ also set by failed transfers

   SerialClose

   OldBuffer>File

   MakeBeep
;

: .DownloadStatus
exit
   DisplayOnly @ if       	\ wait for a click to start the transfer
      OldFormat @ if
         MSG< 
         ." Click Start button to start Downloading from file to Unit, Old Format..." >MSG
      else
         MSG< 
         ." Click Start button to start Downloading from file to Unit..." >MSG
      then
   else				\ show status of transfer
      OldFormat @ if
         MSG< 
         ." Press 'In' on unit to download file 'old' format..." >MSG
      else
         MSG< 
         ." Downloading from file to Unit..." >MSG
      then        
   then		 
;

: .UploadStatus
exit
   DisplayOnly @ if       	\ wait for a Click to start the transfer
      OldFormat @ if
         MSG< 
         ." Click Start button to start Uploading from Unit to File NewUpload.cvs Old Format..." >MSG
      else
         MSG< 
         ." Click Start button to start Uploading from Unit to File NewUpload.cvs ..." >MSG
      then
   else				\ show status of transfer
      OldFormat @ if
        MSG< 
        ." Uploading from Unit to File NewUpload.cvs 'old' format..." >MSG
      else
        MSG< 
        ." Uploading from Unit to File NewUpload.cvs ..." >MSG
      then        
   then		 
;

: ShowPathString ( $ -- )
   Font8x16x1 FontNew   \ set the default font       
   0 2 At-XY  #80 spaces  0 2 At-XY  ." Path : "  BLUE.  ( $ -- )  80 min  type  RESTORE.
   FontRestore
;


: SetCurveSetFilenames ( $ -- )
   JustCurveSetFilename[] $10 BL fill
   -Ext 
   2dup -PATH JustCurveSetFilename[] place
        -NAME JustCurveSetPath[] place

   JustCurveSetFilename[] count ShowFilenameCamel 
   JustCurveSetPath[] count ShowPathString

\   BLUE. #32 5 At-XY CHECKSUM @ 4.hex 	RESTORE.	\ the curve set checksum
;

\ #31 TOP_OF_CURVESET_DISPLAY + constant RESULT_LINE#

: ChecksumsMatch ( -- f )   \ or are not available
   LocalChecksumValid @ RemoteChecksumValid @ and if
      RemoteChecksum @ CHECKSUM @ = 
   else
      -1
   then
;

: .UploadResult
   ChecksumsMatch 0= if
      s"  ***  Not all curves uploaded! ***     "  1 ( red) 2 ShowMessage
   else
\      CurrentCurveSetName count SetCurveSetFilenames
      s" ***  Curves Uploaded successfully ***" 1 ( green) 7 ShowMessage
      1500 ms  ?ShowFolderNotFound  \ warn if the folder was not found     
   then
;

: .DownloadResult

\   BLUE. #32 5 At-XY CHECKSUM @ 4.hex 	RESTORE.	\ the curve set checksum

   RemoteChecksumValid @ 0= if  exit  then

   CHECKSUM @ RemoteChecksum @ = 
   CHECKSUM @ RemoteChecksum @ - $3CA3 = or 
   if 
      CurveFailed @ if
          s"  ***  Not all curves downloaded! ***     "  1 ( red) 2 ShowMessage
      else
          s" ***  Curves Downloaded successfully ***" 1 ( green) 7 ShowMessage
      then
   else
     s"  ***  Curve checksum error - please try again *** "  1 ( magenta) 5 ShowMessage
   then
;

: DisplayCurves ( -- )

   page 
 
   Font12x16x1 FontNew
   
   UploadOrDownload @ 0= if \ only for download
      CurrentCurveSetName count SetCurveSetFilenames
   else
      s" .\CurveSets\NewUpload.SinarCurveSet" SetCurveSetFilenames      
   then 

   Font12x16x1 FontNew

   #CURVES 0 do
      i .Curve
   loop

   File>OldBuffer

   ShowCurveSetChecksum

   vLeaveEarly @ if
\      4 yMSG<  ." Transfer cancelled." >MSG
      0 vLeaveEarly !
   then
   
   InvalidateAll

   FontRestore
;

: DisplayCurveSetFile ( -- )
   CurrentCurveSetName c@ 0= if
      SelectCurveSetFile
   else
\      0 CurrentCurveSetName c!
   then

 \  CurrentCurveSetName count LastCurrentCurveSetName[] over compare 0= not if \ if different, load the new file
      CurrentCurveSetName count ['] GetCurveSetFile catch if 
         0 CurrentCurveSetName c! 
         2drop exit
      then
      CurrentCurveSetName count LastCurrentCurveSetName[] place
\   then

   CurveSetDone @ if
\      MSG< ." Close this window to return to the main menu screen..." >MSG
   else
      UploadOrDownload @ if   \ upload
         .UploadStatus
      else
         .DownloadStatus
      then
   then

   DisplayCurves
;

\ *****************************************************************************

variable v_NeedToTransferToCurve
variable v_PointsTransferredToCurve

: InitTransfer ( -- )
   GetRegValues
\   StopTransfer
\   DisplayAll
\   page
   0 stopping !      \ allow curve to finish transferring if false
   0 vLeaveEarly !  \ clear the flag
   0 CurrentlyDisplayedCurve ! 
   0 v_NeedToTransferToCurve !
   0 v_PointsTransferredToCurve !
;

: StartTransfer ( -- )

   0 CurveSetDone !  \ not all done yet
   -1 DisplayOnly ! \ display only, no transfer of curves yet
   0 Transferring !

   0 RemoteChecksumValid ! \ no longer valid - a new unit may be connected
   pad 48 erase  pad 48 SerialType

   0 vLeaveEarly !    \ set the abort marker flag

\   GetCurrentFile

   -1 Transferring !
   0 DisplayOnly !		\ transfer of curves now
   0 CurveSetDone ! 		\ not all done yet

   UploadOrDownload @ if   \ upload
      0 LocalChecksumValid !
      DisplayNone
      DisplayCurveSetFile
      Upload
      -1 LocalChecksumValid !
      ShowCurveSetChecksum
      .UploadResult
   else                    \ download
      DisplayAll
      DisplayCurveSetFile
      Download
      .DownloadResult
   then

   1 CurveSetDone ! 		\ all done
;

\ *****************************************************************************
\ Automatic curve calibration)                               
\ *****************************************************************************

$40 constant NUMBER_OF_DATA_POINTS

NUMBER_OF_DATA_POINTS     cells Buffer: SPLURGE-COUNT                                 
NUMBER_OF_DATA_POINTS 2 * cells Buffer: SPLURGE-VALUES                      

NUMBER_OF_DATA_POINTS     cells Buffer: REAL-COUNT                                 
NUMBER_OF_DATA_POINTS 2 * cells Buffer: REAL-VALUES                      

|curve| Buffer: AutoCalibrationCurve[]    \ : >AUTO-CURVE ( - a)   LM @ >curve ;                            
: >AUTO-OFFSET ( -- a)   AutoCalibrationCurve[] $1E + ;                      
                                                                
: InitAutoCalibration ( -- )   

   CopyDataPointsToOrderedArray  \ also sets the number if data point pairs in  vNumberDataPointPairsOrdered

   SPLURGE-COUNT  NUMBER_OF_DATA_POINTS     cells erase   
   SPLURGE-VALUES NUMBER_OF_DATA_POINTS 2 * cells erase   
   REAL-COUNT     NUMBER_OF_DATA_POINTS     cells erase   
   REAL-VALUES    NUMBER_OF_DATA_POINTS 2 * cells erase   
   
   OneCurveBuffer[] #16 + AutoCalibrationCurve[]  $20 move
   AutoCalibrationCurve[] $1D 3 /string $A0 fill  
   $14 AutoCalibrationCurve[] $1F + c!  
   $00 AutoCalibrationCurve[] $1C + w!      
;

: ShowSplurges ( -- )
   cr ." Splurge count : "
   SPLURGE-COUNT  NUMBER_OF_DATA_POINTS     cells 4 / dump   
   cr ." Splurge values : "
   SPLURGE-VALUES NUMBER_OF_DATA_POINTS 2 * cells 4 / dump   
   cr ." Real count : "
   REAL-COUNT     NUMBER_OF_DATA_POINTS     cells 4 / dump   
   cr ." Real values : "
   REAL-VALUES    NUMBER_OF_DATA_POINTS 2 * cells 4 / dump   
   cr ." AutoCalibrationCurve[] : "
   AutoCalibrationCurve[] $20 dump
;

variable CURRENT-SAMPLE   
variable SAMPLE

{ 

: LeaveEarly? ( -- )
   CURRENT-SAMPLE @  SAMPLE ! 
\   .SAMPLES  
   -1 ?LEAVE !       
\   BELL $400 MS 
;                                                

: InitLeaveEarly? ( -- )   0 ?LEAVE ! ;                                            

: \MES ( -- )    0 #38 At-XY  Clean  Bell ;

   ( Automatic curve calibration)                            

: ?RANGE ( n -- )  
   ?LEAVE @ 0= IF  
      dup 0< IF  
         \MES  ." Oven % too low - Decrease Offset!"  LeaveEarly?  
      THEN           
      $FF > IF  
         \MES  ." Oven % too high - Increase Offset!"  LeaveEarly?   
      THEN  
   ELSE  
      drop  
   THEN 
;  
}

: 1Average ( l -- x )   
   >r  
   REAL-COUNT r@ cells + @ 0= if  
      SPLURGE-COUNT r@ cells + @  0= if  
         0 \ #160 \ >AUTO-OFFSET c@  
         r> drop  exit  
      else     
         SPLURGE-COUNT  
         SPLURGE-VALUES
      then   
   else  
      REAL-COUNT  
      REAL-VALUES
   then                 
   r@ cells 2* + 2@  rot  
   r> cells + @  M/ 
;                 

: 3AVERAGES ( l -- x )   
   >r  
   r@             1Average 8 *   \ The 8 * gives a much higher weighting of the central value  
   r@ 1-   0 max  1Average +     \ and a lower weighting of the values on either side
   r> 1+ $3F min  1Average + 
   $0A /                         \ take the average 
;                               

: !CODE ( diff offset -- ) 
   >r 
   >AUTO-OFFSET c@ +  
   r> 1- AutoCalibrationCurve[] + c!    
;

: !CODES ( -- )
   $1B $03 do  
      I 2* 3AVERAGES  I !CODE    
   loop 
;                          
                                                                
: tt3av ( -- )
   $1B $03 do  
      cr i 4 .dml 2 spaces 
      i 2* 1Average    .
      I 2* 3AVERAGES   .    
   loop 
; 

   ( Binomial pairs)                                         
\ 
: AddToSplurgeCount ( n offset -- )   
   dup $03 $3F 1+ within 
   if  
      cells SPLURGE-COUNT + +!  
   else  
      2drop  
   then 
;                                             
                                                                
: NPAIR ( x offset n' -- )   
   >r 
   2DUP  
   r@ - AddToSplurgeCount 
   r> + AddToSplurgeCount 
;    
                                                                
: NPAIRS ( l -- )   
   >r  
\   $01 r@ 3 NPAIR  
\   $08 r@ 2 NPAIR  
   $10 r> 1 NPAIR 
;   
                                                                
: M+! ( n a -- )   >r  r@ 2@  ROT M+  r> 2! ;                        

: DP1 ( x offset -- )  
   dup 0 $3F 1+ within if  
      cells 2* SPLURGE-VALUES + M+!  
   else  
      2drop 
   then 
;                                                 
                                                                
: DPAIR ( value offset n -- )   
   >r  
   2DUP 
   r@ -  DP1 
   r> +  DP1 
;
                                                                
: DPAIRS ( value offset -- )   
   >r 
   dup r@ 3 DPAIR  
   dup r@ 2 DPAIR  
   $10 *  r> 1 DPAIR 
;                                                 

: Code0ToOffset ( Code0 -- offset )
   NUMBER_OF_DATA_POINTS #1000 */    \ n entries in the table
   \ #10 #64 */
   \ 3 -
   0 max
\   #32 min
;

( Add-entry)                                              
: ADD-ENTRY ( Code0 %Moisture -- )  
\   cr 2dup swap .P1  space  .P1
   8 * 
   swap
   8 *
   >r   
   r@ 1+ 2/ -   4 + #08 /           
\   dup ?RANGE  
   r> 8 / Code0ToOffset
\   8 + $10 /  
   dup NPAIRS 
   dup cells REAL-COUNT +  1 SWAP +!  
   2DUP DPAIRS                      
   cells 2* REAL-VALUES + M+!  
   !CODES 
;                            
                                                                
: LOOSE-END ( -- )  
   REAL-COUNT 2-  BEGIN  cell+ dup @  UNTIL  REAL-COUNT - 2/ >r   
   r@ 1Average 2* ( 2P2)  r@ 1+ 1Average ( P1)  -  ( P3)      
   r@ 1 AND 0= IF  
      r@ 1Average ( P2)  r@ 1+ 1Average ( P1) -  + 
   THEN    
   r>  1+  2/  1 - 
   !CODE 
;                                      
                                                                
: TOP-END ( -- )
   REAL-COUNT  NUMBER_OF_DATA_POINTS cells +  BEGIN  cell- dup @  UNTIL  REAL-COUNT - 2/ >r  
   r@ 1Average 2* ( 2P2')  r@ 1 - 1Average ( P1') - ( P3')     
   r@ 1 AND 0= IF  r@ 1Average ( P2')  r@ 1 - 1Average ( P1') -  +  THEN 
   r>  2/ 1+  !CODE 
; 

( Calibration)                                                  
: 1CURVE-CAL ( -- )
\   96 48 ADD-ENTRY   \ set a dummy point in the bottom left corner of the graph
   vNumberDataPointPairsOrdered @ 0 ?do
      i GetDataPointPairOrdered 
      dup if
         1 0 do  2dup ADD-ENTRY  loop 2drop
      else
         2drop
      then
   loop
\   #31 >SAMPLE 2+ @ #9999 -     \ test to see if there are more points in the next curve's data
;                            

: DataPointPairsAreNotEmpty ( -- f )
   0
   NUMBER_OF_DATAPOINTS 0 ?do
      i GetDataPointPair or
      or
   loop
;

{ 
\ old system - check if there is more data in the next curve
: NCURVE-CAL ( -- )
   CURVE @ dup dup 2>R 
   #13 OVER + SWAP DO 
      I CURVE !
      I J 1+ = IF  ." Linking curve(s) "  THEN  I J - IF  I .  THEN
      1CURVE-CAL IF  LEAVE  THEN  LOOP   2R> drop CURVE ! ;    
}

( Curve straightening)
: TWIST ( a -- )   >R  r@ C@ r@ 1+ C@ -  #10 > IF
      r@ 1+ C@ 1+  #255 MIN r@ 1+ C!  r@ C@ 1- 0 MAX  r@ C!  THEN
   r@ 1+ C@ r@ C@ - 21 > IF
      r@ C@ 1+  #255 MIN r@ C!  r@ 1+ C@ 1- 0 MAX  r@ 1+ C!  THEN
   R> DROP ;

: FIX ( -- ) 
   #16 0 DO  
   AutoCalibrationCurve[] 26 OVER + SWAP  3 + DO  I TWIST  LOOP
   LOOP 
\   AutoCalibrationCurve[] OneCurveBuffer[] #16 + $20 move
;


\ : SMOOTH   \MES  ." Smooth curve"  Y/N-EXIT  FIX ;
: xxxSmoothTheCurve ( -- )  
   #16 0 DO  
   0 >CurveValue 26 OVER + SWAP  3 + DO  I TWIST  LOOP
   LOOP 
;

: xxxNewCurveFit ( -- )
   vNumberDataPointPairsOrdered @ 0 ?do
      i GetDataPointPairOrdered 2dup or 0= if  
         2drop
      else
         cr i .  2dup  . . 
         
         2drop 
      then
   loop
;

\ *****************************************************************************
\ *****************************************************************************

#24  constant X_LEFT
#04  constant Y_TOP

#50 constant X_OFFSET
#16 constant Y_OFFSET

#640 constant X_ACTIVE_SIZE
#480 constant Y_ACTIVE_SIZE

: x_convert ( n -- x )
   #400 X_ACTIVE_SIZE */
;

: y_convert ( n -- x )
   #800 Y_ACTIVE_SIZE */
;

: x_unconvert ( n -- x )
   X_ACTIVE_SIZE #400 */
;

: y_unconvert ( n -- x )
   Y_ACTIVE_SIZE #800 */
;

: Show_'X' ( x y -- )   \ draw an 'X'in the current colour, centred on the given coordinates
   2>r   \ the centre of the glyph
   2r@ 6 5 v-  
   #11 0 do 
      2dup #03 HorizontalLine
      1 1 v+
   loop
   2drop
   \ #10 HorizontalLine 
   2r@ 6 -5 v-  
   #11 0 do 
      2dup #03 HorizontalLine
      1 -1 v+
   loop
   2drop
   2r> 2drop
;

: ShowGraphPoint'X' ( oven% code0 -- )
   \ Y axis
   y_unconvert    \ ... pixels per 1.6% Moisture
   Y_ACTIVE_SIZE swap - 0 max
   Y_TOP +
   Y_OFFSET +
   swap  
   \ X axis
   x_unconvert   \ ... pixels per 1.6 Code0 units
   X_LEFT +
   X_OFFSET +
   swap
   
\   RED BGR>RGB vDrawingForegoundColour !
   PixelAt-XY  WinX @ WinY @ Show_'X'
;

: 4dup  2over 2over ; 

\ isqrt.f  Howerd Oakford  2008 Jun 04
\ Square Root of unsigned 32 bit integer 
\ from James Ulery's Computing Integer Square Roots
\ http://www.azillionmonkeys.com/qed/ulerysqroot.pdf

decimal 

variable Result
variable BitMask

: sqrt ( u - u' ) 
   0 Result !
   $8000 BitMask !
   0 15 do
        BitMask @  Result @ 2*  +  i lshift 
        2dup u< not 
        if
           -
           BitMask @ Result +!
        else
           drop
        then
        BitMask @ 2/ Bitmask !
   -1 +loop
   drop
   Result @
;

{
( 15-bit Square Root of a 31-bit integer)         
CODE SQRT ( n - i)   0 POP   2 2 SUB   2 W MOV   16 # 1 MOV     
   BEGIN   0 SHL   2 RCL   0 SHL   2 RCL   W SHL   W SHL   W INC
      W 2 CMP   CY NOT IF   W 2 SUB   W INC   THEN   W SHR      
   LOOP   W PUSH   NEXT  
}

: distance ( v1 v2 -- n )  \ returns the linear distance between v1 and v2
    rot - dup *  >r - dup * r> +  sqrt
;

: distanceFromXYline ( Code0 Moisture% -- n )   \ signed distance of point from X=Y line 
   2dup 2>r
   over dup 2/ distance   
   2r> 2* < if  negate  then
;

2000 constant CURVATURE_VALUE  \ increase to make less curved

: AddMidPoint ( v1 v2 -- )
   4dup distance dup 5 < if  drop 2drop 2drop  exit  then  \ do not insert mid points if the two points are too close
   >r  
   v+  2/ swap 2/ swap  \ ( -- Code0 Moisture% )
   
   2dup distanceFromXYline >r 

   swap r> r> CURVATURE_VALUE */ + swap
   
   MAGENTA BGR>RGB vDrawingForegoundColour ! 

\   2dup swap ShowGraphPoint'X'  \ ToDo: remove this!!!

   ( Code0 %Moisture -- ) PutDataPointPairOrdered
;

\ 2variable 2v_LastDataPoint

: AddMidPoints ( -- )
   50 0                                    \ start at the origin ( 50 , 0% )
   vNumberDataPointPairsOrdered @ 0 ?do
      i GetDataPointPairOrdered           \ : GetDataPointPairOrdered ( n -- Code0 Moisture% )
         2>r   \ save the current point
         2r@ AddMidPoint  
 \        2r@ 2v_LastDataPoint 2! \ save the latest point for the calculation after the loop
         2r>
   loop
   850 400 AddMidPoint  \ end at the top right corner of the graph
;

0 
   4 struct o_AverageBinNumber
   8 struct o_AverageBinSum
   4 struct o_AverageBinDummy
constant |AverageBinEntry|    

5 2** constant NUMBER_OF_AVERAGES  \ must be a power of 2

NUMBER_OF_AVERAGES |AverageBinEntry| * Buffer: AverageBin[]

: InitAverageBin[] ( -- )
   AverageBin[]  NUMBER_OF_AVERAGES |AverageBinEntry| *  erase
;

: >AverageBin ( n -- a )   NUMBER_OF_AVERAGES 1- and  |AverageBinEntry| * AverageBin[] + ;

: v+! ( v a -- )  \ like  +!  , but works on a pair of values in 2 consecutive cells 
   >r 
   r@ +!
   r@ cell+ +!
   r> drop  
;

: v/ ( v n -- )  \ like  /  , but works on a pair of values  
   >r 
   r@ /
   swap
   r@ /
   swap
   r> drop  
;

: StoreIntoAverageBin ( v n -- )
   dup 1 27 1+ within not if  drop  2drop  exit then
   >AverageBin >r 
          1 r@ o_AverageBinNumber + +!    \ increase the count
   ( v -- ) r@ o_AverageBinSum + v+!      \ increase the sum
   r> drop
;

: LoadAverageBins ( -- )
   InitAverageBin[]
   vNumberDataPointPairsOrdered @ 0 ?do
      i GetDataPointPairOrdered           \ : GetDataPointPairOrdered ( n -- Code0 Moisture% )
      over #16 + #32 /    \ convert code0 in 0.1 units ( 0 to 800 ) to bin capture area 
      StoreIntoAverageBin
   loop
;

: ReadAverageBin ( n -- v )
   >AverageBin >r 
   r@ o_AverageBinSum + 2@          \ fetch the average data pair
   r@ o_AverageBinNumber + @        \ fetch the count
   dup 0= if 
      drop 2drop  0 0  
   else
      \ -rot r@ o_AverageBinNumber + @ 2/ dup v+  rot    \ correct truncation errors
      V/
   then
   r> drop
;

: ShowAverageBin ( -- )
   NUMBER_OF_AVERAGES 0 do
      cr i 3 u.r  2 spaces  i 32 * .P1  2 spaces i ReadAverageBin swap .P1 2 spaces .P1
   loop
;

: !MoistureOffset ( Moisture% bin# -- ) 
   >r
   r@ 16 * -      \ the Moisture% value of the Y=X line
   >AUTO-OFFSET c@ +  
   0 max #255 min
   r> AutoCalibrationCurve[] + c!    
;

: SaveCodeOffsetsFromAverageBinToAutoCalCurve ( -- )
   $1B $03 do
      i ReadAverageBin swap drop i !MoistureOffset
   loop
;

: CopyAutoCalToCurve ( -- )
   $1B $03 do
      AutoCalibrationCurve[] i + c@ 
      i >CurveValue c!
   loop
;

: ConvertDataPointsToCurve ( -- )
   InitAutoCalibration 
   8 0 do 
      AddMidPoints 
      BubbleSort
   loop

   LoadAverageBins

   SaveCodeOffsetsFromAverageBinToAutoCalCurve

   CopyAutoCalToCurve
;


: ttAutoCal ( -- )
   ConvertDataPointsToCurve
   cr
   AutoCalibrationCurve[] $20 3 /string over + swap ?do
      i c@ 2 .hex space 
   loop
;

\ *****************************************************************************
\ *****************************************************************************

\ ******************************************************************************
\ Graph drawing
\ ******************************************************************************
{ 
\ Old new graph
: ShowGraphOuterBorder ( -- )
   X_OFFSET Y_OFFSET  1 1 V- #768  Y_ACTIVE_SIZE  2 2 V+  BLUE BGR>RGB Box
   X_OFFSET Y_OFFSET  2 2 V- #768  Y_ACTIVE_SIZE  4 4 V+  BLUE BGR>RGB Box
   X_OFFSET Y_OFFSET  3 3 V- #768  Y_ACTIVE_SIZE  6 6 V+  BLUE BGR>RGB Box
   X_OFFSET Y_OFFSET  4 4 V- #768  Y_ACTIVE_SIZE  8 8 V+  BLUE BGR>RGB Box
   X_OFFSET Y_OFFSET  5 5 V- #768  Y_ACTIVE_SIZE  10 10 V+  WHITE Box

\   X_OFFSET Y_OFFSET  3 3 V- #768  Y_ACTIVE_SIZE  6 6 V+  YELLOW Box
;
}
{ 
( Grids And Ticks)                                              
: ORIENT ( n a - v)   () @ IF  CELL- @  ELSE                    
      CELL+ @  SWAP  THEN ;                                     
: <AXIS> ( n)   *SCALE 'ORIGIN ORIENT AT  0 FRAME ORIENT        
   +DRAW ;                                                      
                                                                
: GRIDS ( n)   'DRAW @  +CLIP  OFF  SWAP  DUP 1+ -1 DO          
      DUP REACH @  I ROT */  LOW @ +  <AXIS>  LOOP              
   DROP  'DRAW ! ;                                              
                                                                
: NICK ( n)   'DRAW @  -CLIP  SWAP *SCALE 'ORIGIN ORIENT        
   () @  2/ DUP  2 XOR  SWAP V+ AT  () @  DUP 4 XOR             
   SWAP VNEGATE +DRAW  'DRAW ! ;                                
                                                                
: NICKS ( n)   DUP 1+ 0 DO  DUP REACH @  I ROT */  LOW @ +      
      NICK  LOOP  DROP ;                                        
   ( Graph Labels)                                              
: (ADJUST) ( n - n)   DUP  () @ IF  P/C @  2/  SWAP 1+          
      (X) P/C @ *  VNEGATE  (Y)  ELSE  P/C 2@ >R                
   SWAP R> -2 */  THEN  OFF +DRAW ;                             
                                                                
: (LABEL) ( n - a n)   DUP ABS  0  <#  #PLACES @ ?DUP IF        
      0 DO  #  LOOP  [ASCII] . HOLD  THEN  #S  SIGN  #> ;       
                                                                
: LABEL ( n)   DUP *SCALE  'ORIGIN ORIENT AT  'LABEL @EXECUTE   
   (ADJUST) WRITE ;                                             
                                                                
: LABELS ( n)   DUP 1+ 0 DO  DUP  REACH @  I ROT */  LOW @ +    
      LABEL  LOOP  DROP ;                                       
                                                                
: DECIMAL-LABELS   'LABEL IS (LABEL) ;                          
                                                                
   ( Auto-sized Grids & Axes)                                   
: STEP   REACH @  10 / 2 SWAP  DUP DUP ABS /  SWAP 1 DO         
      OVER 3 MOD  0= 5 AND  20 +  10 */  1 M+  DUP I -  +LOOP   
   NEGATE  NIP ;                                                
: STEPS ( - n l h)   STEP >R  LOW @  R@ ABS  OVER REACH @ +     
   OVER / *  R@ 0> IF  SWAP  R@ + SWAP  THEN  R>  ROT ROT ;     
                                                                
: LIMITS ( - h l)   REACH @  LOW @  2DUP < IF  SWAP  THEN ;     
                                                                
: (GRID)   STEPS DO  I <AXIS>  DUP +LOOP  DROP ;                
: AXIS   STEPS DO  I NICK  I LABEL  DUP +LOOP  DROP ;           
                                                                
: AXES   (X) AXIS  (Y) AXIS ;                                   
: GRID   'DRAW @  +CLIP  (X) (GRID)  (Y) (GRID)  'DRAW ! ;      
                                                                
: PAPER   GRID AXES ;  

: G-TITLE   
   \ >G  BLACK WASH  1 \PAL  (X)  1 X-PLACES  1 Y-PLACES
   \ 280 520 SIZE  306 48 ORIGIN  \RANGES
   \ BIG 2 2 !POINT  9 PEN  >CURVE 16 - DUP 16 -TRAILING SWAP DROP
   \ 0 OVER  NEGATE 16 + 8 * 184 + AT  *WRITE
   \ 2 PEN ACROSS (X) CORNER 2@ 14 184 V+ AT  ." Oven Test Result"
   \ DOWN  (X) CORNER 2@ -252 -48 V+ AT  ." Code 0 Reading"
   THIN LINES (X) CORNER 2@ AT  FRAME 2@ +DRAW  GRID
   3 PEN  SMALL  ACROSS  X-AXIS  Y-AXIS  BIG  7 PEN ;
}
\ *****************************************************************************
{
 0 constant o_GRAPH_BACKGROUND_COLOUR
 1 constant o_GRAPH_GRID_COLOUR      
 2 constant o_GRAPH_LABEL_COLOUR     
 3 constant o_GRAPH_AXES_LABEL_COLOUR
 4 constant o_GRAPH_LINE_COLOUR      
 5 constant o_GRAPH_POINTS_COLOUR
 6 constant o_GRAPH_DATA_POINTS_COLOUR
 7 constant o_GRAPH_CODE0s_COLOUR
 8 constant o_GRAPH_OVERLAY_COLOUR_0 
 9 constant o_GRAPH_OVERLAY_COLOUR_1 
10 constant o_GRAPH_OVERLAY_COLOUR_2 
11 constant o_GRAPH_OVERLAY_COLOUR_3 

BLUE    constant c_GRAPH_BACKGROUND_COLOUR
RED     constant c_GRAPH_GRID_COLOUR
RED     constant c_GRAPH_LABEL_COLOUR
GREEN   constant c_GRAPH_AXES_LABEL_COLOUR
YELLOW  constant c_GRAPH_LINE_COLOUR
YELLOW  constant c_GRAPH_POINTS_COLOUR
RED     constant c_GRAPH_DATA_POINTS_COLOUR
CYAN    constant c_GRAPH_CODE0s_COLOUR
MAGENTA constant c_GRAPH_OVERLAY_COLOUR_0
BLACK   constant c_GRAPH_OVERLAY_COLOUR_1                 
GREEN   constant c_GRAPH_OVERLAY_COLOUR_2
RED     constant c_GRAPH_OVERLAY_COLOUR_3
}
0
   4 struct o_GRAPH_BACKGROUND_COLOUR
   4 struct o_GRAPH_GRID_COLOUR      
   4 struct o_GRAPH_LABEL_COLOUR     
   4 struct o_GRAPH_AXES_LABEL_COLOUR
\   4 struct o_GRAPH_LINE_COLOUR      
   4 struct o_GRAPH_POINTS_COLOUR 
   4 struct o_GRAPH_DATA_POINTS_COLOUR 
   4 struct o_GRAPH_CODE0s_COLOUR
   4 struct o_GRAPH_OVERLAY_COLOUR_0 
   4 struct o_GRAPH_OVERLAY_COLOUR_1 
   4 struct o_GRAPH_OVERLAY_COLOUR_2 
   4 struct o_GRAPH_OVERLAY_COLOUR_3 
constant |GRAPH_COLOUR_TABLE|

create GRAPH_COLOUR_TABLE[]
\ create GRAPH_COLOUR_TABLE_WINDOWS[]
WHITE   ,  \ GRAPH_BACKGROUND_COLOUR
BLUE    ,  \ GRAPH_GRID_COLOUR
BLUE    ,  \ GRAPH_LABEL_COLOUR
GREEN   ,  \ GRAPH_AXES_LABEL_COLOUR
\ GREEN   ,  \ GRAPH_LINE_COLOUR
RED     ,  \ GRAPH_POINTS_COLOUR
BLACK   ,  \ GRAPH_DATA_POINTS_COLOUR
CYAN    ,  \ GRAPH_CODE0s_COLOUR
GREEN   ,  \ GRAPH_OVERLAY_COLOUR_0
CYAN    ,  \ GRAPH_OVERLAY_COLOUR_1                 
MAGENTA ,  \ GRAPH_OVERLAY_COLOUR_2
YELLOW  ,  \ GRAPH_OVERLAY_COLOUR_3
\ create GRAPH_COLOUR_TABLE_RETRO_DOS[]
BLUE    ,  \ GRAPH_BACKGROUND_COLOUR
RED     ,  \ GRAPH_GRID_COLOUR
RED     ,  \ GRAPH_LABEL_COLOUR
GREEN   ,  \ GRAPH_AXES_LABEL_COLOUR
\ YELLOW  ,  \ GRAPH_LINE_COLOUR
YELLOW  ,  \ GRAPH_POINTS_COLOUR
RED     ,  \ GRAPH_DATA_POINTS_COLOUR
CYAN    ,  \ GRAPH_CODE0s_COLOUR
GREEN   ,  \ GRAPH_OVERLAY_COLOUR_0
CYAN    ,  \ GRAPH_OVERLAY_COLOUR_1                 
MAGENTA ,  \ GRAPH_OVERLAY_COLOUR_2
RED     ,  \ GRAPH_OVERLAY_COLOUR_3

: >GRAPH_COLOUR ( n -- )   GRAPH_COLOUR_TABLE[] + >RegGraphColourTable @ |GRAPH_COLOUR_TABLE| * +  @ ;
\ : >GRAPH_COLOUR ( n -- )   GRAPH_COLOUR_TABLE_RETRO_DOS[] + @ ;

: GraphColourTableToggle ( -- )
   >RegGraphColourTable @ 0= if  1  else  0  then >RegGraphColourTable ! SetRegValues
;

: InitGraphColourTable ( -- )
   >RegGraphColourTable @ if  1  else  0  then >RegGraphColourTable ! SetRegValues
;

: GRAPH_BACKGROUND_COLOUR  o_GRAPH_BACKGROUND_COLOUR  >GRAPH_COLOUR ;  
: GRAPH_GRID_COLOUR        o_GRAPH_GRID_COLOUR        >GRAPH_COLOUR ;
: GRAPH_LABEL_COLOUR       o_GRAPH_LABEL_COLOUR       >GRAPH_COLOUR ;
: GRAPH_AXES_LABEL_COLOUR  o_GRAPH_AXES_LABEL_COLOUR  >GRAPH_COLOUR ;
\ : GRAPH_LINE_COLOUR        o_GRAPH_LINE_COLOUR        >GRAPH_COLOUR ;
: GRAPH_POINTS_COLOUR      o_GRAPH_POINTS_COLOUR      >GRAPH_COLOUR ;
: GRAPH_DATA_POINTS_COLOUR o_GRAPH_DATA_POINTS_COLOUR >GRAPH_COLOUR ;
: GRAPH_CODE0s_COLOUR      o_GRAPH_CODE0s_COLOUR      >GRAPH_COLOUR ;
: GRAPH_OVERLAY_COLOUR_0   o_GRAPH_OVERLAY_COLOUR_0   >GRAPH_COLOUR ;
: GRAPH_OVERLAY_COLOUR_1   o_GRAPH_OVERLAY_COLOUR_1   >GRAPH_COLOUR ;
: GRAPH_OVERLAY_COLOUR_2   o_GRAPH_OVERLAY_COLOUR_2   >GRAPH_COLOUR ;
: GRAPH_OVERLAY_COLOUR_3   o_GRAPH_OVERLAY_COLOUR_3   >GRAPH_COLOUR ;

: ShowColours ( - )
   cr GRAPH_BACKGROUND_COLOUR $8.hex
   cr GRAPH_GRID_COLOUR       $8.hex
   cr GRAPH_LABEL_COLOUR      $8.hex
   cr GRAPH_AXES_LABEL_COLOUR $8.hex
\   cr GRAPH_LINE_COLOUR       $8.hex
   cr GRAPH_POINTS_COLOUR     $8.hex
   cr GRAPH_CODE0s_COLOUR     $8.hex
   cr GRAPH_OVERLAY_COLOUR_0  $8.hex
   cr GRAPH_OVERLAY_COLOUR_1  $8.hex
   cr GRAPH_OVERLAY_COLOUR_2  $8.hex
   cr GRAPH_OVERLAY_COLOUR_3  $8.hex
;

\ *****************************************************************************

: HorizontalLine+ ( x y length  -- )
   0 max X_ACTIVE_SIZE min 
   >r  swap X_OFFSET + X_LEFT + swap  Y_ACTIVE_SIZE swap -  Y_TOP +  Y_OFFSET +  r> HorizontalLine
;

: VerticalLine+ ( x y length  -- )
   0 max Y_ACTIVE_SIZE min
   >r  swap X_OFFSET + X_LEFT + swap  Y_TOP +  Y_OFFSET +  r> VerticalLine
;

: DrawY=X ( -- )
   X_ACTIVE_SIZE 0 do
      vDrawingForegoundColour @ 
      X_OFFSET X_LEFT + i +   
      0  X_ACTIVE_SIZE 1- i -  Y_ACTIVE_SIZE  X_ACTIVE_SIZE */ -  >Pixel !
   loop
;

: (ShowFilenameCamelOnGraph) ( y -- )
   >r
   GRAPH_BACKGROUND_COLOUR BGR>RGB vTextBackgroundColour !
   #21 JustFilename[] count swap drop - 0 max #06 + 16 * r> PixelAt-XY  
   JustFilename[] count ShowFilenameCamel-
;

: ShowGraphGrid ( -- )
\   X_LEFT  Y_TOP  X_ACTIVE_SIZE X_OFFSET 2* +  Y_ACTIVE_SIZE Y_OFFSET 2* +  
\   GRAPH_BACKGROUND_COLOUR BGR>RGB Rectangle
   
   GRAPH_BACKGROUND_COLOUR wash 
   
   \ The y = x line
   GRAPH_GRID_COLOUR BGR>RGB vDrawingForegoundColour !
   X_LEFT X_OFFSET + Y_TOP Y_OFFSET +  Y_ACTIVE_SIZE + PixelAt-XY  
   X_ACTIVE_SIZE  Y_ACTIVE_SIZE negate lineRelative

   X_ACTIVE_SIZE 1+ 0 do
      i 0 Y_ACTIVE_SIZE VerticalLine+
   X_ACTIVE_SIZE 8 /  +loop

   GRAPH_GRID_COLOUR BGR>RGB vDrawingForegoundColour !
   Y_ACTIVE_SIZE 1+ 0 do
      0 i X_ACTIVE_SIZE HorizontalLine+
   Y_ACTIVE_SIZE 8 /  +loop

   \ DrawY=X
;

: FlipYaxis ( v -- v' ) 
   Y_ACTIVE_SIZE swap - 
   X_OFFSET Y_OFFSET V+
   X_LEFT Y_TOP V+
;

: ShowGraphPoint5x5 ( x y -- )
   2 2 V- 5 5  
   vDrawingForegoundColour @ Rectangle
;

: SetGraphTextColour ( n -- )   
   BGR>RGB vTextForegroundColour !
   GRAPH_BACKGROUND_COLOUR BGR>RGB vTextBackgroundColour !  
;

\ *****************************************************************************
\ Display the DataPointPair array
\ *****************************************************************************

variable vGraphScaling  \ 1 , 2 or 4 divider for x and y axes, and number of points displayed, scale up for values
1 vGraphScaling ! 

: VscaleUp ( v -- v' )   vGraphScaling @ * swap vGraphScaling @ * swap ;

create Scales[]  1 c, 2 c, 4 c, 1 c,

variable vGraphScalingIndex

: SetScale ( -- )   vGraphScalingIndex @ 3 and Scales[] + c@ vGraphScaling ! ;

: InitScale ( -- )  0 vGraphScalingIndex !  SetScale ;

: +Scale ( -- )   vGraphScalingIndex @ 1+ dup 3 = if  drop 0  then  vGraphScalingIndex !  SetScale ;

: -Scale ( -- )   vGraphScalingIndex @ 1- dup 1 < if  drop 3  then  vGraphScalingIndex !   SetScale ;

: ShowDataPointPair ( oven% code0 -- )
   swap 
   ShowGraphPoint'X'    
;

variable vShowDataPoints


: ShowAllDataPoints ( -- )    \ including the calculated mid points. Test only !!!
   MAGENTA BGR>RGB vDrawingForegoundColour ! 
\   GRAPH_DATA_POINTS_COLOUR SetGraphTextColour

   vShowDataPoints @ 0= if  exit  then 
   vNumberDataPointPairsOrdered @ 0 ?do
      i GetDataPointPairOrdered 
      dup if
         VscaleUp ShowDataPointPair
      else
         2drop
      then
   loop

\   AddMidPoints
;

: ShowDataPoints ( -- )
   GRAPH_DATA_POINTS_COLOUR BGR>RGB vDrawingForegoundColour ! 
\   GRAPH_DATA_POINTS_COLOUR SetGraphTextColour

   vShowDataPoints @ 0= if  exit  then 
   NUMBER_OF_DATAPOINTS 0 do
      i GetDataPointPair 
      dup if
         VscaleUp ShowDataPointPair
      else
         2drop
      then
   loop
;


variable PointsFileExists?

#36 constant VALUES_TOP_LINE


: 3U.M ( c -- )   \ display a 1 to 3 digit value in the middle of the field
   3 (u.r) 
   -Leading -Trailing
   case
   0 of  endof
   1 of  space  1 type  space  endof
   2 of  2 type  space  endof
   3 of  3 type  endof
   drop  3 spaces
   endcase
;

variable CurrentValue
variable CurrentIndex

: IndexToPosition ( index -- col row )
   dup 3 #25 1+ within if
      3 - 0 max  4 * 2 +  #39    
      exit 
   then
   dup #26 #31 1+ within if 
      #22 -  \ join onto the 0 to 3 indices   
   then
   10 * 2 -  #44
;
variable CurrentKeypress

: EnterCurveValue ( i -- )

   MAGENTA.  4 #41 At-XY  ." Press 0 to 9 , Enter , or Esc to leave the value unchanged : "  RESTORE. 
   DoNotOverwriteGraph @ >r
   -1 DoNotOverwriteGraph !

   CurrentIndex !
   CurrentIndex @ >CurveValue c@ CurrentValue ! 

   begin
      CurrentIndex @ IndexToPosition At-XY 
      MAGENTA.  CurrentValue @ 3u.m  RESTORE. 
      ekey CurrentKeypress ! 
      CurrentKeypress @ [char] 0 [char] 9 1+ within if  
         CurrentKeypress @ [char] 0 -
         CurrentValue @ 10 * 1000 mod  +  CurrentValue !    
      then
      CurrentKeypress @ $08 = if   \ Backspace
         CurrentValue @ 10 / CurrentValue ! 
      then
      CurrentKeypress @ $1002E = if   \ Delete key
         CurrentIndex @ #31 = if 
            #20 CurrentValue !     \ set the default Temperature Coefficient = 20
         else   
            #160 CurrentValue !     \ set the default 160 value
         then
      then

      CurrentKeypress @ $0D = if   \ Enter key ( carriage return )
         CurrentValue @ 0 max 255 min  CurrentIndex @ >CurveValue c!
         $1B CurrentKeypress !    \ exit the loop now
      then
      CurrentKeypress @ $1B =    \ Escape key  
   until 

   r> DoNotOverwriteGraph !

   4 #41 At-XY  Clean
;

: GetCurveOffset ( -- c )   #30 GetCurveValue ;
: GetCurveOffsetNegated ( -- c )   160 GetCurveOffset 160 - - ;

2variable vLastPointCoordinates

: ShowGraphPointsInColour ( -- )

   vTextForegroundColour @ vDrawingForegoundColour !

   
   0 0 FlipYaxis  swap WinX ! WinY ! 

   #26 vGraphScaling @ / 3 do  
      i X_ACTIVE_SIZE #25 */   \ X pixel
      i GetCurveValue  GetCurveOffset -  \ 0.1% oven test result 0 to 40.0
      #40 #25 */ 
      +           
      i Y_ACTIVE_SIZE #25 */ 
      \ Y value to display
      VscaleUp
      FlipYaxis
      2dup 2>r 
      ShowGraphPoint5x5
\      GRAPH_LINE_COLOUR BGR>RGB vDrawingForegoundColour !
      2r@ lineBresenham
\      vTextForegroundColour @ vDrawingForegoundColour !
      2r> WinY ! WinX !
   loop   
;

: ShowGraphPoints ( -- )
   GRAPH_POINTS_COLOUR BGR>RGB vTextForegroundColour !
   ShowGraphPointsInColour  
   #36 (ShowFilenameCamelOnGraph)  
\   RESTORE.
;

: ShowGraphPointsOverlay ( -- )
   vOverlayIndex @ cNumberOfOverlays 1- and  dup >r case
   0 of  GRAPH_OVERLAY_COLOUR_0  endof
   1 of  GRAPH_OVERLAY_COLOUR_1  endof
   2 of  GRAPH_OVERLAY_COLOUR_2  endof
   3 of  GRAPH_OVERLAY_COLOUR_3  endof
      GRAPH_OVERLAY_COLOUR_0
   endcase
   BGR>RGB vTextForegroundColour !
      
   ShowGraphPointsInColour
   #36 #32 + r> #18 * + 
   >r
   GRAPH_BACKGROUND_COLOUR BGR>RGB vTextBackgroundColour !
\   #21 JustFilename[] count swap drop - 0 max #06 + 16 * r> PixelAt-XY  \ centered
   #21 8 12 * + r> PixelAt-XY      \ left justified

   Font12x16x1 FontNew   
   JustFilename[] count -Trailing 0 max $10 min type
   FontRestore
;


\ set the pixel coordinate for the top-left corner of the glyph to draw
: SetPixelXY ( x y -- )     
   Y_TOP + 2+ WinY !  
   X_LEFT + WinX !
;

\ set the character coordinate for the glyph to draw
: GraphAt-XY ( col row -- )
   vPixelsPerGlyphY @ * 2 + WinY !  vPixelsPerGlyphX @ * WinX !
;

: GraphTypeHorizontal ( $ -- )
   over + swap ?do  i c@ emit(f)  vPixelsPerGlyphX @ WinX +!  loop 
;

: GraphTypeVertical ( $ -- )
   over + swap ?do  i c@ emit(f)  vPixelsPerGlyphY @ WinY +!  vPixelsPerGlyphX @ negate  WinX +!  loop 
;

: ShowGraphAxisTitles ( -- )
   ScreenResource get

\   Font24x32ubuntu
\   Font9X14
\   Font8x16

   decimal

   GRAPH_LABEL_COLOUR SetGraphTextColour 

   #30 8 * 4 + #29 #16 * PixelAt-XY  s" Oven Test Result %" GraphTypeHorizontal 
   #2 8 *      #09 #16 * PixelAt-XY  s" Code 0 Reading"   GraphTypeVertical 

   GRAPH_AXES_LABEL_COLOUR SetGraphTextColour 

   \ Y axis labels
   Y_ACTIVE_SIZE 1+ 0 do  
      #16 ( x -- )  Y_ACTIVE_SIZE i - Y_TOP +  #04 + ( -- y )  SetPixelXY  i y_convert vGraphScaling @ / .P1
   Y_ACTIVE_SIZE 8 / +loop

   \ X axis labels
   X_ACTIVE_SIZE 1+ 0 do 
      i 0= not if 
         i X_LEFT + #08 + ( x -- )  #488 ( -- y )  SetPixelXY  i x_convert vGraphScaling @ / 4 R.P1 
      then
   X_ACTIVE_SIZE 8 / +loop

   GRAPH_CODE0s_COLOUR SetGraphTextColour 
   #26 3 do  
      X_ACTIVE_SIZE  X_LEFT +  #54 +                   \ X pixel coordinate
      #26 i - Y_ACTIVE_SIZE #25 */  Y_TOP +   #08 -    \ Y pixel coordinate
      PixelAt-XY 
      i vGraphScaling @ / 3 U.R
   vGraphScaling @ +loop  

   ScreenResource release
;

: ShowGraph ( -- )

   ShowGraphGrid
   ShowGraphPoints
\   ShowAllDataPoints   \ Test only!!!
   ShowDataPoints

   ShowGraphAxisTitles 
;

$20 Buffer: TempBuffer[]

: GraphPreview ( -- )
   0 >CurveValue  TempBuffer[] $20 move    
   ConvertDataPointsToCurve
   ShowGraph
   TempBuffer[] 0 >CurveValue  $20 move    
;

\ *****************************************************************************
\ Convert the Data Point Array into a Data Points file
\ *****************************************************************************
{ 

$1000 Buffer: MySavedPointsFile[]

: AddPointsLine ( n n -- )
   swap 
   (.P1) MySavedPointsFile[] append
   s"   " MySavedPointsFile[] append
   (.P1) MySavedPointsFile[] append
   $CRLF count MySavedPointsFile[] append
;
}


\ *****************************************************************************
\ Curve Points file read
\ read the current CurvePoints into the file buffer
\ *****************************************************************************

variable vCurvePointsSize
$10000 constant |OneCurvePointsBuffer|
|OneCurvePointsBuffer| Buffer: OneCurvePointsBuffer[]
\ variable vSavedCurvePointsSize
|OneCurvePointsBuffer| Buffer: SavedOneCurvePointsBuffer[]
|OneCurvePointsBuffer| Buffer: SavedOneCurvePointsBuffer2[]

: SaveCurvePointsCopy ( -- )
   OneCurvePointsBuffer[] SavedOneCurvePointsBuffer[] |OneCurvePointsBuffer| cmove  \ save the file
;

: RestoreCurvePointsCopy ( -- )
   SavedOneCurvePointsBuffer[] OneCurvePointsBuffer[] |OneCurvePointsBuffer| cmove  \ save the file
;

: SaveCurvePointsCopy2 ( -- )
   OneCurvePointsBuffer[] SavedOneCurvePointsBuffer2[] |OneCurvePointsBuffer| cmove  \ save the file
;

: RestoreCurvePointsCopy2 ( -- )
   SavedOneCurvePointsBuffer2[] OneCurvePointsBuffer[] |OneCurvePointsBuffer| cmove  \ save the file
;

: CurvePointsHaveChanged? ( -- f )
   OneCurvePointsBuffer[] |OneCurvePointsBuffer| SavedOneCurvePointsBuffer[] |OneCurvePointsBuffer| compare 0= not
;

MAX_PATH Buffer: SavedCurrentCurveFilename[]

: SaveCurveFilename ( -- )
  CurrentCurveName count MAX_PATH 1- min  SavedCurrentCurveFilename[] place 
;

: RestoreCurveFilename ( -- )
   SavedCurrentCurveFilename[] count MAX_PATH 1- min  CurrentCurveName place 
;

MAX_PATH Buffer: SavedCurrentCurvePointsFilename[]

: SaveCurvePointsFilename ( -- )
  CurrentPointsName count MAX_PATH 1- min  SavedCurrentCurvePointsFilename[] place 
;

: RestoreCurvePointsFilename ( -- )
   SavedCurrentCurvePointsFilename[] count MAX_PATH 1- min  CurrentPointsName place 
;

variable vSelectOverlayEnable

: ShowOverlay ( -- )  
   SaveCurveCopy2
   SaveCurvePointsCopy2
   SaveCurveFilename
   SaveCurvePointsFilename
   
   vSelectOverlayEnable @ if
      SelectCurveFile-
   then

   vOverlayIndex @ >r 

   cNumberOfOverlays 0 ?do
      i vOverlayIndex ! 
      i >OverlayFilename count dup if
         CurrentCurveName MAX_PATH BL fill 0 CurrentCurveName c!
         CurrentCurveName place
         (GetCurveFile-overlay)
         ShowGraphPointsOverlay
      else
          2drop
      then      
   loop

   r> vOverlayIndex !

   RestoreCurveFilename
   RestoreCurvePointsFilename

   UpdateJustFilenamesFromCurveName
   RestoreCurvePointsCopy2
   RestoreCurveCopy2

   0 vSelectOverlayEnable !   \ only select a new overlayh file if the Overlay button has been pressed

   \ re-draw the main curve, on top
   ShowGraphPoints
;

: ShowFilenames ( -- )
   cr ." CurrentCurveName  = " CurrentCurveName count type
   cr ." CurrentPointsName = " CurrentPointsName count type
   cr ." JustFilename[]    = " JustFilename[] count type
   cr ." JustPath[]        = " JustPath[]     count type
   cr
   cr ." CurrentCurveSetName    = " CurrentCurveSetName    count type
   cr ." JustCurveSetFilename[] = " JustCurveSetFilename[] count type
   cr ." JustCurveSetPath[]     = " JustCurveSetPath[]     count type
;

variable vCurvePointsFileHandle

\ *****************************************************************************
\ Convert Data Point file into DataPointPair array
\ *****************************************************************************

\ process the accumulated line buffer
: ProcessDataPointPairLine ( c-addr u -- )
   over c@ [char] \ = if  2drop  exit  then  \ comment line
   over c@ [char] # = if  2drop  exit  then  \ comment line
\   cr -Trailing 2dup type
   -Leading -Trailing  BL split 
   -Leading -Trailing  ['] Dnumber catch if  2drop  0  else  drop  then  >r
   -Leading -Trailing  ['] Dnumber catch if  2drop  0  else  drop  then
   r> 
   2dup 0= swap 0= or if
      2drop    \ ignore this line if either value is 0
   else
      PutDataPointPair
   then
;

: ttProcessDataPointPairLine ( -- )
\   s" 12.3" ['] Dnumber catch if  2drop  else  space drop  10 u.r  then
   s"    12.3   26.6   " ProcessDataPointPairLine
;

9 2** constant |DataPointPair|       \ the size of our line buffer - must be a power of 2
|DataPointPair| Buffer: DataPointPair[]     \ our send buffer
variable #DataPointPair              \ how many bytes to send so far
variable #LastDataPointPair          \ how many bytes to send so far


: ValidChar? ( c -- f )
   >r
   r@ [char] \ =
   r@ [char] # = or
   r@ [char] . = or 
   r@ [char] 0 [char] 9 1+ within or
   r> drop
;

: DataPointPairEmit ( c -- )
   dup ValidChar? not if  drop  BL  then    \ ignore commas etc., replace by spaces
   DataPointPair[] #DataPointPair @ |DataPointPair| 1- and + c!    \ store the character in the send buffer
   1 #DataPointPair +!             \ increment the count
;

: InitDataPointPairEmit ( -- )   0 #DataPointPair ! ;

\ accumulate characters, act on Backspace, ctrlC, Escape and Enter
\ Note the  ProcessDataPointPairLine  which does all of the work when you read a <CR>
: DataPointPairLineEmit ( c -- )
\   dup emit
   dup <CR> = if  drop  exit  then  \ ignore CRs
   dup <LF> = if
      drop
      #DataPointPair @ #LastDataPointPair !
      #DataPointPair @ if
         DataPointPair[] #DataPointPair @ ProcessDataPointPairLine 
         InitDataPointPairEmit
      then
   else  
\   dup $08 = if  08 emit  bl emit  08 emit  #DataPointPair @ 1- 0 max  #DataPointPair !  drop  exit  then
\   dup <ESC> = if  cr ." Done."  #1006 throw  then 
\   dup $03 = if  cr ." ctrlC"  #1007 throw  then 
\   dup emit
      DataPointPairEmit
   then
;

: ConvertFileToDataPointArray ( -- )

   DataPointArrayInit

   OneCurvePointsBuffer[]  vCurvePointsSize @ 
   over + swap ?do
      i c@ DataPointPairLineEmit
   loop
   <CR> DataPointPairLineEmit    \ in case the last line was not terminated with a CR

   0 vNoDataPointFileLoaded !
;

\ *****************************************************************************
\ Edit array, used by F1 to change values before (optionally) saving to a file
\ *****************************************************************************
0 
   cell struct oOven% 
   cell struct oCode0 
\   cell struct oOven%_undo 
\   cell struct oCode0_undo
constant |Datum_F1|   

|Datum_F1| NUMBER_OF_DATAPOINTS * constant |Data_F1| 
|Data_F1| Buffer: Data_F1[]
|Data_F1| Buffer: SavedData_F1[]

Data_F1[] |Data_F1| erase
SavedData_F1[] |Data_F1| erase

|Data_F1| Buffer: SavedData_F1_2[]

: SaveDataPoints_F1Copy ( -- )
   Data_F1[] SavedData_F1_2[] |Data_F1| cmove  \ save the file
;

: RestoreDataPoints_F1Copy ( -- )
   SavedData_F1_2[] DataPointPairs[] |Data_F1| cmove  \ save the file
;

: DataPoints_F1HaveChanged? ( -- f )
   DataPointPairs[] |Data_F1| SavedData_F1_2[] |Data_F1| compare 0= not
;

: LoadEditBuffer_F1 ( -- )
   DataPointPairs[] Data_F1[] |Data_F1| move
   Data_F1[] SavedData_F1[] |Data_F1| move
;

\ : DataF1_changed? ( -- f )
\    Data_F1[] |Data_F1| SavedData_F1[] |Data_F1| compare    
\ ;

\ copy the newly edited values to the main array
: SaveEditBuffer_F1 ( -- )
   Data_F1[] DataPointPairs[] |Data_F1| move
;

variable Data_F1_Ptr          \ which data item we are pointing to
variable Data_F1_HighLight    \ which data item we are pointing to

: >Datum_F1 ( -- a )   Data_F1[]  Data_F1_Ptr @ 2/ |Datum_F1| * + ;

: Oven%      ( -- a )      >Datum_F1 oOven%      + ;   
: Code0_F1      ( -- a )   >Datum_F1 oCode0      + ;  
\ : Oven%_undo ( -- a )      >Datum_F1 oOven%_undo + ;  
\ : Code0_F1_undo ( -- a )   >Datum_F1 oCode0_undo + ; 

$1000 Buffer: MyBigAppendBuffer[]

: SavePointsFile_F1 ( -- )   \ 

   CurrentPointsName count swap drop 0= if  exit  then 
   
   SaveEditBuffer_F1    \ copy the edited values to the main array, before saving them to the file

   CopyPointsToCurveFilename
      
\   BigAppendInit
   s" \ %moisture Code0   Points data for file : " MyBigAppendBuffer[] BigPlace
   CurrentCurveName count -Ext MyBigAppendBuffer[] BigAppend  
   
   $CRLF count MyBigAppendBuffer[] BigAppend
   
   NUMBER_OF_DATAPOINTS 0 do
      i GetDataPointPair 
      (.P1) MyBigAppendBuffer[] BigAppend  
      s"  , " MyBigAppendBuffer[] BigAppend
      (.P1) MyBigAppendBuffer[] BigAppend 
       $CRLF count MyBigAppendBuffer[] BigAppend
   loop

\   CurrentCurveName count -Ext CurrentPointsName place
\   s" .SinarPoints" CurrentPointsName append

   MyBigAppendBuffer[] BigCount MyFileSize !
   MyFileBuffer[] MyFileSize @ move
   CurrentPointsName count WriteMyFile
;

: xxxCheckNeedToTransfer ( -- )
   DataPoints_F1HaveChanged? 
   if
      -1 v_NeedToTransferToCurve !
\      z" Press Ok to save the Points" z" Points have been edited" Dialogue
\      1 = if 
\         SavePointsFile_F1 
\         SaveDataPoints_F1Copy
\      then
   then
;

: SavePointsFileIfEdited? ( -- )
   SaveEditBuffer_F1    \ copy the newly edited values to the main array
   DataPoints_F1HaveChanged? 
   if
      z" Press Ok to save the Points" z" Points have been edited" Dialogue
      1 = if 
         -1 v_NeedToTransferToCurve !
         0 v_PointsTransferredToCurve !
         SavePointsFile_F1 
         SaveDataPoints_F1Copy
      then
   then
;

: ClearAllDataPointPairs ( -- )
   0 vNumberDataPointPairs !
   NUMBER_OF_DATAPOINTS 0 do
      0 0 PutDataPointPair
   loop
   LoadEditBuffer_F1
   SaveDataPoints_F1Copy
;

: CreateBlankPointsFile_F1 ( -- )   
   ClearAllDataPointPairs
   SavePointsFile_F1
;

: ttCreateBlank ( -- )
   S" MyBlankPointsfile.SinarCurve" CurrentCurveName place
   CreateBlankPointsFile_F1
;

: TransferToCurve ( -- )     
   SaveCurveCopy2 
   CopyPointsToCurveFilename
   CurrentCurveName count LastCurrentCurveName[] place
   DefaultCurve  0 >CurveValue $20 move   \ set bias to correct value
   ConvertDataPointsToCurve
;


: TransferToCurve? ( -- )
   Font12x16x1 FontNew

   SavePointsFileIfEdited?

   DataPointPairsAreNotEmpty ( v_NeedToTransferToCurve @ and ) if 
      z" Overwrite curve with new data?" z" Press Ok to continue..." Dialogue
      1 = if
         SavePointsFile_F1
         TransferToCurve
         CurrentCurveName count SaveCurveFile_F2  
         
         0 v_NeedToTransferToCurve !
         -1 v_PointsTransferredToCurve !
         \ [ 0 CHAR_n B ] literal  ForceKeyPress
\         [ 2  CHAR_n H ] literal  ForceKeyPress
      else
         s" Transfer cancelled" 2 4 ShowMessage
      then
   else
      DataPointPairsAreNotEmpty if
         \ nothing changed, so no need to transfer data to curve
      else
         v_NeedToTransferToCurve @ if
         \   s" No data points to transfer! " 2 4 ShowMessage
         then
      then
   then
   FontRestore
;

: TransferToCurve?? ( -- )
   SavePointsFileIfEdited?
   v_PointsTransferredToCurve @ 0= if
      TransferToCurve?
   then
\   0 v_PointsTransferredToCurve !   \ mark as not transferred now, as we are exiting F1
;

: ttShowDataPointPairs
   NUMBER_OF_DATAPOINTS 0 do
      i GetDataPointPair 
      dup if
         cr swap .P1  space  .P1
      else
         2drop
      then
   loop
;

: GetPointsFile() ( -- )  
   CurrentPointsName count  r/o open-file if  
      4 yMSG<  RED. ." Could not find the SinarPoints file : "  RESTORE. CurrentPointsName count type   >MSG  exit \ #9010 throw   
   then
   vCurvePointsFileHandle !
   OneCurvePointsBuffer[] |OneCurvePointsBuffer| vCurvePointsFileHandle @ read-file  if 
      cr ." GetPointsFile could not read from the file = "  CurrentPointsName count type  space  #9011 throw  
   then
   vCurvePointsSize !
   vCurvePointsFileHandle @ close-file drop

   ConvertFileToDataPointArray

   LoadEditBuffer_F1
   SaveDataPoints_F1Copy
;

: SelectPointsFile ( -- )
   MNET8-Points DEFAULT-OPEN-FLAGS  s" .\Curves" ChooseFile+ dup if
      CurrentPointsName place
      CopyPointsToCurveFilename
      SetRegValues
\      GetPointsFile()   \ from file to OneCurvePointsBuffer[] - done later
   else
      2drop
   then
   UpdateJustFilenamesFromCurveName
;


: GetPointsFile ( -- )  
   
   CurrentCurveName count FILE-STATUS swap drop 0= if    \ exists

   else
      exit  \ do nothing if we don't have a curve file
   then

   UpdateJustFilenamesFromCurveName

   CurrentPointsName count FILE-STATUS swap drop 0= 
   if    \ the file exists
      ['] GetPointsFile() catch if  4 yMSG< ." 1111 >>>>>>>>>>>>>>><><><> "  >MSG  then
   else  \ the file does not exist
      z" Press OK to create new, empty, Points file" z" Press Ok to continue..." Dialogue
      1 = if
         CreateBlankPointsFile_F1
      else
         SelectPointsFile
      then
      ['] GetPointsFile() catch if  4 yMSG< ." 2222 >>>>>>>>>>>>>>><><><> " >MSG  then
   then
;

: ttShowDataPointPairs+
   SelectPointsFile
   ttShowDataPointPairs
;

: CopyCurrentCurveFileToCurveSet ( n -- )
   >r 
   JustFilename[] count  r@ >CurveName  dup $10 $20 fill  swap move  \ copy the name
   OneCurveBuffer[] #16 +    r@ >CurveName #16 + $20 move  \ skip the #16 byte curve name to copy the curve data
   OneCurvePointsBuffer[]    r@ >CurveName $80 + $80 move  \ copy 128 bytes of the Points data
   r> drop
;

variable vCurveSetHasBeenEdited

: EditCurveInCurveSet ( -- )
   SelectCurveFile   \ select the Curve to put into the CurveSet
   GetCurveFile()    \ get the curve data for the currently selected Curve filename
   CopyCurveToPointsFilename  \ setup the associated Points filename      
   CurrentPointsName count FILE-STATUS swap drop 0= 
   if    \ the file exists
      ['] GetPointsFile() catch if  4 yMSG< ." 3333 >>>>>>>>>>>>>>><><><> "  >MSG  then
   then  \ the file does not exist, so do nothing - the points area was cleared by GetCurveFile() 
   EditCurveNumber @ CopyCurrentCurveFileToCurveSet
;

\ *****************************************************************************
\ Show a curve
\ *****************************************************************************

: ShowCurve ( -- )
   page   
   
   NORMAL.

   CurrentCurveName count swap drop 6 < if  RESTORE.  exit  then 
  
   CurrentPointsName count FILE-STATUS swap drop 0= if  \ the file exists

   else
      0 vCurvePointsSize !    \ do not display the last file we read... 
   then

   ShowGraph

\   UpdateTextWindow
;

\ ***** program end *****
