\ Help.F  HPO  2014 Dec 21  Display a help file window

\ convert character position to pixel position
: c>p ( c l - x y )   char-height * 6 + swap  char-width * 6 + swap ;
: p>c ( x y - c l )   char-height / swap  char-width / swap ;

defer Htype \ x y a n Htype
: (Htype)  2swap  p>c  gotoXY type ;
: norm   ['] (Htype) is Htype ;

\ variable c# \ the column number
\ variable l# \ the line number
8 constant charX \ character width in pixels
16 constant charY \ character height in pixels

defer Wtype  \ Windows text type
' (Htype) is Wtype \ default console version

: Wcr   l# @ 1+ 25 min  0 max  l# !  0 c# ! ;

: Wpage   0 l# !  0 c# ! 


;

defer Hpage  \
' Wpage is Hpage \ default console version

defer Hcr \
' Wcr is Hcr \ default console version

: HelpText   Wpage
    Wcr  Wcr
    Wcr   s"                  Sinar F4 Help Screen "  Wtype
    Wcr   s"                    V3.4 2022 Jul 31   "  Wtype
    Wcr   s"                  ==================== "  Wtype
    Wcr  
    Wcr   s"       1. Hold down the F3 or kg/hl key and turn on the" Wtype 
    Wcr   s"          Sinar Moisture Meter"  Wtype
    Wcr   s"       2. Select the COM port on your PC or laptop "  Wtype
    Wcr   s"       3. Select Download (to the meter) or Upload (from the meter) : "  Wtype
    Wcr   s"        "  Wtype
    Wcr   s"       4. If Downloading to the meter, Select the File "  Wtype
    Wcr   s"       5. Click on the Start button to start the transfer. "  Wtype
    Wcr
    Wcr   s"       When Uploading, the File is always saved as NewUpload.cvs "  Wtype
    Wcr
    Wcr   s"       Press the Stop button to stop the transfer "  Wtype
;

variable TextBuffer $100 allot

\ : Append   +place ;

: HelpCom ( -- )  
   Wpage
   Wcr  Wcr
   Wcr   s" Cannot open serial port COM" TextBuffer place
   ComPort (.) TextBuffer append  s"  !!!"  TextBuffer append
   TextBuffer count Wtype
   Wcr
   Wcr   s" Please check that it is not being used by another application."  Wtype
   Wcr   s" If it is, you must close that application to allow this program" Wtype
   Wcr   s" to use the port..." Wtype
   Wcr Wcr
   Wcr   s" Close this window to continue... "
;

:object HelpWindow <super window

Font fdFont \ font data structure

int screen-cols
int screen-rows
int SlowCounter         \ increments when the window is updated

int cur-first-line \ current first line position
0 value first-line#    \ first line number
200 value last-line#   \ last line number
last-line# 20 - value last-top-line#
16 value bytes/line

: Wtype() ( a n )
   c# @ l# @ c>p
   2swap  TextOut: dc
;

:m home: ( -- )
   first-line# to cur-first-line
   paint: self
;m

:M On_Init:     ( -- )          \ things to do at the start of window creation

   ['] Wtype() is Wtype		\ set up the type function

   On_Init: super             \ do anything superclass needs

    12 to char-width
    16 to char-height

   charX  Width: fdFont
   charY Height: fdFont
   Create: fdFont
   s" SYSTEM_FONT" SetFaceName: fdFont
\   s" Freestyle Script" SetFaceName: fdFont
\   s"       Courier" SetFaceName: fdFont
   s" Sinar F4 Help"  SetTitle: self
   on_paint: self
;M

:m on_move: ( w -- )
   On_Paint: self
;m

:m startpos: ( StartPos: parent 20 20 v+ ) 22 100 ;m

:m startsize: 68 charX *  20 charY * ;m

:m on_paint: ( -- )
   SaveDC: dc                      \ save device context

   Handle: fdFont SetFont: dc      \ set the font to be used
\   s"  Arial" SetFaceName: fdFont
   BLUE SetTextColor: dc

   HelpText
 
   RestoreDC: dc	
;m						     

(( 
:m WM_PAINT ( -- )
   SaveDC: dc                   \ save device context
   SetupFont
\   .ComData 			\ update the status bar
   DisplayCurves
   RestoreDC: dc	
\   WM_PAINT super
;m
))

:M WM_LBUTTONDOWN     ( -- res )
   On_Done: self
   DestroyWindow: self
   0
;M

:M WM_RBUTTONDOWN     ( -- res )
   On_Done: self
   DestroyWindow: self
   0
;M
   

:M On_Done:    ( h m w l -- res )
    Delete: fdFont			\ remove the font
\   On_Done: Super	\ the class's close function
\   DestroyWindow: self
   0 
;M

:M WindowStyle: ( -- style )            \ return the window style
                WindowStyle: super
		WS_BORDER  or
                ;M
		   
:M ExWindowStyle: ( -- extended_style )
                ExWindowStyle: super
                WS_EX_TOPMOST or        \ if so, lock on top
;M

((
:m vposition: ( n -- ) \ move to position n
        0max last-top-line# min
        to cur-first-line paint: self ;m

:m vscroll: ( n -- ) \ move n lines up or down
        cur-first-line + vposition: self ;m

:m end: ( -- ) \ move to end, in this case it's 100 bytes down to pad
        last-top-line# to cur-first-line paint: self ;m

:m vpage: ( n -- )  \ down or up n pages
        screen-rows 1- * vscroll: self ;m

:M WM_VSCROLL   ( h m w l -- res )
                swap word-split >r
        CASE
                SB_BOTTOM        of          End: self   endof
                SB_TOP           of         Home: self   endof
                SB_LINEDOWN      of    1 VScroll: self   endof
                SB_LINEUP        of   -1 VScroll: self   endof
                SB_PAGEDOWN      of    1   VPage: self   endof
                SB_PAGEUP        of   -1   VPage: self   endof
                SB_THUMBPOSITION of r@ VPosition: self   endof
                SB_THUMBTRACK    of r@ VPosition: self   endof
        ENDCASE r>drop
        \ position the vertical button in the scroll bar
        TRUE cur-first-line SB_VERT
        GetHandle: self Call SetScrollPos drop

   0 
;M

))

;object

:object ComWindow <super window
	     
Font fdFont \ font data structure

int screen-cols
int screen-rows
int SlowCounter         \ increments when the window is updated

int cur-first-line \ current first line position
0 value first-line#    \ first line number
200 value last-line#   \ last line number
last-line# 20 - value last-top-line#
16 value bytes/line
  
: Wtype() ( a n )
\   c# @ charX *  l# @   charY * 
   c# @ l# @ c>p
   2swap  TextOut: dc 
;

:m home: ( -- )
   first-line# to cur-first-line 
   paint: self 
;m

:M On_Init:     ( -- )          \ things to do at the start of window creation

   ['] Wtype() is Wtype \ set up the type function

   On_Init: super             \ do anything superclass needs

   charX  Width: fdFont
   charY Height: fdFont
\   s"  Arial" SetFaceName: fdFont
\   s"       Courier" SetFaceName: fdFont
   Create: fdFont
   s" Sinar F4 Error Message"  SetTitle: self
   on_paint: self
;M

:m on_move: ( w -- )
   On_Paint: self
;m

:m startpos: ( StartPos: parent 20 20 v+ ) 22 100 ;m

:m startsize: 68 charX *  20 charY * ;m

:m on_paint: ( -- )
   SaveDC: dc                      \ save device context

   Handle: fdFont SetFont: dc      \ set the font to be used
\   s"  Arial" SetFaceName: fdFont
   RED SetTextColor: dc

   HelpCom
 
   RestoreDC: dc	
;m


:M WM_LBUTTONDOWN     ( -- res )
   On_Done: self
   DestroyWindow: self
   0
;M

:M WM_RBUTTONDOWN     ( -- res )
   On_Done: self
   DestroyWindow: self
   0
;M
   

:M On_Done:    ( h m w l -- res )
    Delete: fdFont			\ remove the font
\   On_Done: Super	\ the class's close function
\   DestroyWindow: self
   0 
;M

:M WindowStyle: ( -- style )            \ return the window style
                WindowStyle: super
WS_BORDER  or
                ;M

:M ExWindowStyle: ( -- extended_style )
                ExWindowStyle: super
                WS_EX_TOPMOST or        \ if so, lock on top
;M

;object



\ *********************** Dialog and Monologue boxes **************************
\ GetHandle: mainWindow to  GetHandleOfMainWindow

0 value GetHandleOfMainWindow

variable my_temp$ 128 allot

: Monologue ( $message $title )     \ presents some information

   temp$ place
   temp$ +NULL

   my_temp$ place
   my_temp$ +NULL

   MB_TASKMODAL
\   MB_ICONWARNING or   \ add a big "!"
   MB_ICONINFORMATION or   \ add an "i"

   my_temp$ 1+ rel>abs
   temp$ 1+ rel>abs

   GetHandleOfMainWindow

   Call MessageBox drop
;

\ presents some information and offers a choice
: Dialogue  ( $message $title - result ) 
   temp$ place
   temp$ +NULL

   my_temp$ place
   my_temp$ +NULL

   MB_TASKMODAL
   MB_OKCANCEL or  \ Ok or cancel?
   MB_ICONQUESTION or  \ add a big "?"

   my_temp$ 1+ rel>abs

   temp$ 1+ rel>abs

   GetHandleOfMainWindow
   
   Call MessageBox         \ returns 1 for OK, 2 for Cancel
;

: PasswordEntry  ( $message $title - result ) 
   temp$ place
   temp$ +NULL

   my_temp$ place
   my_temp$ +NULL

   MB_TASKMODAL
   MB_OKCANCEL or  \ Ok or cancel?
   MB_ICONQUESTION or  \ add a big "?"

   my_temp$ 1+ rel>abs

   temp$ 1+ rel>abs

   GetHandleOfMainWindow
   
   Call MessageBox         \ returns 1 for OK, 2 for Cancel
;

: TTD  s" Title bar" s" Message" Dialogue . ;


:object ForcePaintWindow <super window

:M On_Init:     ( -- )          \ things to do at the start of window creation
\   On_Init: super             \ do anything superclass needs
\   on_paint: super
   on_paint: self
;M

\ :m startpos: ( StartPos: parent 20 20 v+ ) 2 80 ;m

\ :m startsize: 1 1 ;m

:m on_paint: ( -- )
   DestroyWindow: self
;m

;object

\ ***** program end *****

