\ LanguageDialog.f    Language selection Dialog        2021 Jan 11

\ The data returned by reading curves 1004 to 1007 is
\ 'CONFIG  2 bytes  Bit 7 = GrainPro/-BeanPro, bits 0 to 3 enable language options 0 to 3
\ LANG     2 bytes  0, 1, 2 or 3
\ MyLanguage string : byte 0, bits 0 to 5 available space on LCD for this text, followed by
\ four counted strings for the four available languages, in WinStar LCD format, mixed UNICODE 16 (sort of) and ASCII 

include PasswordBox.f

$30 8 * Buffer: LanguageConfigBuffer[]  
variable v_LanguageConfigBitsSelected
\ variable v_LanguageConfigBitsReceived

variable v_LanguageConfigBitsSelectedLast
variable v_LanguageConfigBitsReceivedLast

$30 Buffer: LanguageName0[]  
$30 Buffer: LanguageName1[]  
$30 Buffer: LanguageName2[]  
$30 Buffer: LanguageName3[]  

: RemoveUnicode16chars ( $ -- $ )
   2dup  
   over + swap ?do  i c@ $80 and if  $2E2E i w!  then  2 +loop
;

\ : ttRUNI ( -- )    s" 12345678012345678" RemoveUnicode16chars  type ;

: ParseLanguageInfo ( -- )
\   LanguageConfigBuffer[] w@ v_LanguageConfigBitsReceived !
   LanguageConfigBuffer[] @ v_LanguageConfigBitsSelected !
   LanguageConfigBuffer[] 4 +  1+ >r 
   r@                         count RemoveUnicode16chars  LanguageName0[] place
   r@ count +                 count RemoveUnicode16chars  LanguageName1[] place
   r@ count + count +         count RemoveUnicode16chars  LanguageName2[] place
   r@ count + count + count + count RemoveUnicode16chars  LanguageName3[] place
   r> drop
;

: GetLangaugeInfo ( -- f )   \ from the four special "curves", return true if OK
   0 
   LanguageConfigBuffer[] $00 + 1004 RemoteCurveIn if  1+  then  1 ms
   LanguageConfigBuffer[] $30 + 1005 RemoteCurveIn if  1+  then  1 ms
   LanguageConfigBuffer[] $60 + 1006 RemoteCurveIn if  1+  then  1 ms
   LanguageConfigBuffer[] $90 + 1007 RemoteCurveIn if  1+  then  1 ms
   4 = not if  
      cr ." GetLangaugeInfo failed!!!"  12349 throw   
   then

   \ cr LanguageConfigBuffer[] $90 dump  
   ParseLanguageInfo
   -1
;

$30 Buffer: LanguageOptionCurve1008[]

decimal

variable vLanguageOptionsSet

: SetOptionBit ( f mask -- )
   dup invert v_LanguageConfigBitsSelected @ and v_LanguageConfigBitsSelected !
   swap if  v_LanguageConfigBitsSelected @ or  v_LanguageConfigBitsSelected !  else  drop  then
;

: SetMainLanguageOptionBits ( c -- )
   3 and 4 lshift 
   v_LanguageConfigBitsSelected @ $03 4 lshift invert and v_LanguageConfigBitsSelected !
   v_LanguageConfigBitsSelected @ or  v_LanguageConfigBitsSelected !  
;

: SetHLWOptionBits ( c -- )
   3 and 6 lshift 
   v_LanguageConfigBitsSelected @ $03 6 lshift invert and v_LanguageConfigBitsSelected !
   v_LanguageConfigBitsSelected @ or  v_LanguageConfigBitsSelected !  
;

: (4.hex) ( w -- $ )
   base @ >r hex
   0 <# # # # # #> 
   r> base !
;

: ttsob ( -- )
   SetOptionBit cr v_LanguageConfigBitsSelected @ H.2
;

: SetLanguageOptions ( -- f )
   0 vLanguageOptionsSet !
   LanguageOptionCurve1008[] $30 erase
   v_LanguageConfigBitsSelected @ LanguageOptionCurve1008[] w!
   LanguageOptionCurve1008[] 1008 RemoteCurveOut if  -1 vLanguageOptionsSet !  then
   vLanguageOptionsSet @
;

: v+ ( x y x' y' -- x+x' y+y' )
   >r  rot + swap r> +   
;

: v- ( x y x' y' -- x-x' y-y' )
   >r  rot swap -  swap r> -   
;

    200 constant OneField_x
     20 constant OneField_y
OneField_x  OneField_y 2constant OneField
     7 constant Gap_x
     2 constant Gap_y
 Gap_x  Gap_y 2constant Gap

  Gap_x 2* OneField_x + constant Size_x
  Gap_y 2* OneField_y + constant Size_y

 50  30 2constant Group_org
\ Size_x Gap_x + 27 - Size_y 4 1+ * 2constant Col_x
 Size_x Gap_x 2* + constant Col_x

: >Group_field ( fx fy -- x y dx dy )   
   >r 
   Col_x  Gap_x 8 * +  * 0  Group_org v+ \ 0 20 v+ 
   Size_y r> 1+ * + OneField ;

: >Group_org ( fx fy #lines -- x y )   
   >r
   >Group_field 2drop 
   Gap Size_y + v-  \ 10 20 v- 
   Col_x Gap_x 3 * - 7 +
   r> 1+ Size_y *  Gap_y 3 * + \ Size_y 2/ +
;

variable v_DataSaved

: clrSaved ( -- )   0 v_DataSaved ! ;

:OBJECT LanguageDialog <SUPER WINDOW 

   PasswordBox MyPasswordBox

   TextBox MyTextBox
                                 \ Bits set in 'CONFIG variable :
   GroupControl Group_0           \ a frame around a group   Bits 0 to 3
   CheckControl Language_0        \ check box to select or de-select language 0
   CheckControl Language_1        \ check box to select or de-select language 1
   CheckControl Language_2        \ check box to select or de-select language 2
   CheckControl Language_3        \ check box to select or de-select language 3

   GroupControl Group_0b          \ a frame around a group   Bits 4 and 5
   RadioControl Main_Language_0   \ a radio button
   RadioControl Main_Language_1   \ a radio button
   RadioControl Main_Language_2   \ a radio button
   RadioControl Main_Language_3   \ a radio button

   GroupControl Group_2           \ a frame around a group   Bits 6 and 7
   RadioControl HLWunits_0        \ a radio button
   RadioControl HLWunits_1        \ another radio button
   RadioControl HLWunits_2        \ another radio button
   RadioControl HLWunits_3        \ another radio button

   GroupControl Group_1           \ a frame around a group   Bit 8
   RadioControl Select_BeanPro    \ a radio button
   RadioControl Select_GrainPro     \ another radio button

   GroupControl Group_3             \ a frame around a group  
   CheckControl Option_USA          \ check box USA            Bit 9
   CheckControl Option_ShowLang     \ check box ShowLang       Bit 10 

   GroupControl Group_PW            \ a frame around a group   for the password textbox

   ButtonControl Button_1              \ a button
   ButtonControl Button_2              \ a button
   ButtonControl Button_3              \ a button

:M Refresh: ( -- )

   \ Bits 0 to 3 Enable Languages, Bits 4 and 5  MainLanguage, Bits 6 and 7 HLW units, 
   \ Bit 8 GrainPro/-BeanPro, Bit 9 USA option degF/-degC, Bit 10 ShowLang option
   v_LanguageConfigBitsSelected @ 1 and if  1  else  0  then 
   GetID: Language_0 CheckDlgButton: self 
   v_LanguageConfigBitsSelected @ 2 and if  1  else  0  then 
   GetID: Language_1 CheckDlgButton: self 
   v_LanguageConfigBitsSelected @ 4 and if  1  else  0  then 
   GetID: Language_2 CheckDlgButton: self 
   v_LanguageConfigBitsSelected @ 8 and if  1  else  0  then 
   GetID: Language_3 CheckDlgButton: self 

   v_LanguageConfigBitsSelected @ 4 rshift 3 and case \ bits 4 and 5
   0 of 1 0 0 0  endof
   1 of 0 1 0 0  endof
   2 of 0 0 1 0  endof
   3 of 0 0 0 1  endof
   endcase
   GetID: Main_Language_3 CheckDlgButton: self
   GetID: Main_Language_2 CheckDlgButton: self
   GetID: Main_Language_1 CheckDlgButton: self
   GetID: Main_Language_0 CheckDlgButton: self

   v_LanguageConfigBitsSelected @ 6 rshift 3 and case \ bits 6 and 7
   0 of 1 0 0 0  endof
   1 of 0 1 0 0  endof
   2 of 0 0 1 0  endof
   3 of 0 0 0 1  endof
   endcase
   GetID: HLWunits_3 CheckDlgButton: self
   GetID: HLWunits_2 CheckDlgButton: self
   GetID: HLWunits_1 CheckDlgButton: self
   GetID: HLWunits_0 CheckDlgButton: self

   v_LanguageConfigBitsSelected @ 8 rshift 1 and 0=
   GetID: Select_BeanPro CheckDlgButton: self 
   v_LanguageConfigBitsSelected @ 8 rshift 1 and 
   GetID: Select_GrainPro CheckDlgButton: self 

   v_LanguageConfigBitsSelected @ 9 rshift 1 and 
   GetID: Option_USA CheckDlgButton: self 

   v_LanguageConfigBitsSelected @ 10 rshift 1 and  
   GetID: Option_ShowLang CheckDlgButton: self 

   LanguageName0[] count SetText: Language_0
   LanguageName1[] count SetText: Language_1
   LanguageName2[] count SetText: Language_2
   LanguageName3[] count SetText: Language_3

   LanguageName0[] count SetText: Main_Language_0
   LanguageName1[] count SetText: Main_Language_1
   LanguageName2[] count SetText: Main_Language_2
   LanguageName3[] count SetText: Main_Language_3

   v_DataSaved @ if
      s" Saved."  SetText: Button_2 
   else
        s" Save to Unit"  SetText: Button_2 
   then

   v_LanguageConfigBitsSelectedLast @ v_LanguageConfigBitsSelected @ - if
      \ v_LanguageConfigBitsSelected @ (4.hex) 2drop 
      \ s" Hello" SetText: MyTextBox
      v_LanguageConfigBitsSelected @ v_LanguageConfigBitsSelectedLast !
   then

;M

:M ClassInit:   ( -- )
                ClassInit: super
                ;M

:M ExWindowStyle: ( -- style )
                ExWindowStyle: SUPER
                ;M

:M WindowStyle: ( -- style )
                WindowStyle: SUPER
                WS_BORDER OR
                WS_OVERLAPPED OR
                ;M

:M WindowTitle: ( -- title )
                z" Select Unit Options" ;M

:M StartSize:   ( -- width height )
                640 480 ;M

:M StartPos:    ( -- x y )
                30 30 ;M

:M Close:       ( -- )
\                GetText: Edit_1  cr type cr
   Close: SUPER
   ;M

:M On_Init:     ( -- )

   0 v_DataSaved !

   self               Start: MyPasswordBox
   1 12 >Group_field  Move: MyPasswordBox
      WS_GROUP        +Style: MyPasswordBox
      SetFocus: MyPasswordBox

   self               Start: Group_PW
   1 12 1 >Group_org   Move: Group_PW                                   
   s" Password :                               " SetText: Group_PW
   self                Start: Language_0
   0 0 >Group_field  Move: Language_0
   LanguageName0[] count SetText: Language_0

((
   self               Start: MyTextBox
   1 14 >Group_field  2>r  40 10 v+ 2r> Move: MyTextBox
   BS_DEFPUSHBUTTON   +Style: MyTextBox
   s" 'CONFIG " SetText: MyTextBox
   IDOK3               SetID: MyTextBox
))
     
   self                Start: Language_1
   0 1 >Group_field  Move: Language_1
   LanguageName1[] count SetText: Language_1

   self                Start: Language_2
   0 2 >Group_field  Move: Language_2
   LanguageName2[] count SetText: Language_2

   self                Start: Language_3
   0 3 >Group_field  Move: Language_3
   LanguageName3[] count SetText: Language_3

   self                 Start: Group_0
   0 0 4 >Group_org       Move: Group_0                                   \
   s"    Enable Language Options    " SetText: Group_0

   self                Start: Main_Language_0
   1 0 >Group_field  Move: Main_Language_0
   LanguageName0[] count SetText: Main_Language_0

   self                Start: Main_Language_1
   1 1 >Group_field  Move: Main_Language_1
   LanguageName1[] count SetText: Main_Language_1

   self                Start: Main_Language_2
   1 2 >Group_field  Move: Main_Language_2
   LanguageName2[] count SetText: Main_Language_2

   self                Start: Main_Language_3
   1 3 >Group_field  Move: Main_Language_3
   LanguageName3[] count SetText: Main_Language_3

   self                 Start: Group_0b
   1 0 4 >Group_org       Move: Group_0b                                   \
\  s"    Enable Language Options    " SetText: Group_0
   s"       Main startup language      " SetText: Group_0b

   self                Start: Select_BeanPro
   0 6 >Group_field    Move: Select_BeanPro
   s" BeanPro"         SetText: Select_BeanPro
   WS_GROUP            +Style: Select_BeanPro 

   self                Start: Select_GrainPro
   0 7 >Group_field    Move: Select_GrainPro
   BS_LEFT             +Style: Select_GrainPro 
   s" GrainPro"        SetText: Select_GrainPro

   self                 Start: Group_1
\   50 175 180 100         Move: Group_1
   0 6 2 >Group_org    Move: Group_1
   s"           Select Unit Type            " SetText: Group_1


   self                Start: HLWunits_0
   1 6 >Group_field   Move: HLWunits_0
   s" g/l"            SetText: HLWunits_0    \ g/l kg/hl  lb/bu  lbcf
   WS_GROUP            +Style: HLWunits_0 

   self                Start: HLWunits_1
   1 7 >Group_field  Move: HLWunits_1
   BS_LEFT             +Style: HLWunits_1 
   s" kg/hl"           SetText: HLWunits_1

   self                Start: HLWunits_2
   1 8 >Group_field  Move: HLWunits_2
   s" lb/bu"            SetText: HLWunits_2    \ g/l kg/hl  lb/bu  lbcf
   BS_LEFT            +Style: HLWunits_2 

   self                Start: HLWunits_3
   1 9 >Group_field  Move: HLWunits_3
   BS_LEFT             +Style: HLWunits_3 
   s" lbcf"           SetText: HLWunits_3

   self                 Start: Group_2
   1 6 4 >Group_org      Move: Group_2
   s"           Select HLW Units          " SetText: Group_2


   self                Start: Option_USA
   0 10 >Group_field  Move: Option_USA
   s" Enable degF" SetText: Option_USA

   self                Start: Option_ShowLang
   0 11 >Group_field  Move: Option_ShowLang
   s" Show 'Lang' text" SetText: Option_ShowLang

   self                 Start: Group_3
   0 10 2 >Group_org       Move: Group_3                                   \
   s"         Select Defaults                " SetText: Group_3

\   IDCANCEL                SetID: Button_1
\   self                Start: Button_1
\   230 285 70 25        Move: Button_1
\   s" Close"    SetText: Button_1
\                  GetStyle: Button_1
\   BS_DEFPUSHBUTTON   +Style: Button_1

   IDOK                SetID: Button_2
   self                Start: Button_2
   45 400 120 40        Move: Button_2
   s" Save to Unit"  SetText: Button_2
                  GetStyle: Button_2
   BS_DEFPUSHBUTTON   +Style: Button_2


   IDOK2               SetID: Button_3
   self                Start: Button_3
   240 400 120 40       Move: Button_3
   s" Refresh"       SetText: Button_3
                  GetStyle: Button_3
   BS_DEFPUSHBUTTON   +Style: Button_3


   Refresh: self
                
;M

:M On_Paint:    ( -- )          \ screen redraw procedure

                0 0 StartSize: self LTGRAY FillArea: dc
                ;M

:M WM_COMMAND   ( hwnd msg wparam lparam -- res )
        over LOWORD ( ID )
        case
              \  IDOK3 of 
              \     v_LanguageConfigBitsSelected @ (4.hex) SetText: MyTextBox
              \  endof

                IDCANCEL           of Close: self          endof
                IDOK               of     \ "Save to unit"
                  \ $0F SetLanguageOptions if
                        \ s" " SetText: MyPasswordBox
                        GetText: MyPasswordBox ReceivedPassword[] place
                        PasswordOK? if
                           s" Password OK                           " SetText: Group_PW
                           SetLanguageOptions 0= if 

                           then
                           bell -1 v_DataSaved ! 
                        else
                           s" Please enter the password :      " SetText: Group_PW
                           bell bell
                           SetFocus: MyPasswordBox
                        then                   
                  \ then
                endof
                IDOK2              of     \ "Refresh"
                  GetLangaugeInfo if  bell  2 v_DataSaved !  then  
                endof
         GetID: Language_0  of  clrSaved  GetID: Language_0 IsDlgButtonChecked: self $01 SetOptionBit  endof
         GetID: Language_1  of  clrSaved  GetID: Language_1 IsDlgButtonChecked: self $02 SetOptionBit  endof
         GetID: Language_2  of  clrSaved  GetID: Language_2 IsDlgButtonChecked: self $04 SetOptionBit  endof
         GetID: Language_3  of  clrSaved  GetID: Language_3 IsDlgButtonChecked: self $08 SetOptionBit  endof

         GetID: Main_Language_0  of  clrSaved   $00 SetMainLanguageOptionBits  endof
         GetID: Main_Language_1  of  clrSaved   $01 SetMainLanguageOptionBits  endof
         GetID: Main_Language_2  of  clrSaved   $02 SetMainLanguageOptionBits  endof
         GetID: Main_Language_3  of  clrSaved   $03 SetMainLanguageOptionBits  endof

         GetID: HLWunits_0 of  clrSaved   $00 SetHLWOptionBits  endof
         GetID: HLWunits_1 of  clrSaved   $01 SetHLWOptionBits  endof
         GetID: HLWunits_2 of  clrSaved   $02 SetHLWOptionBits  endof
         GetID: HLWunits_3 of  clrSaved   $03 SetHLWOptionBits  endof

         GetID: Select_BeanPro  of  clrSaved  GetID: Select_BeanPro  IsDlgButtonChecked: self 0= $100 SetOptionBit  endof
         GetID: Select_GrainPro of  clrSaved  GetID: Select_GrainPro IsDlgButtonChecked: self    $100 SetOptionBit  endof
 
         GetID: Option_USA      of  clrSaved  GetID: Option_USA      IsDlgButtonChecked: self    $200 SetOptionBit  endof
         GetID: Option_ShowLang of  clrSaved  GetID: Option_ShowLang IsDlgButtonChecked: self    $400 SetOptionBit  endof
          
     \    GetID: MyTextBox of  clrSaved  GetID: MyTextBox IsDlgButtonChecked: self   xxxxx  endof
                                         
        endcase

        v_DataSaved @ 2 = if 
            0 v_DataSaved !
        else
            Refresh: self
        then
        0 ;M

;OBJECT


