\ PasswordBox.f  2021 Jan 11

decimal
1109 constant EM_CANREDO  
1108 constant EM_REDO     
EM_LIMITTEXT constant EM_SETLIMITTEXT \ 197

$100 Buffer: ReceivedPassword[]
2variable lastMessage
variable PasswordEntered

\ ------------------------------------------------------------------------
\ *W <a name="TextBox"></a>
\ *S TextBox class
\ ------------------------------------------------------------------------
:Class TextBox		<Super EditControl
\ *G Class for Edit controls.
\ ** An edit control is a rectangular control window typically used in a dialog
\ ** box to permit the user to enter and edit text by typing on the keyboard.
\ *P This is an enhanced Version of the EditControl class.

int pWmSetFocus	\ pointer to WM_SETFOCUS handler

:M ClassInit:   ( -- )
\ *G Initialise the class.
                ClassInit: super
                0 to pWmSetFocus ;M

:M SetWmSetFocus: ( xt -- )
\ *G Set handler for WM_SETFOCUS messages.
                to pWmSetFocus ;M

:M WM_SETFOCUS  ( h m w l -- res )
                pWmSetFocus
                if   self pWmSetFocus execute
                then old-wndproc CallWindowProc  \ still needs to be called
                ;M

:M WindowStyle: ( -- style )
\ *G Get the window style of the control.
                WindowStyle: super ;M

:M ReadOnly:    ( f -- )
\ *G Set or remove the read-only style of the edit control.
\ ** A value of TRUE sets the read-only style; a value of FALSE removes it.
                0 SWAP EM_SETREADONLY hWnd call SendMessage ?Win-error ;M

:M SetSelection: ( nEnd nStart -- )
\ *G Selects a range of characters in the edit control. \i nEnd \d specifies the
\ ** ending character position of the selection. \i nStart \d specifies the
\ ** starting character position of the selection.
\ *P The start value can be greater than the end value. The lower of the two values
\ ** specifies the character position of the first character in the selection. The
\ ** higher value specifies the position of the first character beyond the selection.
\ *P The start value is the anchor point of the selection, and the end value is the
\ ** active end. If the user uses the SHIFT key to adjust the size of the selection,
\ ** the active end can move but the anchor point remains the same.
\ *P The control displays a flashing caret at the end position regardless of the relative
\ ** values of start and end.
                EM_SETSEL hWnd call SendMessage drop ;M

:M GetSelection: ( -- nEnd nStart )
\ *G Get the starting and ending character positions of the current selection in the edit control
		NULL NULL EM_GETSEL hWnd call SendMessage
		DUP HIWORD SWAP LOWORD ;M

:M SelectAll: 	( -- )
\ *G Set the focus to the edit control and select all the text in the control.
                SetFocus: Self   \ Change the focus for deselecting.
                -1 0 SetSelection: self
                ;M

:M RemoveSelection: ( -- )
\ *G Remove any selection.
		0 -1 SetSelection: self ;M

:M GetCursor:   ( -- n )
\ *G Get location of cursor (chars from start)
                GetSelection: self
                RemoveSelection: self
                GetSelection: self  >R
                SetSelection: self  R>
                ;M

:M IsModified?: ( -- f )
\ *G Get the state of an edit control's modification flag. The flag indicates whether
\ ** the contents of the edit control have been modified.
                0 0 EM_GETMODIFY hWnd call SendMessage
                ;M

:M SetModify:	( f -- )
\ *G Sets or clears the modification flag for an edit control. The modification flag
\ ** indicates whether the text within the edit control has been modified.
                swap 0 EM_SETMODIFY hWnd call SendMessage drop ;M

\ DPR-WARNING-OFF
:M NotModified: ( -- )
\ *G Depreacted method. Use \i SetModify: \d instead.
                0 0 EM_SETMODIFY hWnd call SendMessage drop ;M \ DEPRECATED
\ DPR-WARNING-ON

:M Undo?:       ( -- f )
\ *G Check is there is an operation in the control's undo queue.
		0 0 EM_CANUNDO hWnd call SendMessage ;M

:M Undo:	( -- )
\ *G Undoes the last edit control operation in the control's undo queue.
                Undo?: self
		IF   0 0 EM_UNDO hWnd call SendMessage drop
		ELSE MB_OK
		     Z" Undo notification"
		     Z" Nothing to undo."
		     GetHandle: self Call MessageBox DROP
		THEN ;M

:M Redo?:       ( -- f )
\ *G Check is there is an operation in the control's redo queue.
		0 0 EM_CANREDO hWnd call SendMessage ;M

:M Redo:	( -- )
\ *G Redoes the last edit control operation in the control's redo queue.
		Redo?: self
		IF   0 0 EM_REDO hWnd call SendMessage drop
		ELSE MB_OK
		     Z" Redo notification"
		     Z" Nothing to redo."
		     GetHandle: self Call MessageBox DROP
		THEN ;M

:M Cut: 	( -- )
\ *G Delete (cut) the current selection, if any, in the edit control and
\ ** copy the deleted text to the clipboard in CF_TEXT format.
		0 0 WM_CUT hWnd call SendMessage drop ;M

:M Copy: 	( -- )
\ *G Copy the current selection to the clipboard in CF_TEXT format.
		0 0 WM_COPY hWnd call SendMessage drop ;M

:M Paste: 	( -- )
\ *G Copy the current content of the clipboard to the edit control at the current
\ ** caret position. Data is inserted only if the clipboard contains data in CF_TEXT
\ ** format.
		0 0 WM_PASTE hWnd call SendMessage drop ;M

:M Clear: 	( -- )
\ *G Delete selected text.
		0 0 WM_CLEAR hWnd call SendMessage drop ;M

:M SetFont:     ( handle -- )
\ *G Set the font in the control.
                1 swap WM_SETFONT hWnd call SendMessage drop ;M

:M SetTextLimit: ( n -- )
\ *G Set the text limit of an edit control.
\ *P \i n \d Specifies the maximum number of characters the user can enter. This number does
\ ** not include the null terminator. \n
\ ** Edit controls on Windows NT/ 2000: If this parameter is zero, the text length is set
\ ** to 0x7FFFFFFE characters for single-line edit controls or -1 for multiline edit controls. \n
\ ** Edit controls on Windows 95/98: If this parameter is zero, the text length is set to 0x7FFE
\ ** characters for single-line edit controls or 0xFFFF for multiline edit controls.
\ *P The SetTextLimit: method limits only the text the user can enter. It does not affect any text
\ ** already in the edit control when the message is sent, nor does it affect the length of the text
\ ** copied to the edit control by the SetText: method. If an application uses the SetText: method
\ ** to place more text into an edit control than is specified in the SetTextLimit: method, the user can
\ ** edit the entire contents of the edit control.
\ *P Before the SetTextLimit: method is called, the default limit for the amount of text a user can enter
\ ** in an edit control is 32,767 characters.
\ *P Edit controls on Windows NT/ 2000: For single-line edit controls, the text limit is either 0x7FFFFFFE bytes
\ ** or the value of \i n \d, whichever is smaller. For multiline edit controls, this value is either
\ ** -1 bytes or the value of \i n \d, whichever is smaller.
\ *P Edit controls on Windows 95/98: For single-line edit controls, the text limit is either 0x7FFE bytes or
\ ** the value of \i n \d, whichever is smaller. For multiline edit controls, this value is either 0xFFFF bytes
\ ** or the value of \i n \d, whichever is smaller.
                0 swap EM_SETLIMITTEXT hWnd call SendMessage drop ;M

:M GetTextEx:  \ { buffer$ maxlen -- buffer$ len }
\ *G Copies the text of the edit control into a buffer.
\ *P \i buffer$ \d is the address of the buffer that will receive the text.
\ *P \i maxlen \d Specifies the maximum number of characters to copy to the
\ ** buffer, including the NULL character. If the text exceeds this limit, it
\ ** is truncated.
\ *P If the method succeeds, \i len \d is the length, in characters, of the copied
\ ** string, not including the terminating null character.
\                maxlen buffer$ hwnd Call GetWindowText buffer$ swap ;M
                over hwnd Call GetWindowText ;M

:M SetTextZ:    ( addrz -- )
\ *G Copy the text from the 0 terminated string \i addrz \d into the edit control.
                hwnd
                if   hwnd Call SetWindowText ?win-error
		else drop
                then ;M

;Class
\ *G End of TextBox class

\ ------------------------------------------------------------------------
\ *W <a name="PassWordBox"></a>
\ *S PassWordBox class
\ ------------------------------------------------------------------------
:Class PassWordBox	<super TextBox
\ *G Class for Edit controls.
\ ** All characters in the control are displayed as an asterisk (*).

:M WindowStyle: ( -- style )
\ *G Get the window style of the control. The default style is ES_PASSWORD.
	WindowStyle: super ES_PASSWORD or ;M

;Class
\ *G End of PassWordBox class

:OBJECT PasswordDialog <SUPER WINDOW 

   PasswordBox MyPasswordBox
   ButtonControl Button_1          \ a button
\   ButtonControl Button_2          \ a button
\   ButtonControl Button_3          \ a button

:M Refresh: ( -- )   
;M

:M ClassInit:   ( -- )
                ClassInit: super
                ;M

:M ExWindowStyle: ( -- style )
                ExWindowStyle: SUPER
                ;M

:M WindowStyle: ( -- style )
                WindowStyle: SUPER
                WS_BORDER OR
                WS_OVERLAPPED OR
                WS_EX_TOPMOST or
                ;M

:M WindowTitle: ( -- title )
                z" Please enter the password :" ;M

:M StartSize:   ( -- width height )
                640 480 ;M

:M StartPos:    ( -- x y )
                30 30 ;M

:M Close:       ( -- )
\                GetText: Edit_1  cr type cr
   \ Close: SUPER
   ;M

:M On_Init:     ( -- )

   0 PasswordEntered !

   self          Start: MyPasswordBox
   20 20 160 30  Move: MyPasswordBox
   s" " SetText: MyPasswordBox
   SetFocus: MyPasswordBox

\   IDCANCEL                SetID: Button_1
\   self                Start: Button_1
\   230 285 70 25        Move: Button_1
\   s" Close"    SetText: Button_1
\                  GetStyle: Button_1
\   BS_DEFPUSHBUTTON   +Style: Button_1

   IDOK2                SetID: Button_1
   self                Start: Button_1
   60 80 80 30          Move: Button_1
   s" OK"  SetText: Button_1
                    GetStyle: Button_1
   BS_DEFPUSHBUTTON   +Style: Button_1

   Refresh: self
                
;M

:M On_Paint:    ( -- )          \ screen redraw procedure

                0 0 StartSize: self LTGRAY FillArea: dc
                ;M

:M WM_COMMAND   ( hwnd msg wparam lparam -- res )
        over LOWORD ( ID )
        case
                \ IDCANCEL of  Close: self  endof
                IDOK2     of  
                  GetText: MyPasswordBox ReceivedPassword[] place
                  -1 PasswordEntered !
                  Close: SUPER   
                  bell  
                endof
        \ $72100C of  bell bell  endof
\        HIWORD LOWORD lastMessage 2!                       
        endcase
        Refresh: self
        0 ;M

;OBJECT

: ttpw  start: PasswordDialog ;

: PasswordOK? ( -- f )

   0 PasswordEntered !
 
 ((  
   10000 0 do 
      PasswordEntered @ if  leave  then
      100 sleep
   loop
  )) 
   ReceivedPassword[] count -Trailing  Number? if
      drop
      9876 =   \ just a simple fixed password
   else
      2drop
      0
   then
;
