\ *****************************************************************************
\ serial.f  2022 Jul 31  select, configure and access serial ports 
\ From Mnet8\serialPlus.f 2016 Sep 14  Howerd Oakford  www.inventio.co.uk
\ *****************************************************************************

decimal 

\ *****************************************************************************
\ Library imports
\ *****************************************************************************

\ 2 IMPORT: GetCommState
\ 2 IMPORT: SetCommTimeouts
\ 2 IMPORT: SetCommState
\ 2 IMPORT: GetCommProperties
\ 0 IMPORT: PurgeComm
\ 2 IMPORT: BuildCommDCB
\ FormatMessage

include Qvariables.f

Qvariable qvCOMh        \ handle 

\ *****************************************************************************
\ Windows error display using FormatMessage
\ *****************************************************************************

$200 constant |ErrorString| 
|ErrorString| Buffer: ErrorString[]

: .LastErrorText ( n -- )   \ displays the error message text for a Win32 error code
   dup 
   FORMAT_MESSAGE_IGNORE_INSERTS  FORMAT_MESSAGE_FROM_SYSTEM or  FORMAT_MESSAGE_ARGUMENT_ARRAY or
   NULL 
   rot ( format 0 n -- )
   NULL
   |ErrorString|
   ErrorString[]
   0 
   Call FormatMessage drop     \ check order of arguments
   ?dup if
      ErrorString[] swap 2 - 0 max  type  space 
   then
;

\ Display  GetLastError  values
: .LastError ( n -- )  
   dup . 2 spaces 
   dup case
   0    of  ." Not an error at all ;)"   endof
   1    of  ." ERROR_INVALID_FUNCTION"   endof
   2    of  ." ERROR_FILE_NOT_FOUND"     endof
   5    of  ." ERROR_ACCESS_DENIED"      endof
   6    of  ." ERROR_INVALID_HANDLE"     endof
   13   of  ." ERROR_INVALID_DATA"       endof
   21   of  ." ERROR_NOT_READY"          endof
   31   of  ." ERROR_GEN_FAILURE"        endof
   50   of  ." ERROR_NOT_SUPPORTED"      endof
   87   of  ." ERROR_INVALID_PARAMETER"  endof
   122  of  ." ERROR_INSUFFICIENT_BUFFER" endof
   123  of  ." ERROR_INVALID_NAME"       endof
   259  of  ." ERROR_NO_MORE_ITEMS"      endof
   997  of  ." ERROR_IO_PENDING"         endof
   1110 of  ." ERROR_MEDIA_CHANGED"      endof
   1112 of  ." ERROR_NO_MEDIA_IN_DRIVE"  endof
   1784 of  ." ERROR_INVALID_USER_BUFFER" endof
         ." Unknown Win32 error code = " 
   endcase
   space 
   .LastErrorText
;

\ show the error and abort
: ShowLastError ( a n -- )   
   ."  GetLastError = "  Call GetLastError .LastError  2 spaces 
   1112 throw  
;

: ?ERROR? ( f -- )   0= if  Call GetLastError .LastError  then ;


\ 0  value ComHandle \ the currently open COM port's handle
\ 1  value ComPort \ COM1 to COMn is 1 to n
variable RKEY-val     \ a place to save last key received
0  value RKEY-flg?    \ have we already got a key?

\ define the object for the Device Control Block structure

: bell ( -- )   1320 150 BEEP! BEEP ;

\ Abort the transfer if a C , c or Esc key is pressed
: LeaveEarlyKey? ( - )
   key? if  
      key >r
      r@ 'C' = 
      r@ 'c' = or
      r> 27 = or
      if
      bell
      -1 [LeaveEarly] !
      123 throw
      then
      0  [LeaveEarly] !
    then     
;      

\ : asd  0  begin  1+ dup 1000000 > if  123 throw then   Again ;
\ : asd  begin  KEY? if  key 85 = IF  bell  123 throw  then   THEN  Again ;
: asd  begin  LeaveEarlyKey?  Again ;
: zxc  ['] asd  catch . ;


:Object DCB        <Super Object

Record: AddrOf
        int   DCBlength
        int   BaudRate
        int   BinaryBits                \ a 32bit cell of bit fields
      1 bits  fBinary                   \ define the bit fields
      1 bits  fParity
      1 bits  fOutxCtsFlow
      1 bits  fOutxDsrFlow             \ SDI changed for Rts to Dsr
      2 bits  fDtrControl
      1 bits  fDtrSensitivity
      1 bits  fTXContinueOnXoff
      1 bits  fOutX
      1 bits  fInx
      1 bits  fErrorChar
      1 bits  fNull
      2 bits  fRtsControl
      1 bits  fAbortOnError
     17 bits  fDummy
        short wReserved
        short XonLim
        short XoffLim
        byte  ByteSize
        byte  Parity
        byte  StopBits
        byte  XonChar
        byte  XoffChar
        byte  ErrorChar
        byte  EofChar
        byte  EvtChar
        short wReserved1
;RecordSize: SIZEOFDCB

:M Reset:       ( -- )
                AddrOf SIZEOFDCB erase
                ;M

:M ClassInit:   ( -- )
                ClassInit: super
                Reset: self             \ create structure as Reset
                ;M

;Object

\ define the field names for the Communications Timeout structure

0       CELL Field+ .ReadIntervalTimeout
        CELL Field+ .ReadTotalTimeoutMultiplier
        CELL Field+ .ReadTotalTimeoutConstant
        CELL Field+ .WriteTotalTimeoutMultiplier
        CELL Field+ .WriteTotalTimeoutConstant
CONSTANT COMMTIMEOUTSBYTES


\ ***************************************************************************


\ COMM-TIMEOUTS  is an array of data to tell the open com port to
\ return immediately if no characters are waiting.
CREATE COMM-TIMEOUTS   -1 , 0 , 0 , 1 , 20 ,

: set_ComTimeouts ( -- )  
      COMM-TIMEOUTS rel>abs  qvCOMh @  Call SetCommTimeouts 0= if  16 throw  then
exit

\  { CT -- }      \ Initialize the communications timeouts
\    COMMTIMEOUTSBYTES LocalAlloc: CT    \ allocate a CT structure
\    CT COMMTIMEOUTSBYTES erase          \ initialize it to zeros
\    \ set read timeouts to magic value of don't wait, just poll
\    -1 CT .ReadIntervalTimeout         !
\    0 CT .ReadTotalTimeoutMultiplier   !
\    0 CT .ReadTotalTimeoutConstant     !
\    1 CT .WriteTotalTimeoutMultiplier  !
\    1000000 CT .WriteTotalTimeoutConstant    !
\    CT rel>abs
\    qvCOMh @
\    Call SetCommTimeouts drop 
;

\ ***************************************************************************
\ { ---------------------------------------------------------------------------
\ Data structures
\ 
\ CommConfig  is the data structure for the Windows interface
\ 
\ Field: defines the fields within the CommConfig structure
\ 
\ --------------------------------------------------------------------------- }

\ These Qvars contain data for the currently selected comm port 

\ Qvariable qvComNumber   \ COM port number

\ $40 Qbuffer: qbComFileName   \ null terminated name in "filename" form
Qvariable qvKeyFlag
Qvariable qvKeyChar
Qvariable qvEmitFlag
Qvariable qvEmitChar

\ *****************************************************************************
\ DCB
\ *****************************************************************************

: Field: ( u1 u2 -- u3)   create over , +  does> ( a -- a' )  @ + ;

' noop is stack-check
0
\ CommConfig starts here
   4 Field: ->dwSize
   2 Field: ->wVersion
   2 Field: ->wReserved
\ DCB starts here
   4 Field: ->DCB          \ Length of DCB
   4 Field: ->BaudRate     \ Line speed, bits per second
   4 Field: ->DCBflags     \ See flags below
   2 +                     \ Not currently used
   2 Field: ->XonLim       \ Transmit XON threshold
   2 Field: ->XoffLim      \ Transmit XOFF threshold
   1 Field: ->ByteSize     \ Bits per character (4-8)
   1 Field: ->Parity       \ 0-4 --> no,odd,even,mark,space
   1 Field: ->StopBits     \ 0,1,2 --> 1,1.5,2
   1 Field: ->XonChar      \ Tx and Rx XON char
   1 Field: ->XoffChar     \ Tx and Rx XOFF char
   1 Field: ->ErrorChar    \ Error replacement char
   1 Field: ->EofChar      \ End of input char
   1 Field: ->EvtChar      \ Received event char
   2 Field: ->Reserved1    \ Reserved, do not use
  24 Field: ->Reserved2    \ Reserved, do not use
\   4 +                    \ for 64 bit systems
constant |CommConfig|
' _stack-check is stack-check

((
typedef struct _COMM_CONFIG {
  DWORD dwSize;
  WORD  wVersion;
  WORD  wReserved;
  DCB   dcb;   \ 36 bytes
  DWORD dwProviderSubType;
  DWORD dwProviderOffset;
  DWORD dwProviderSize;
  WCHAR wcProviderData[1];
] COMMCONFIG, *LPCOMMCONFIG;

 9 cells : 
typedef struct _DCB {
  DWORD DCBlength;
  DWORD BaudRate;
     DWORD fBinary  :1;
     DWORD fParity  :1;
     DWORD fOutxCtsFlow  :1;
     DWORD fOutxDsrFlow  :1;
     DWORD fDtrControl  :2;
     DWORD fDsrSensitivity  :1;
     DWORD fTXContinueOnXoff  :1;
     8
     DWORD fOutX  :1;
     DWORD fInX  :1;
     DWORD fErrorChar  :1;
     DWORD fNull  :1;
     DWORD fRtsControl  :2;
     DWORD fAbortOnError  :1;
     7
     DWORD fDummy2  :17;
  total 32
  WORD  wReserved;
  WORD  XonLim;
  WORD  XoffLim;
  BYTE  ByteSize;
  BYTE  Parity;
  BYTE  StopBits;
  char  XonChar;
  char  XoffChar;
  char  ErrorChar;
  char  EofChar;
  char  EvtChar;
  WORD  wReserved1;
] DCB, *LPDCB;
)) 

: /ALLOT ( n -- )   HERE SWAP DUP ALLOT ERASE ;

create CommConfig   ( dwSize ) |CommConfig| , ( wVersion ) 1 w, |CommConfig| 0 ->wReserved -  /ALLOT



: ComClose ( -- )    \ close com port if its open
   qvCOMh @
   if  
      qvCOMh @ Call CloseHandle  drop
      0 qvCOMh !     \ mark as closed
   then
;

\ qvCOMh @ of 0 means failed to open the port
: ComOpen ( z1 -- ) \ Open Com port for z" COM1", or z" COM2"
   qvCOMh @ if  ComClose  then       \ just in case...

   >R
   NULL                            \ no template
   NULL                            \ open file attributes
   OPEN_EXISTING                   \ creation distribution
   NULL                            \ no security attributes
   0                               \ exclusive access
   GENERIC_READ GENERIC_WRITE or   \ desired access modes
   R> rel>abs                      \ zstring filename
   Call CreateFile                 \ returns handle or -1
   dup -1 = if                        \ if -1 then error
      \ CR  ." Failed to open COM port!"  drop  exit
      \ Hcr  s" Failed to open COM port!" Htype  drop  exit
      \ Start: CommWindow  exit
      drop 
   else
   \ dup qvCOMh ! \ save the handle
      qvCOMh !
      set_ComTimeouts  
   then
;


\ realSerialClose closes the current Qselect serial port

: realSerialClose ( -- )   
   qvCOMh @ if  
\      -RTS  \ disable the RTS handshake output from the PC
\      -DTR  \ disable the DTR handshake output from the PC
      qvCOMh @ Call CloseHandle drop 
      100 ms 
   0 qvCOMh !
\   UpdateStatusBuffer
   then 
;


variable COMportName  10 allot

: MakeCOMportName ( n -- z$ )
   s" \\.\COM" COMportName place
   (.) COMportName +place
   0 COMportName count + c!
   COMportName 1+ 
;   

\ *****************************************************************************

\ 7 data bits, Even parity, 1 stop bit
: set7E1 ( -- )  CommConfig >r  $11 r@ ->DCBflags !  $07 r@ ->ByteSize !  2 r@ ->Parity C!  0 r@ ->StopBits C!  r> drop ;
\ 8 data bits, no parity, 1 stop bit
: set8N1 ( -- )  CommConfig >r  $11 r@ ->DCBflags !  $08 r@ ->ByteSize !  0 r@ ->Parity C!  0 r@ ->StopBits C!  r> drop ;
\ 8 data bits, even parity, 1 stop bit
: set8E1 ( -- )  CommConfig >r  $11 r@ ->DCBflags !  $08 r@ ->ByteSize !  2 r@ ->Parity C!  0 r@ ->StopBits C!  r> drop ;

\ { ---------------------------------------------------------------------
\ Port control
\ 
\ The EscapeCommFunction function directs a specified communications device
\ to perform an extended function.
\ 
\ BOOL EscapeCommFunction(
\     HANDLE hFile,	// handle to communications device
\     DWORD dwFunc 	// extended function to perform
\    );
\ 
\ dwFunc	Meaning
\ CLRDTR	Clears the DTR (data-terminal-ready) signal.
\ CLRRTS	Clears the RTS (request-to-send) signal.
\ SETDTR	Sends the DTR (data-terminal-ready) signal.
\ SETRTS	Sends the RTS (request-to-send) signal.
\ SETXOFF	Causes transmission to act as if an XOFF character has 
\   been received.
\ SETXON	Causes transmission to act as if an XON character has 
\   been received.
\ SETBREAK	Suspends character transmission and places the 
\   transmission line in a break state until the ClearCommBreak 
\   function is called (or EscapeCommFunction is called with the 
\   CLRBREAK extended function code). 
\   The SETBREAK extended function code is identical to the 
\   SetCommBreak function. 
\   Note that this extended function does not flush data that 
\   has not been transmitted.
\ CLRBREAK	Restores character transmission and places the
\   transmission line in a nonbreak state. 
\   The CLRBREAK extended function code is identical to the 
\   ClearCommBreak function.
\ 
\ 
\ If the function succeeds, the return value is nonzero.
\ 
\ +DTR asserts DTR, -DTR lowers it.  And so on...
\ 
\ --------------------------------------------------------------------- }

\ 2 IMPORT: EscapeCommFunction

: +RTS   ( -- )   qvCOMh @ 3 Call EscapeCommFunction ?ERROR? ;    \ SETRTS 
: -RTS   ( -- )   qvCOMh @ 4 Call EscapeCommFunction ?ERROR? ;    \ CLRRTS 
: +DTR   ( -- )   qvCOMh @ 5 Call EscapeCommFunction ?ERROR? ;    \ SETDTR 
: -DTR   ( -- )   qvCOMh @ 6 Call EscapeCommFunction ?ERROR? ;    \ CLRDTR 
: +BREAK ( -- )   qvCOMh @ 8 Call EscapeCommFunction ?ERROR? ;    \ SETBREAK
: -BREAK ( -- )   qvCOMh @ 9 Call EscapeCommFunction ?ERROR? ;    \ CLRBREAK


: xxxserialOpen ( -- )  \ from Mnet8 partially working
   realSerialClose
   ComPort MakeCOMportName  
   \ Qselect @ >RegPortname count ComNameTo# MakeComFileName
   \ dup zcount Qselect @ >RegPortname place
   >R
   NULL                            \ no template
   NULL                            \ open file attributes
   OPEN_EXISTING                   \ creation distribution
   NULL                            \ no security attributes
   0                               \ exclusive access
   GENERIC_READ GENERIC_WRITE or   \ desired access modes
   R> rel>abs                      \ zstring filename
   Call CreateFile                 \ returns handle or -1

   dup INVALID_HANDLE_VALUE = if 
\      RED. ." Cannot open port " Qselect @ >RegPortname count type  ."  !  "  RESTORE.
      1002 throw
   else
      qvCOMh !
      qvCOMh @ COMM-TIMEOUTS swap Call SetCommTimeouts 0= if  16 throw  then 
      qvCOMh @ CommConfig ->DCB swap Call GetCommState 0= if  ( ." GetCommState" ) 17 throw  then \ get the system port state
      \ Qselect @ >RegBaudrate @  
      ComBaud
\      ForceFastBaud @ if  drop  115200  then 
      CommConfig ->Baudrate ! \ set our baudrate in the DCB using data from the registry
\     Qselect @ >RegPortSettings @ case
\     8N1 of  set8N1  endof  \  set 8 data bits, no parity, 1 stop bit
\     8E1 of  set8E1  endof  \  set 8 data bits, even parity, 1 stop bit
\     7E1 of  set7E1  endof  \  set 7 data bits, Even parity, 1 stop bit
         set8N1
\     endcase
      qvCOMh @  CommConfig ->DCB  swap Call SetCommState 0= if  ( ." SetCommState" ) 18 throw  then 
\      -RTS  \  the RTS handshake output from the PC
\      -DTR  \  the DTR handshake output from the PC
      +RTS  \  the RTS handshake output from the PC
      +DTR  \  the DTR handshake output from the PC
   then
\   UpdateStatusBuffer
   200 ms 
;

\ : SetBaud ( baud -- )   Qselect @ >RegBaudrate !  SetRegValues ;
\ : SetPortSettings ( c -- )   Qselect @ >RegPortSettings !  SetRegValues ;

\ ************************************************************
\ here is a list of valid parameters for ComSetup.  Except for
\ the ByteSize parameter, these are all windows constants.

\ --- BaudRate    CBR_110       CBR_300       CBR_600
\                 CBR_1200      CBR_2400      CBR_4800
\                 CBR_9600      CBR_14400     CBR_56000
\                 CBR_19200     CBR_38400     CBR_57600
\                 CBR_115200    CBR_128000    CBR_256000
\
\ --- ByteSize    5, 6, 7, 8
\
\ --- Parity      NOPARITY      ODDPARITY     MARKPARITY
\                 EVENPARITY    SPACEPARITY
\
\ --- StopBits    ONESTOPBIT    TWOSTOPBITS   ONE5STOPBITS
\
\ ************************************************************

\ Setup the Communications state to the parameters specified

: ComSetup      { baud size parity stop -- }
  DCB.AddrOf rel>abs
                qvCOMh @
                Call GetCommState ?win-error

                baud       Put: DCB.BaudRate
                size       Put: DCB.ByteSize
                parity 0<> Put: DCB.fParity             \ parity enabled flag
                parity     Put: DCB.Parity
                stop       Put: DCB.StopBits
                1          Put: DCB.fBinary

                DCB.AddrOf rel>abs
                qvCOMh @
                Call SetCommState ?win-error ;


:Object COMSTAT        <Super Object

Record: AddrOf
        int   lpComStatBits             \ a 32bit cell of bit fields
      1 bits  fCtsHold                  \ define the bit fields Low to High
      1 bits  fDsrHold
      1 bits  fRlsdHold
      1 bits  fXoffHold
      1 bits  fXoffSent
      1 bits  fEof
      1 bits  fTxim
     25 bits  fReserved
        int   cbInQue
        int   cbOutQue
;RecordSize: SIZEOFCOMSTAT

:M Reset:       ( -- )
                AddrOf SIZEOFCOMSTAT erase
                ;M

:M ClassInit:   ( -- )
                ClassInit: super
                Reset: self             \ create structure as Reset
                ;M

;Object

: ComErrorClear { lpErrors -- f }      \ true = success
                COMSTAT.addrof  rel>abs
                lpErrors  rel>abs
                qvCOMh @ Call ClearCommError 0<>  ;

\ : pause ( -- )   1 ms ; \ WinPause
: pause ( -- )   1 Call Sleep drop ; \ WinPause

0 [if]

\ Original SinarF4 serial port words

: SerialKey?      ( -- c1 )       \ get a key from serial port, don't wait long
   LeaveEarlyKey?
   RKEY-flg? 0=
   if      RKEY-val 1 qvCOMh @ read-file       \ -- len flag
\ if  drop -1  then
       abort" Failed to read serial port"       \ -- len
       to RKEY-flg?
   then
   RKEY-flg? ;

: SerialKey() ( -- c)   RKEY-val c@                  \ return the key
                0 to RKEY-flg? ;                \ clear the save buffer

: RPURGE   BEGIN  SerialKey? WHILE SerialKey() DROP  REPEAT ;
: RPURGE.   CR  BEGIN  SerialKey? WHILE SerialKey() .  REPEAT ;


: serialKey0       ( -- c1 )       \ must return a key
   begin   SerialKey?  until       \ loop till we get one
   SerialKey() ;

: X?ESC   KEY? IF  27 = ABORT" Escape..."  THEN ;

6 constant RKEY_TIMEOUT \ 

\ returns a 0 if it times out
: SerialKey ( -- c )   
   RKEY_TIMEOUT COUNTER +  
   BEGIN
      DUP COUNTER - 0<
      SerialKey? IF  2DROP  SerialKey()  EXIT  THEN
      LeaveEarlyKey?
   UNTIL  
   DROP  0 
;

\ : serialEmit      { char -- }     \ write a character to the serial port
\                &LOCAL char 1 qvCOMh @ write-file
\                abort" Failed to write serial port" ;

\ : serialType ( a n )   over + swap do  i c@ serialEmit  loop ;	

variable rres

variable 1pad	
variable charpad	

: (serialType1) ( -- f )  	\ tries to send 1 character, returns 0 if OK
    charpad 1 qvCOMh @ write-file ;

: serialEmit ( c)  
   charpad c! (serialType1) drop  pause \ WinPause
\   ComBaud 300 = if  2 ms  then 
   LeaveEarlyKey?
;

\ : serialType ( a n )   OVER C@ serialEmit 	\ abort if can't do it
\    swap 1+ swap  1-  0 max  dup 0= if  2drop  exit  then	
\    over + swap do  i c@ serialType1  loop ;	

: serialType ( a n )
   dup 0= if  2drop  exit  then	
   qvCOMh @ write-file drop
;
\   over + swap do  i c@ serialEmit  loop ;	

\ : slowSerialType ( a n -- )
\    dup 0= if  2drop  exit  then	
\    over + swap do  2 ms  i c@ serialEmit  loop 
\ ;	

: RR   s" Hi there!!! 1" serialType ;
: RRR   s" Hi there!!! 2"  ;

[else]

\ From Mnet8 :

\ { ---------------------------------------------------------------------------
\ 
\ serialKey?  returns true if a character has been received from the currently selected serial port
\ 
\ serialKey  waits for and returns a character from the currently selected serial port
\ 
\ serialKeyT/O waits for 1000 ms or until a key is received. Returns -1 if timed out or a keyboard key is pressed
\ 
\ serialEmit  sends a character to the currently selected serial port
\ 
\ serialPurge  cleans out the receive input buffer for the currently selected serial port
\ 
\ --------------------------------------------------------------------------- }

: 2**   ( c -- u )   \ takes 2 to the power c
   1 swap lshift
;

\ *****************************************************************************
\ *****************************************************************************

$100 Buffer: MyDCB[]

: BuildDCB ( -- a )
   MyDCB[] $100 erase
   z" baud=1200 parity=N data=8 stop=1" MyDCB[] Call BuildCommDCB ?ERROR?
   MyDCB[] |CommConfig| dump
;

\ *****************************************************************************
\ Comm Port character access, via Qvariables
\ *****************************************************************************
11 2** constant |SerialBuffer|
|SerialBuffer| qBuffer: SerialBuffer[]

Qvariable qSerialBufferCharsIn
Qvariable qSerialBufferCharsOut

: SerialBufferInit ( -- )
   SerialBuffer[] |SerialBuffer| erase 
   0 qSerialBufferCharsIn !
   0 qSerialBufferCharsOut ! 
;

SerialBufferInit

: SerialBuffer# ( -- n )   qSerialBufferCharsIn @ qSerialBufferCharsOut @ -  |SerialBuffer| 1- and ;

: >SerialBuffer ( n -- a )   |SerialBuffer| 1- and  SerialBuffer[] + ;

: SerialBufferPut ( c -- )
   qSerialBufferCharsIn @ >SerialBuffer  c!
   1 qSerialBufferCharsIn +!
;

: SerialBufferGet ( -- c )
   qSerialBufferCharsOut @ >SerialBuffer c@
   1 qSerialBufferCharsOut +!
;

: GetSerialBuffer ( a n -- )   SerialBuffer[] qSerialBufferCharsIn @ ;

: SerialBufferKey ( -- c ) 
   begin  pause  SerialBuffer# until  SerialBufferGet   
;

8 2** constant |SerialReadPad| 
|SerialReadPad| QBuffer: SerialReadPad[]

Qvariable SerialReadNumber

: serialKey? ( -- flag )
   SerialBuffer# ?dup if  exit  then 
\   qvKeyFlag @ dup ?EXIT drop
\   qvCOMh @ SerialReadPad[] |SerialReadPad| SerialReadNumber 0 Call ReadFile 
   0  SerialReadNumber rel>abs  |SerialReadPad|  SerialReadPad[] rel>abs  qvCOMh @  Call ReadFile 

   0= if  
      ( [char] Q EMIT(G] #1070 throw ) 
   then 
   
   SerialReadPad[] SerialReadNumber @ over + swap ?do  i c@ SerialBufferPut  loop 
   
   SerialBuffer#
;

: serialKeySimple ( -- char )
   begin  pause  serialKey? until  SerialBufferGet 
;

: EXPIRED ( ms -- t )
   COUNTER -  -86400000 1 WITHIN ;

: serialKey() ( -- char )     \ with timeout
   counter [ 2000 56 / ] literal ( ms ) +
   begin 
      pause
      DUP expired if  drop 1003 throw  then
   serialKey? 0= while
      \ . 100 ms
      2 Call Sleep drop  
   repeat
   drop 
   SerialBufferGet
\   cr ." SerialReadNumber = "  SerialReadNumber ?
;

: serialKeyT/O ( -- char | -1 )     \ with timeout
   ['] serialKey() catch if  -1  then 
;

: serialKey ( -- c )
   serialKeyT/O
;


\ *****************************************************************************
\ Serial character 
\ *****************************************************************************

: serialEmit- ( char -- )
   qvEmitChar !  pause
\   qvCOMh @ qvEmitChar 1 qvEmitFlag 0  Call WriteFile  0= throw 
    0  qvEmitFlag rel>abs  1  qvEmitChar  rel>abs  qvCOMh @  Call WriteFile  0= throw 
;

: serialEmit ( char -- )
   ['] serialEmit- catch if  drop  then 
;

: serialType- ( addr u -- )
   pause  
\   qvCOMh @  rot rot qvEmitFlag  0  Call WriteFile  0= throw 
   0  qvEmitFlag rel>abs 2swap  ( u -- ) swap rel>abs  qvCOMh @  Call WriteFile  0= throw 
;

: serialType ( addr u -- )
  \ over + swap ?do  i c@ serialEmit  loop
   ['] serialType- catch if  2drop  then 	\ this is necessary!!!
;   

: serialKey0       ( -- c1 )       \ must return a key
   begin   SerialKey?  until       \ loop till we get one
   SerialKey() ;

[then]


: 2ms_delay ( -- )
   \ 2 ms
   pause \ pause
;

: slowSerialType ( a n -- )
   dup 0= if  2drop  exit  then	
   over + swap do  2ms_delay  i c@ SerialEmit  loop 
;	

\ Receives n characters from the serial port and puts them straight into address a
: serialStraight ( a n -- )
   dup 0= if  2drop  exit  then
   over + swap do  
      serialKey0 i c!  
   loop 
;

\ serialPurge  cleans out the receive input buffer for the currently selected serial port
: serialPurge ( -- )   begin  serialKey? while  serialKey drop  repeat ;

: serialShow ( -- )   begin  serialKey? while  serialKey emit  repeat ;
\ : serialShow+ ( -- )   begin  serialKey? while  serialKey emit+  repeat ;

\ : serialShowHex ( -- )   begin  serialKey? while  serialKey 2.hex space  repeat ;


(( variable len-written
: serialType ( a n )
   0                                 \ lpOverlapped
   len-written rel>abs               \ lpNumberOfBytesWritten
   rot >r
   rot r> swap  	             \ nNumberOfBytesToWrite(slen)
    rel>abs                          \ *lpBuffer(adr)
   fileid                            \ hFile
   call WriteFile
   ;                                 \ ior = 0 = success
))





: do-RKEY    ( c1 -- f1 )    \ process keys pressed on user keyboard
                dup k_ESC =                     \ was ESC pressed?
                if      drop                    \ if so, discard c1 and shut down
                        ComClose       \ close the com port
                        ." Done"          TRUE  \ return TRUE, we are terminating
                else    serialEmit          FALSE \ else emit char and continue
                then    ;


: xSetComTalk ( baud com# -- ) 	\ setup the simple terminal talker to the serial port
	to ComPort
   to ComBaud
;

\ 300 1 SetComTalk

: 2P.R       ( n -- adr len )    \ display n as a 2 digit decimal number, plus a space
   BASE @ >R DECIMAL
        0 <#  32 HOLD  #S  2 OVER - 0MAX 0 ?DO  32 HOLD  LOOP #>
        R> BASE ! ;

: 5P.R       ( n -- adr len )    \ display n as a 5 digit decimal number, plus a space
   BASE @ >R DECIMAL
        0 <#  32 HOLD  #S  5 OVER - 0MAX 0 ?DO  32 HOLD  LOOP #>
        R> BASE ! ;

: 1P.R       ( n -- adr len )    \ display n as a single digit number, plus a space
        BASE @ >R DECIMAL
        0 <#  32 HOLD  #  #>
        R> BASE ! 
;

 
0 VALUE hStat

: STATUS-ZTYPE ( a - )   
   hStat 0= IF  DROP  EXIT  THEN
   >R hStat SB_SETTEXTA 0 R> Call SendMessage DROP
;           

variable StatusBuffer[] 128 allot
variable MySlowCounter

\ Path.f  filename and path support
\ 
\ -NAME strips the name and trailing \ from the end of a full pathname.
\ -EXT strips the trailing extension from the filename.
\ -PATH strips the path from the beginning of a full pathname.

: ENDING ( addr len char -- addr len )
   >R BEGIN  2DUP R@ SCAN
      ?DUP WHILE  2SWAP 2DROP  1 /STRING
   REPEAT  R> 2DROP ;

: -NAME ( a n -- a n )
   2DUP  [CHAR] \ ENDING  NIP -  1-  0 MAX ;

: -EXT ( a n -- a n )
   2DUP  [CHAR] . ENDING  NIP -  1-  0 MAX ;

: -PATH ( a n -- a n )
   [CHAR] \ ENDING  0 MAX ;

\ create the COM port data status string
: UpdateStatusBuffer ( -- )
   StatusBuffer[] 128 BLANK
   ComPort MakeCOMportName DROP COMportName count 4 /string StatusBuffer[] place
\   s" COM" StatusBuffer[] place ComPort (.) StatusBuffer[] +place 
   s"  " StatusBuffer[] +place
   ComBaud (.)  StatusBuffer[] +place
   s"  Baud "  StatusBuffer[] +place
   s"    File : "  StatusBuffer[] +place
   CurrentFilename dup c@ if  
      count  
      -Path \ only show the filename, not the full path
   else   
      drop  s" -------"  
   then   
   StatusBuffer[] +place

   UploadOrDownload @ if
      s"    Upload"  StatusBuffer[] +place
   else
      s"    Download"  StatusBuffer[] +place
  then

   OldFormat @ if
      s"  Old format"  StatusBuffer[] +place
   else
\      s" . "  StatusBuffer[] +place
  then

\  MySlowCounter @ (.) StatusBuffer[] +place

\   0 StatusBuffer[]  count + c!   \ null terminate 
;

: ComData() UpdateStatusBuffer  StatusBuffer[] count ;
\ \ : .STATUS   UpdateStatusBuffer  StatusBuffer[] 1+ STATUS-ZTYPE ;

: ttmcn  12 0 do  cr i .  space i
      COMportName 10 erase
      MakeCOMportName 10 type
   loop
;

: ttus  
   ComPort 1+ 20 mod to ComPort 
\   .STATUS
;

\ a return of 0 means OK
: InitCom ( -- f )
\   UpdateStatusBuffer  
   ComClose                \ just in case one is open...
\   100 ms 
   ComPort MakeCOMportName  ComOpen
   qvCOMh @ 0= if  -1  exit  then
   Reset: DCB
   FALSE Put: DCB.fOutxCtsFlow      \ Don't pay attention to CTS
   ComBaud 8 NOPARITY ONESTOPBIT ComSetup
   0
;

\ create and destroy a dumy window to get the status bar to update...
: ForcePaint   UpdateStatusBuffer
   RegistryPutValues
   \ exit Start: ForcePaintWindow
;

\ returns 0 if OK
: CheckComPort ( -- )
   InitCom if  Start: ComWindow  bell 10 ms bell 100 ms bell  -99 throw  then
   serialPurge
\   cr ." Comport OK... "  1000 ms
\   pad 48 erase  pad 48 serialType
;

: RR3
   InitCom if  bell  cr ." InitCom Error! "  exit  then
   s" 1 THE QUICK BROWN FOX JUMPED OVER THE LAZY DOG"
   2DUP CR  TYPE  13 serialEmit  qvCOMh @ write-file
   13 serialEmit
   s" 2 THE QUICK BROWN FOX JUMPED OVER THE LAZY DOG"
   2DUP CR  TYPE  serialType  13 serialEmit
   ComClose
   ;

\ ---------------------------------------------------------------------
\ TTY  runs a simple TeleTYpe program
\ TTY+  sets up the COM port and baudrate before running TTY
\ ---------------------------------------------------------------------

\ Block 55
   ( Host terminal emulation)

: BCKS  ( -1 +CP)  ;

: @EXECUTE ( a -- )   @ ?dup if  EXECUTE  then ;

CREATE CTRLS   ' BELL ,  ' BCKS ,  0 , 0 , 0 ,  0 ( ' PAGE(T) ,  ' CR ,
: EMIT+ ( c)   DUP 7 14 WITHIN IF  7 - CELL * CTRLS + @EXECUTE
      ELSE  DUP 0= IF  DROP  127  THEN 127 AND EMIT  THEN ;

: TTY   page ." TTY - Press F1 to exit : " CR
   InitCom drop
   BEGIN
      SerialKey? IF  SerialKey EMIT+  THEN
      EKEY? IF
         EKEY DUP $10000 AND IF  CR ." TTY done"  R> 2DROP  EXIT  THEN
\         DUP serialEmit  EMIT
         serialEmit
      THEN
   AGAIN ;

\ ---------------------------------------------------------------------
\ Tests
\ ---------------------------------------------------------------------

: ttm    begin  SerialKey? UNTIL  SerialKey ;

: UUU   InitCom   BEGIN  128 0 DO [CHAR] E serialEmit LOOP  KEY? UNTIL ;

DECIMAL


(( ---------------------------------------------------------------------
Registry key enumeration
--------------------------------------------------------------------- ))

\ Function: RegOpenKeyEx ( hkey zkey 0 security 'handle -- ior )
\ Function: RegEnumValue ( hkey index 'name 'namelen 0 'type 'data 'datalen -- ior )
\ Function: RegCloseKey ( hkey -- ior )

WinLibrary Advapi32.dll

((
: HKEY_SERIALBAUD ( -- hkey )   0 >R
   HKEY_LOCAL_MACHINE Z" HARDWARE\DEVICEMAP\SERIALBAUD" 0 KEY_READ RP@
   Call RegOpenKeyEX 0= R> AND ;

: EnumBAUDPort ( index -- zaddr )      \ uses pad and here
   256 HERE !  256 PAD !
   HKEY_SERIALBAUD >R
   R@  SWAP  HERE CELL+  HERE  0 0  PAD CELL+ PAD RegEnumValue 0=
   PAD CELL+ AND  R> Call RegCloseKey DROP ;

: TTBP   3 0 DO  CR I EnumBAUDPort zcount TYPE  LOOP ;
))

variable disposition
variable hKey
variable regtype
variable reglen

named-new$ ReturnedKey$

: ZCOUNT ( zaddr -- addr n )
   DUP DUP IF  65535 0 SCAN DROP OVER - THEN ;

: OpenRegistryKey ( -- )  \ open a registry key in hKey , or 0 if error
   >r 
   0 hKey !

   hKey rel>abs       \ PHKEY phkResult        
   KEY_READ KEY_ENUMERATE_SUB_KEYS or \ DWORD ulOptions,     
\   KEY_ENUMERATE_SUB_KEYS \ DWORD ulOptions,     
   NULL                 \ REGSAM samDesired, 
   r> rel>abs           \ LPCTSTR lpSubKey,  
   HKEY_LOCAL_MACHINE   \ HKEY hKey,  
       
   Call RegOpenKeyExA  
   dup if 
      cr ." OpenRegistryKey failed! Error = " .  $1002 throw
   then 
\   cr ." RegKey = "  dup .
;

$100 constant MAX_PATH
create RegComName[] MAX_PATH allot
create RegComData[] MAX_PATH allot
create RegComKeyName[] MAX_PATH allot
create RegComValueName[] MAX_PATH allot

variable vRegComValueSize
variable vRegComDataSize
variable vRegComKeyNameSize

: EnumRegistryKey ( index hkey -- )      
   >r  
   >r 
   0 RegComKeyName[] c!
   MAX_PATH vRegComKeyNameSize ! 
\   s" Enum" RegComData[] place 
\   0 RegComData[] count +  c!

\  Note the reversed order for Win32Forth :
   NULL                     \ PFILETIME lpftLastWriteTime 
   NULL                     \ LPDWORD lpcClass,           
   NULL                     \ LPTSTR lpClass,             
   NULL                     \ LPDWORD lpReserved,         
   vRegComKeyNameSize rel>abs \ LPDWORD lpcName,           
   RegComKeyName[] rel>abs \ LPTSTR lpName,              
   r> ( index )             \ DWORD dwIndex,              
   r>                       \ HKEY hKey,                  
   Call RegEnumKeyEx
   cr ." RegEnumKeyEx = " . 
exit
   234 =  
   dup 
   if
      cr ." RegEnumKeyEx failed! error = "  . 
\      $1001 throw 
      0  0 
   else
      drop
      RegComKeyName[] 
      -1 
   then
;

: EnumRegistryValue ( index -- )      
   >r 
   0 RegComData[] c!
   MAX_PATH vRegComValueSize ! 
   MAX_PATH vRegComDataSize ! 

\  Note the reversed order for Win32Forth :
   vRegComDataSize rel>abs  \ LPDWORD lpcbData       
   RegComData[] rel>abs     \ LPBYTE lpData,         
   NULL                     \ LPDWORD lpType,        
   NULL                     \ LPDWORD lpReserved,    
   vRegComValueSize rel>abs \ LPDWORD lpcchValueName,
   RegComValueName[] rel>abs \ LPTSTR lpValueName,    
   r> ( index )             \ DWORD dwIndex,         
   hKey @                   \ HKEY hKey,             
   Call RegEnumValue
   ?dup if  
      cr ." RegEnumValue = " . 
      hKey @ Call RegCloseKey drop
      5000 throw 
   then 
;

: TTCP   
   256 0 do 
      0 RegComData[] c!
      0 RegComValueName[] c! 
      i z" HARDWARE\DEVICEMAP\SERIALCOMM" OpenRegistryKey
      I EnumRegistryValue 
      hKey @ Call RegCloseKey drop 
      cr i . RegComData[] zcount TYPE 
      cr ."    Value = " RegComValueName[] zcount TYPE 
   loop 
;

6 constant #ITEMS

: TTCP2 ( -- ) 
\   drop
   100000 0 do 
      0 RegComData[] c!
      0 RegComValueName[] c!
\      i z" SYSTEM\CurrentControlSet\Enum\USB"  OpenRegistryKey
\      i z" HARDWARE\DEVICEMAP\SERIALCOMM" OpenRegistryKey
      i z" SYSTEM\CurrentControlSet\Services\serenum\Enum" OpenRegistryKey
      i EnumRegistryValue 
\      i EnumRegistryKey
      hKey @ Call RegCloseKey drop 
      cr i . RegComData[] zcount TYPE 
\      cr ."    Value = " RegComValueName[] zcount TYPE 
\      cr ."    Name  = " RegComKeyName[] zcount TYPE 
\      cr ."    Value = " RegComValueName[] zcount TYPE 
   loop 
;

: TTCP3   
   256 0 do 
      0 RegComData[] c!
      0 RegComValueName[] c! 
      
      i z" SYSTEM\CurrentControlSet\Enum\USB\Vid_1a86&Pid_7523\5&38bcb265&0&2\Device Parameters"  OpenRegistryKey
\      i z" SYSTEM\CurrentControlSet\Enum\ACPI\PNP0501\1\Device Parameters"  OpenRegistryKey

      I EnumRegistryValue 
      hKey @ Call RegCloseKey drop 
      cr i 2 u.r 2 spaces  RegComValueName[] zcount TYPE  
      cr 4 spaces  RegComData[] zcount TYPE
      RegComData[] comopen  qvCOMh @ ."   >>> " .  comclose
   loop 
;
: TTU1 z" USB\Vid_1a86&Pid_7523\5&38bcb265&0&2" comopen  qvCOMh @ .  comclose ;
: TTU2 
z" \??\USB#Vid_1a86&Pid_7523#5&38bcb265&0&2#{a5dcbf10-6530-11d2-901f-00c04fb951ed}" comopen  qvCOMh @ .  comclose ;
: TTU3 z" COM5" comopen  qvCOMh @ .  comclose ;
: TTU4 
z" \\?\USB#Vid_1a86&Pid_7523#5&38bcb265&0&2#{a5dcbf10-6530-11d2-901f-00c04fb951ed}" comopen  qvCOMh @ .  comclose ;


\s

LONG WINAPI RegOpenKeyEx(
  __in        HKEY hKey,
  __in_opt    LPCTSTR lpSubKey,
  __reserved  DWORD ulOptions,
  __in        REGSAM samDesired,
  __out       PHKEY phkResult
);

LONG WINAPI RegEnumKey(
  __in   HKEY hKey,
  __in   DWORD dwIndex,
  __out  LPTSTR lpName,
  __in   DWORD cchName
);

LONG WINAPI RegEnumValue(
  __in         HKEY hKey,
  __in         DWORD dwIndex,
  __out        LPTSTR lpValueName,
  __inout      LPDWORD lpcchValueName,
  __reserved   LPDWORD lpReserved,
  __out_opt    LPDWORD lpType,
  __out_opt    LPBYTE lpData,
  __inout_opt  LPDWORD lpcbData
);

LONG WINAPI RegEnumKeyEx(
  __in         HKEY hKey,
  __in         DWORD dwIndex,
  __out        LPTSTR lpName,
  __inout      LPDWORD lpcName,
  __reserved   LPDWORD lpReserved,
  __inout      LPTSTR lpClass,
  __inout_opt  LPDWORD lpcClass,
  __out_opt    PFILETIME lpftLastWriteTime
);


LONG WINAPI RegQueryValueEx(
  __in         HKEY hKey,
  __in_opt     LPCTSTR lpValueName,
  __reserved   LPDWORD lpReserved,
  __out_opt    LPDWORD lpType,
  __out_opt    LPBYTE lpData,
  __inout_opt  LPDWORD lpcbData
);


\s

(( ---------------------------------------------------------------------
Configuration dialog
--------------------------------------------------------------------- ))

DIALOG Configure_Port-TEMPLATE
[MODAL " Select Serial Port" 10 20 80 74 (FONT 10, MS SANS SERIF)]
 [DEFPUSHBUTTON     " OK"                         1    44   58   32   12 ]
 [PUSHBUTTON        " Cancel"                     2    4    58   32   12 ]
 [LISTBOX                                         100  4    14   72   40 (+STYLE LBS_NOINTEGRALHEIGHT ) ]
 [LTEXT             " Select port"                101  4    4    64   10 ]
END-DIALOG

\  [LISTBOX                                         100  4    14   72   40 (+STYLE LBS_SORT LBS_NOINTEGRALHEIGHT	) ]
variable mycount
variable myPort

128 BUFFER: MYLIST

\ ZSKIP moves to the end of an ASCIIZ string.
: ZSKIP ( addr -- addr' )
   zcount + 1+ ;

: FILL-COMMPORTS ( -- )
   0 myPort !  0 mycount !
   HWND 100 GetDlgItem >R  0
   BEGIN
      DUP EnumRegistryKey
   WHILE
      1 MYCOUNT +!
      dup zcount ComName over COMPARE 0= if
         mycount @ myPort !  then
      R@ LB_ADDSTRING ROT 0 SWAP SendMessage DROP
      1+
   REPEAT DROP DROP
   myPort @ if   R@ LB_SETCURSEL  myPort @ 1- 0 max  0 SendMessage DROP
   R@ LB_GETITEMDATA 0 0 SendMessage  MYLIST !
   then
   r> drop
   ;

: ttf  cr ." myPort  ComName comname2 "
   cr  myPort @ 6 u.r   4 spaces
   ComName zcount type 2 spaces ;

: DONE ( --  flag)
   HWND 100 GetDlgItem DUP
   LB_GETCURSEL 0 0 SendMessage  DUP 0< IF  2DROP FALSE EXIT THEN
   LB_GETTEXT SWAP ComName SendMessage DROP  TRUE ;

[SWITCH Configure_Port-COMMANDS DROP ( wparam -- )
   IDOK RUN: HWND DONE  EndDialog ;
   IDCANCEL RUN: HWND FALSE  EndDialog ;
\   IDCANCEL RUN: TTT  ;
SWITCH]

[SWITCH Configure_Port-MESSAGES ZERO ( msg -- flag )
   WM_COMMAND RUN: WPARAM LOWORD Configure_Port-COMMANDS ;
   WM_INITDIALOG RUN: FILL-COMMPORTS 0 ;
SWITCH]

:NONAME ( -- flag )   MSG LOWORD Configure_Port-MESSAGES ;  4 CB: Configure_Port-CALLBACK

\ SELECT-PORT returns true if a port was selected, false if the cancel button was pressed
: SELECT-PORT() ( - f )   HWND
   HINST Configure_Port-TEMPLATE ROT Configure_Port-CALLBACK
   0 DialogBoxIndirectParam
   DUP IF  COM-OPEN() ComName zcount !PORT  THEN
   UpdateStatusBuffer
;

: SELECT-PORT   ['] SELECT-PORT() catch if  0  then ;


\\

BOOL CEnumerateSerial::UsingRegistry(CStringArray& ports)
{
    //Make sure we clear out any elements which may already be in the array(s)
    ports.RemoveAll();

    //What will be the return value
    BOOL bSuccess = FALSE;

    HKEY hSERIALCOMM;
    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("HARDWARE\\DEVICEMAP\\SERIALCOMM"), 0, KEY_QUERY_VALUE, &hSERIALCOMM) == ERROR_SUCCESS)
    {
        //Get the max value name and max value lengths
        DWORD dwMaxValueNameLen;
        DWORD dwMaxValueLen;
        DWORD dwQueryInfo = RegQueryInfoKey(hSERIALCOMM, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &dwMaxValueNameLen, &dwMaxValueLen, NULL, NULL);
        if (dwQueryInfo == ERROR_SUCCESS)
        {
            DWORD dwMaxValueNameSizeInChars = dwMaxValueNameLen + 1; //Include space for the NULL terminator
            DWORD dwMaxValueNameSizeInBytes = dwMaxValueNameSizeInChars * sizeof(TCHAR);
            DWORD dwMaxValueDataSizeInChars = dwMaxValueLen/sizeof(TCHAR) + 1; //Include space for the NULL terminator
            DWORD dwMaxValueDataSizeInBytes = dwMaxValueDataSizeInChars * sizeof(TCHAR);

            //Allocate some space for the value name and value data
            ATL::CHeapPtr<TCHAR> szValueName;
            ATL::CHeapPtr<BYTE> byValue;
            if (szValueName.Allocate(dwMaxValueNameSizeInChars) && byValue.Allocate(dwMaxValueDataSizeInBytes))
            {
                bSuccess = TRUE;

                //Enumerate all the values underneath HKEY_LOCAL_MACHINE\HARDWARE\DEVICEMAP\SERIALCOMM
                DWORD dwIndex = 0;
                DWORD dwType;
                DWORD dwValueNameSize = dwMaxValueNameSizeInChars;
                DWORD dwDataSize = dwMaxValueDataSizeInBytes;
                memset(szValueName.m_pData, 0, dwMaxValueNameSizeInBytes);
                memset(byValue.m_pData, 0, dwMaxValueDataSizeInBytes);
                LONG nEnum = RegEnumValue(hSERIALCOMM, dwIndex, szValueName, &dwValueNameSize, NULL, &dwType, byValue, &dwDataSize);
                while (nEnum == ERROR_SUCCESS)
                {
                    //If the value is of the correct type, then add it to the array
                    if (dwType == REG_SZ)
                    {
                        TCHAR* szPort = reinterpret_cast<TCHAR*>(byValue.m_pData);
                        ports.Add(szPort);
                    }

                    //Prepare for the next time around
                    dwValueNameSize = dwMaxValueNameSizeInChars;
                    dwDataSize = dwMaxValueDataSizeInBytes;
                    memset(szValueName.m_pData, 0, dwMaxValueNameSizeInBytes);
                    memset(byValue.m_pData, 0, dwMaxValueDataSizeInBytes);
                    ++dwIndex;
                    nEnum = RegEnumValue(hSERIALCOMM, dwIndex, szValueName, &dwValueNameSize, NULL, &dwType, byValue, &dwDataSize);
                }
            }
            else
            SetLastError(ERROR_OUTOFMEMORY);
        }

        //Close the registry key now that we are finished with it
        RegCloseKey(hSERIALCOMM);

        if (dwQueryInfo != ERROR_SUCCESS)
        SetLastError(dwQueryInfo);
    }

    return bSuccess;
}

\ THIS IS BACKWARDS FOR WIN32FORTH!
: EnumRegistryKey ( index -- zaddr f )      \ uses pad and here
   0 RegComData[] c!
   MAX_PATH vRegComValueSize ! 
   MAX_PATH vRegComDataSize ! 

   
   OpenRegistryKey >R  
   R@                       \ HKEY hKey,             
   SWAP                     \ DWORD dwIndex,         
   RegComName[] rel>abs     \ LPTSTR lpValueName,    
   vRegComValueSize rel>abs \ LPDWORD lpcchValueName,
   NULL                     \ LPDWORD lpReserved,    
   NULL                     \ LPDWORD lpType,        
   RegComData[] rel>abs     \ LPBYTE lpData,         
   vRegComDataSize rel>abs  \ LPDWORD lpcbData       
   Call RegEnumValue
   R> Call RegCloseKey DROP 
   dup if  cr ." RegEnumValue failed! error = "  dup . $1001 throw then 
   drop

   RegComData[] 
;
: TTCP   
   z" HARDWARE\DEVICEMAP\SERIALCOMM" OpenRegistryKey 0= if \ ( -- hKey ) 
   255 0 do     
         i ( hKey ) over EnumRegistryKey 
      0= if  
         drop  leave 
      else   
         cr i .  
         zcount TYPE  
      then  
   loop 
   ( hKey ) Call RegCloseKey drop 
   then
;
: TTCP   
   z" HARDWARE\DEVICEMAP\SERIALCOMM" OpenRegistryKey 0= if \ ( -- hKey ) 
   256 0 do     
         i ( hKey ) over EnumRegistryKey 
      0= if  
         drop  leave 
      else   
         cr i .  
\         dup CheckUSBport
         zcount TYPE  
      then  
   loop 
   ( hKey ) Call RegCloseKey drop 
   else 
      drop
   then
;

\ read registry key value string 'vadr,vlen' 
\ from            section string 'sadr,slen'
\ return             data string 'dadr,dlen'

\ sadr,slen = the registry key section string
\ vadr,vlen = the registry key value string to read
\ dadr,dlen = the registry key data string returned

: RegGetString  { vadr vlen sadr slen -- dadr dlen }
                ReturnedKey$ off                \ initially clear return buffer
                sadr slen RegGetKey dup -1 =
                if      drop
                        ReturnedKey$ count
                        EXIT                    \ return on error, empty data
                then    >r
                MAXCOUNTED reglen !             \ init max length of string

                reglen          rel>abs
                ReturnedKey$ 1+ rel>abs
                regtype         rel>abs
                0
                vadr            rel>abs
                r@
                Call RegQueryValueEx
                if      ReturnedKey$ off
                else    reglen @ 1- 0max ReturnedKey$ c!
                then    ReturnedKey$ count
                r> Call RegCloseKey drop ;



\ zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz
#include  <windows.h>
#include <stdio.h>
void main()
{
HKEY hKey;
LONG lReturn;
DWORD dwValue = 0, dwType, dwData;
DWORD dwValName;
TCHAR szValueName[256];
TCHAR szData[1024];
BOOL bIsPath;
if(RegOpenKeyEx(HKEY_LOCAL_MACHINE,
"SYSTEM\\CurrentControlSet\\Services\\usbaudio",
0, KEY_READ, &hKey) != ERROR_SUCCESS)
{
RegCloseKey(hKey);
return;
}
else
bIsPath=TRUE;
while(bIsPath==TRUE)  {
dwData = sizeof(szData);
dwValName = sizeof(szValueName);
lReturn = RegEnumValue(hKey, dwValue++, szValueName, &dwValName,
NULL, &dwType, (UCHAR*)szData, &dwData);
if(lReturn == ERROR_NO_MORE_ITEMS)
break;
printf("%s = %s\n", szValueName, szData);
}
RegCloseKey(hKey);
}
