\ SinarF4.f  2022 Jul 31

only forth also definitions

decimal

: buffer:   dup  create  here  swap  allot  swap erase  ;
: Turnkeyed? ( -- f )   sys-free 0= ;   \ for compatability with v5.2

1     value ComPort \ COM1 to COM99
4800  value ComBaud \ the currently open COM port's baud rate

\ Compatibility
: COUNTER ( - n)   ms@ 182 10000 */ ;

variable [LeaveEarly]    0 [LeaveEarly] !  \ don't leave early yet
\ variable [StartTransfer]   0 [StartTransfer] ! \ don't start the transfer yet
VARIABLE CurrentFilename  max-path allot \ the current file name

variable Checksum 0 Checksum !
variable RemoteChecksum  0 RemoteChecksum !

\ What we want to do in the Transfer Window :
variable dummy  1024 allot
variable UploadOrDownload  	0 UploadOrDownload  !	\ true for upload of curves
variable OldFormat  0 OldFormat !		\ true for Old Format mode
variable DisplayOnly   -1 DisplayOnly !	\ true for display only, no transfer of curves

variable CurveFailed  \ A CURVE TRANSFER HAS FAILED
variable CurveSetDone \ set when all 26 curves have been transferred

variable [curve]
variable WhichKey
variable LastLine  0 LastLine !
variable TransferWindowOpen    0 TransferWindowOpen !
variable Transferring          0 Transferring !

\ fload Path.f
fload registry2.f
fload Help.f
fload Serial.f
fload Curves.f
fload Graphics.f  \ line drawing etc
fload Menu.f

fload LanguageDialog.f
fload SinarRemF5.f

variable MyEmitChar

: Hemit ( c -- )   MyEmitChar c!  MyEmitChar 1 Htype ;
 
: process_keyboardKey ( c1 -- f1 )    \ process keys pressed on user keyboard
   dup k_ESC =          \ was ESC pressed?
   if      
      drop              \ if so, discard c1 and shut down
      ComClose          \ close the com port
      ." Done"          
      TRUE              \ return TRUE, we are terminating
   else    
      SerialEmit        
      FALSE             
      \ else emit char and continue
   then    
;

: ComTalk ( -- )
 \  Init-Console
    InitCom if  bell  cr ." InitCom Error!"  exit  then
        Hpage ." Test Terminal : Press ESC to stop talking." cr
        BEGIN   key? if
          key process_keyboardKey ?EXIT
                then
                SerialKey? if
                  SerialKey
          dup 13 = if  drop HCR  else
          Hemit  Refresh: TheMainWindow  then
        then
        AGAIN
\   Hide-Console
;

: RunDemo ( -- )
   FALSE TO walking?
   next-color
   screen-height 1- to -vdots
   screen-width  1- to -hdots
   BEGIN   
      -hdots random -vdots random -hdots random -vdots random
      -vdots 0
      DO      I + swap i - 2swap
         I - swap i + 2swap
         2over 2over line
         15 random 2 <
         IF      next-color
         THEN
         I 31 and 0=
         IF      Refresh: TheMainWindow
                WINPAUSE
         THEN
      LOOP    2drop   2drop
      key?
      do-printing?
      IF      
         line-count save-count >= OR
      THEN
   UNTIL
   do-printing? 0=
   IF      
      line-count to save-count
   THEN
;

\ ---------------------------------------------------------------
\ Top Level program starts here
\ ---------------------------------------------------------------

: xInitTransfer
   0 RemoteChecksum !
   0 CurveSetDone !
   0 CurveFailed !
   0 CurrentFilename c!
;

: SetFun ( OldFormat Upload )   UploadOrDownload !  OldFormat !
   InitTransfer
   UploadOrDownload @ if
      DisplayNone  \ upload, so no curves received yet...
   else
      DisplayAll  \ download, so show all of the curves - they are all from the file
   then

   0 CurrentFilename c! 	\ force a new file to be selected

   CurrentFilename c@ 0= if
      SelectFile \ only if not already selected
   then
;

: \SetFun
  OldFormat @
  UploadOrDownload @
  SetFun
;

: keyPlus ( -- c )
   begin
      I'mDone @ if  0 exit  then 
      10 ms
   key? until
   key
;

: SinarF4() ( -- )
   Start: TheMainWindow
   GetHandle: TheMainWindow  to GetHandleOfMainWindow

   StartPos: TheMainWindow 50 + swap 50 + swap message-origin
   blue line-color
   RANDOM-INIT             \ initialize random number generator
   erase-demo

   ComClose  \ just in case it was left open

   begin
      Refresh: TheMainWindow
      keyPlus           \ handle keyboard presses
      dup 'N' = not if  bell  then
      dup 'a' 'z' 1+ within if   223 and  then	\ ignore case
      case
      'O' +k_control  of  open-demo-bitmap      endof
      '1'             of   1 to ComPort CheckComPort  ForcePaint  endof
      '2'             of   2 to ComPort CheckComPort  ForcePaint  endof
      '3'             of   3 to ComPort CheckComPort  ForcePaint  endof
      '4'             of   4 to ComPort CheckComPort  ForcePaint  endof
      '5'             of   5 to ComPort CheckComPort  ForcePaint  endof
      '6'             of   6 to ComPort CheckComPort  ForcePaint  endof
      '7'             of   ComPort 1+ 99 min to ComPort InitCom drop  ForcePaint  endof
      '8'             of   ComPort 1-  1 max to ComPort InitCom drop  ForcePaint  endof
      '9'             of   300 to ComBaud       ForcePaint  endof
      '0'             of   4800 to ComBaud      ForcePaint  endof

      'S'             of  StopTransfer          endof
      k_F1            of  Start: HelpWindow     endof
      k_F1 +k_control of  about-demo            endof
 \    k_F2            of  Start: ComWindow     endof
      k_F4 +k_control of  cr ." Quit ... "   cr  quit  endof

      k_F8 +k_control of  CheckComPort ResetServiceTimer   endof
      k_F10 +k_control of  CheckComPort SetLanguageConfig  endof

      'D'             of   0  0 SetFun ForcePaint endof \ Download mode
      'O'             of  -1  0 SetFun ForcePaint endof \ Download Old format mode
      'U'             of   0 -1 SetFun ForcePaint endof \ Upload mode
      'P'             of  -1 -1 SetFun ForcePaint endof \ Upload Old Format mode

      'N'             of                        endof \ a null button
\      'T'             of  ComTalk  endof
\      'T' +k_control  of  ComTalk  endof
      'T' +k_control  of  ResetUnitTime  endof
      'R'             of  CheckComPort ( \SetFun )  StartTransfer  endof
      'r'             of  CheckComPort ( \SetFun )  StartTransfer  endof
      'C'             of  StopTransfer ForcePaint endof
      'c'             of  StopTransfer ForcePaint endof
      'Z'             of  RunDemo  endof

\      'Q'             of  abort" Test mode..." quit  endof
      'P' +k_control  of  ttus ForcePaint endof

     k_esc            of  StopTransfer ForcePaint  endof
     k_esc +k_control of                           endof
\      'P' +k_control +k_shift of  GetHandle: TheMainWindow
\                                  Setup: ThePrinter           endof
      endcase
      Refresh: TheMainWindow		\ update variables in Status bar etc.
   I'mDone @ until 

   ComClose

   12 to char-width
   16 to char-height
;

: SinarF4 ( -- )
   0 I'mDone !

   InitRegistry

   RegistryGetValues
   OpenCurrentFile
   0 OldFormat ! 

   UploadOrDownload @ if \ upload
      DisplayNone
   else                  \ download
      DisplayAll
   then

  \ \SetFun

\ these are now kept in the registry :
\  0 CurrentFilename c!
\  1 to ComPort
\  4800 to ComBaud
\  0 UploadOrDownload !  

\ default selection :
\   DisplayAll  \ download, so show all of the curves - they are all from the file
   Begin ['] SinarF4() Catch drop  I'mDone @ until 

   cr ." Done " cr ."  ok" cr
;

: SinarF4+     ['] Sinarf4 catch drop ;

' SinarF4+ turnkey SinarF4       \ build an application on disk

: ttt ( -- )   SinarF4 ;

\ test serial port words
: tttt ( -- ) 
   RegistryGetValues
   7 to ComPort
   cr ." Opening COM port " ComPort MakeCOMportName zcount type
   InitCom 
   if  
      cr ." InitCom failed !!!" 
   else
      cr ." InitCom opened port " ComPort MakeCOMportName zcount type space 
   then

\   serialOpen

   $55 serialEmit-
   $56 serialEmit-
   $57 serialEmit-
   
 \  10 0 do
 \     cr ." serialKey? = "  serialKey? .
 \     500 ms 
 \  loop
 \  cr ." serialKey? = "  serialKey?  .

   cr serialKey .

   .s
   
   cr ." RemoteCurveIn : " pad 5 RemoteCurveIn .
   cr ." RemoteCurveOut : " pad 5 RemoteCurveOut .

   .s
\   cr ." NewUpload : " NewUpload

   cr ." Done"

   ComClose
;

\ 5 pause-seconds

\ MODULE          \ end of the module
