\      Example code for registry use

\ The following word is executed at compile time to setup the
\ current programs base registry key.

: InitRegistry  ( -- )  \ intialize the program base registry key
               s" Sinar\F4\" progreg place ;

\ InitRegistry

variable MyNumberPad  128 allot

: ($ton) ( caddr n - u )
   -TRAILING
   MyNumberPad place 
   MyNumberPad NUMBER 
   drop
;

: $ton ( caddr n - u )
  ['] ($ton) catch if  2drop 0  then 
;

: RegistryGetValues ( -- )
   s" Filename" GetSetting CurrentFilename place
   s" ComPort"  GetSetting $ton  1 max 99 min to ComPort 
   s" Baudrate" GetSetting $ton  dup 0 = if  drop 4800  then  300 max 115200 min to ComBaud 
   s" UploadMode" GetSetting $ton 0= 0= UploadOrDownload !
;

: RegistryPutValues
    CurrentFilename count s" Filename" SetSetting
    ComPort (.) s" ComPort" SetSetting
    ComBaud (.) s" Baudrate" SetSetting
    UploadOrDownload @ (.) s" UploadMode" SetSetting
;

\s


: test          ( -- )
                s" MyVals" GetSetting type ;

\ Write the 'MyVals' value in section 'Settings' 
\ to a data string of '5,9' which would presumably be the x,y coordinate
\ of where the window should be placed next time the application starts up.

: test!         ( -- )
                s" 5,9" s" MyVals" SetSetting ;

\s


(( registry.f   The Registry Interface for Win32Forth            by Tom Zimmer

        Andrew implemented the functionality in C, and I translated it
        into Forth

))

cr .( Loading Windows Registry...)

\ Default registry key.  Change this string to put your programs registry
\ information in a place other than "Win32For\". Look for PROGREG in WinEd
\ for an example of how to change the program base registry key to a value
\ that will be specific not only to your program, but to the particular
\ directory instance of your program that is running.

create PROGREG ," F4\"   MAX-PATH allot-to

INTERNAL

  create basereg ," Sinar\"
variable disposition
variable regkey
variable regtype
variable reglen

named-new$ ReturnedKey$

EXTERNAL

\ sadr,slen = the registry section to get the key of
\ return -1 if we could not get the key

: RegGetKey     { sadr slen \ key$ -- regkey | -1 }  \ read the key of a section
                MAXSTRING localAlloc: key$
                basereg count         key$  place
                progreg count         key$ +place
                sadr slen             key$ +place
                                      key$ +NULL
                disposition rel>abs         \ we get it, but don't use it
                regkey      rel>abs         \ the return value
                NULL
                KEY_ALL_ACCESS
                REG_OPTION_NON_VOLATILE
                NULL
                0
                key$ 1+  rel>abs
                HKEY_CURRENT_USER
                Call RegCreateKeyEx 
                if      -1
                else    regkey @
                then    ;


\ read registry key value string 'vadr,vlen' 
\ from            section string 'sadr,slen'
 \ return             data string 'dadr,dlen'

\ sadr,slen = the registry key section string
\ vadr,vlen = the registry key value string to read
\ dadr,dlen = the registry key data string returned

: RegGetString  { vadr vlen sadr slen -- dadr dlen }
                ReturnedKey$ off                \ initially clear return buffer
                sadr slen RegGetKey dup -1 =
                if      drop
                        ReturnedKey$ count
                        EXIT                    \ return on error, empty data
                then    >r
                MAXCOUNTED reglen !             \ init max length of string

                reglen          rel>abs
                ReturnedKey$ 1+ rel>abs
                regtype         rel>abs
                0
                vadr            rel>abs
                r@
                Call RegQueryValueEx
                if      ReturnedKey$ off
                else    reglen @ 1- 0max ReturnedKey$ c!
                then    ReturnedKey$ count
                r> Call RegCloseKey drop ;

\ Write to the registry, a key value string 'vadr,vlen'
\ in                         section string 'sadr,slen'
\ the                           data string 'dadr,dlen'

: RegSetString  { dadr dlen vadr vlen sadr slen \ val$ khdl -- }
                MAXSTRING localAlloc: val$      \ allocate a dynamic string
                dlen "CLIP" to dlen             \ clip key to 255 characters
                sadr slen RegGetKey to khdl
                khdl -1 =
                if      EXIT                    \ just return, ignore error
                then
                dadr dlen val$ place
                          val$ +NULL
                dlen 1+                         \ data length including NULL
                val$ 1+ rel>abs                 \ null terminated data string
                REG_SZ
                0
                vadr    rel>abs
                khdl Call RegSetValueEx drop 
                khdl Call RegCloseKey drop ;

: SetSetting    ( a1 n1 a2 n2 -- )      \ a1,n1=value string, a2,n2=key string
                s" Settings" RegSetString ;

: GetSetting    ( a1,n1 -- a2 n2 )      \ a1,n1-key string, a2,n2=value string
                s" Settings" RegGetString ;

\s      Example code for registry use

\ The following word is executed at compile time to setup the
\ current programs base registry key.

: app-key-init  ( -- )  \ intialize the program base registry key
               s" F4\" progreg place

app-key-init

: test          ( -- )
                s" MyVals" GetSetting type ;

\ Write the 'MyVals' value in section 'Settings' 
\ to a data string of '5,9' which would presumably be the x,y coordinate
\ of where the window should be placed next time the application starts up.

: test!         ( -- )
                s" 5,9" s" MyVals" SetSetting ;


