\ *****************************************************************************
\ serialPlus.f  select, configure and access serial ports 
\ Howerd Oakford 2016 Sep 14  www.inventio.co.uk
\ Note : sometimes unplugging a USB serial port on Windows 7 causes a crash :
\ http://connect.microsoft.com/VisualStudio/feedback/details/140018/serialport-crashes-after-disconnect-of-usb-com-port
\ To fix this, update the driver...
\ Note : some Prolific chips will not work with the latest Windows drivers : http://www.miklor.com/COM/UV_Drivers.php
\ *****************************************************************************

decimal 

\ *****************************************************************************
\ Library imports
\ *****************************************************************************

\ 2 IMPORT: GetCommState
\ 2 IMPORT: SetCommTimeouts
\ 2 IMPORT: SetCommState
\ 2 IMPORT: GetCommProperties
0 IMPORT: PurgeComm
2 IMPORT: BuildCommDCB

\ *****************************************************************************
\ Windows error display using FormatMessage
\ *****************************************************************************

$200 constant |ErrorString| 
|ErrorString| Buffer: ErrorString[]

: .LastErrorText ( n -- )   \ displays the error message text for a Win32 error code
   dup 
   FORMAT_MESSAGE_IGNORE_INSERTS  FORMAT_MESSAGE_FROM_SYSTEM or  FORMAT_MESSAGE_ARGUMENT_ARRAY or
   NULL 
   rot ( format 0 n -- )
   0 
   ErrorString[]
   |ErrorString|
   NULL
   FormatMessage
   ?dup if
      ErrorString[] swap 2- 0 max  type  space 
   then
;

\ Display  GetLastError  values
: .LastError ( n -- )  
   dup . 2 spaces 
   dup case
   #0    of  ." Not an error at all ;)"   endof
   #1    of  ." ERROR_INVALID_FUNCTION"   endof
   #2    of  ." ERROR_FILE_NOT_FOUND"     endof
   #5    of  ." ERROR_ACCESS_DENIED"      endof
   #6    of  ." ERROR_INVALID_HANDLE"     endof
   #13   of  ." ERROR_INVALID_DATA"       endof
   #21   of  ." ERROR_NOT_READY"          endof
   #31   of  ." ERROR_GEN_FAILURE"        endof
   #50   of  ." ERROR_NOT_SUPPORTED"      endof
   #87   of  ." ERROR_INVALID_PARAMETER"  endof
   #122  of  ." ERROR_INSUFFICIENT_BUFFER" endof
   #123  of  ." ERROR_INVALID_NAME"       endof
   #259  of  ." ERROR_NO_MORE_ITEMS"      endof
   #997  of  ." ERROR_IO_PENDING"         endof
   #1110 of  ." ERROR_MEDIA_CHANGED"      endof
   #1112 of  ." ERROR_NO_MEDIA_IN_DRIVE"  endof
   #1784 of  ." ERROR_INVALID_USER_BUFFER" endof
         ." Unknown Win32 error code = " 
   endcase
   space 
   .LastErrorText
;

\ show the error and abort
: ShowLastError ( a n -- )   
   RED.  ."  GetLastError = "  GetLastError .LastError  NORMAL.  2 spaces 
   qSelect @ case
   0 of   #1110 throw  endof
   1 of   #1111 throw  endof
   #1112 throw  
   endcase
;

: ?ERROR? ( f -- )   0= if  GetLastError .LastError  then ;

{ --------------------------------------------------------------------
Com port and baud rate selection
-------------------------------------------------------------------- }
0 VALUE hAPP

\ VARIABLE mybaud
2VARIABLE WindowSize  624 400 WindowSize 2!

0 VALUE hwndFILE

{ ---------------------------------------------------------------------------
Data structures

CommConfig  is the data structure for the Windows interface

Field: defines the fields within the CommConfig structure

--------------------------------------------------------------------------- }

\ These Qvars contain data for the currently selected comm port 
Qvariable qvCOMh        \ handle 

\ Qvariable qvComNumber   \ COM port number

\ $40 Qbuffer: qbComFileName   \ null terminated name in "filename" form
Qvariable qvKeyFlag
Qvariable qvKeyChar
Qvariable qvEmitFlag
Qvariable qvEmitChar


\ *****************************************************************************
\ DCB
\ *****************************************************************************

: Field: ( u1 u2 -- u3)   create over , +  does> ( a -- a' )  @ + ;

0
\ CommConfig starts here
   4 Field: ->dwSize
   2 Field: ->wVersion
   2 Field: ->wReserved
\ DCB starts here
   4 Field: ->DCB          \ Length of DCB
   4 Field: ->BaudRate     \ Line speed, bits per second
   4 Field: ->DCBflags     \ See flags below
   2 +                     \ Not currently used
   2 Field: ->XonLim       \ Transmit XON threshold
   2 Field: ->XoffLim      \ Transmit XOFF threshold
   1 Field: ->ByteSize     \ Bits per character (4-8)
   1 Field: ->Parity       \ 0-4 --> no,odd,even,mark,space
   1 Field: ->StopBits     \ 0,1,2 --> 1,1.5,2
   1 Field: ->XonChar      \ Tx and Rx XON char
   1 Field: ->XoffChar     \ Tx and Rx XOFF char
   1 Field: ->ErrorChar    \ Error replacement char
   1 Field: ->EofChar      \ End of input char
   1 Field: ->EvtChar      \ Received event char
   2 Field: ->Reserved1    \ Reserved, do not use
  24 Field: ->Reserved2    \ Reserved, do not use
\   4 +                    \ for 64 bit systems
constant |CommConfig|

{ 
typedef struct _COMM_CONFIG {
  DWORD dwSize;
  WORD  wVersion;
  WORD  wReserved;
  DCB   dcb;   \ 36 bytes
  DWORD dwProviderSubType;
  DWORD dwProviderOffset;
  DWORD dwProviderSize;
  WCHAR wcProviderData[1];
] COMMCONFIG, *LPCOMMCONFIG;

 9 cells : 
typedef struct _DCB {
  DWORD DCBlength;
  DWORD BaudRate;
     DWORD fBinary  :1;
     DWORD fParity  :1;
     DWORD fOutxCtsFlow  :1;
     DWORD fOutxDsrFlow  :1;
     DWORD fDtrControl  :2;
     DWORD fDsrSensitivity  :1;
     DWORD fTXContinueOnXoff  :1;
     8
     DWORD fOutX  :1;
     DWORD fInX  :1;
     DWORD fErrorChar  :1;
     DWORD fNull  :1;
     DWORD fRtsControl  :2;
     DWORD fAbortOnError  :1;
     7
     DWORD fDummy2  :17;
  total 32
  WORD  wReserved;
  WORD  XonLim;
  WORD  XoffLim;
  BYTE  ByteSize;
  BYTE  Parity;
  BYTE  StopBits;
  char  XonChar;
  char  XoffChar;
  char  ErrorChar;
  char  EofChar;
  char  EvtChar;
  WORD  wReserved1;
] DCB, *LPDCB;
} 

create CommConfig   ( dwSize ) |CommConfig| , ( wVersion ) 1 h, |CommConfig| 0 ->wReserved -  /ALLOT

\ *****************************************************************************
\ *****************************************************************************

$100 Buffer: MyDCB[]

: BuildDCB ( -- a )
   MyDCB[] $100 erase
   z" baud=1200 parity=N data=8 stop=1" MyDCB[] BuildCommDCB ?ERROR?
   MyDCB[] |CommConfig| dump
;

\ *****************************************************************************

\ 7 data bits, Even parity, 1 stop bit
: set7E1 ( -- )  CommConfig >r  $11 r@ ->DCBflags !  $07 r@ ->ByteSize !  2 r@ ->Parity C!  0 r@ ->StopBits C!  r> drop ;
\ 8 data bits, no parity, 1 stop bit
: set8N1 ( -- )  CommConfig >r  $11 r@ ->DCBflags !  $08 r@ ->ByteSize !  0 r@ ->Parity C!  0 r@ ->StopBits C!  r> drop ;
\ 8 data bits, even parity, 1 stop bit
: set8E1 ( -- )  CommConfig >r  $11 r@ ->DCBflags !  $08 r@ ->ByteSize !  2 r@ ->Parity C!  0 r@ ->StopBits C!  r> drop ;

{ ---------------------------------------------------------------------
Port control

The EscapeCommFunction function directs a specified communications device
to perform an extended function.

BOOL EscapeCommFunction(
    HANDLE hFile,	// handle to communications device
    DWORD dwFunc 	// extended function to perform
   );

dwFunc	Meaning
CLRDTR	Clears the DTR (data-terminal-ready) signal.
CLRRTS	Clears the RTS (request-to-send) signal.
SETDTR	Sends the DTR (data-terminal-ready) signal.
SETRTS	Sends the RTS (request-to-send) signal.
SETXOFF	Causes transmission to act as if an XOFF character has 
  been received.
SETXON	Causes transmission to act as if an XON character has 
  been received.
SETBREAK	Suspends character transmission and places the 
  transmission line in a break state until the ClearCommBreak 
  function is called (or EscapeCommFunction is called with the 
  CLRBREAK extended function code). 
  The SETBREAK extended function code is identical to the 
  SetCommBreak function. 
  Note that this extended function does not flush data that 
  has not been transmitted.
CLRBREAK	Restores character transmission and places the
  transmission line in a nonbreak state. 
  The CLRBREAK extended function code is identical to the 
  ClearCommBreak function.


If the function succeeds, the return value is nonzero.

+DTR asserts DTR, -DTR lowers it.  And so on...

--------------------------------------------------------------------- }

2 IMPORT: EscapeCommFunction

: +RTS   ( -- )   qvCOMh @ 3 EscapeCommFunction ?ERROR? ;    \ SETRTS 
: -RTS   ( -- )   qvCOMh @ 4 EscapeCommFunction ?ERROR? ;    \ CLRRTS 
: +DTR   ( -- )   qvCOMh @ 5 EscapeCommFunction ?ERROR? ;    \ SETDTR 
: -DTR   ( -- )   qvCOMh @ 6 EscapeCommFunction ?ERROR? ;    \ CLRDTR 
: +BREAK ( -- )   qvCOMh @ 8 EscapeCommFunction ?ERROR? ;    \ SETBREAK
: -BREAK ( -- )   qvCOMh @ 9 EscapeCommFunction ?ERROR? ;    \ CLRBREAK

{ ---------------------------------------------------------------------------

serialKey?  returns true if a character has been received from the currently selected serial port

serialKey  waits for and returns a character from the currently selected serial port

serialKeyT/O waits for 1000 ms or until a key is received. Returns -1 if timed out or a keyboard key is pressed

serialEmit  sends a character to the currently selected serial port

serialPurge  cleans out the receive input buffer for the currently selected serial port

--------------------------------------------------------------------------- }

\ *****************************************************************************
\ Comm Port character access, via Qvariables
\ *****************************************************************************
#11 2** constant |SerialBuffer|
|SerialBuffer| qBuffer: SerialBuffer[]

Qvariable qSerialBufferCharsIn
Qvariable qSerialBufferCharsOut

: SerialBufferInit ( -- )
   SerialBuffer[] |SerialBuffer| erase 
   0 qSerialBufferCharsIn !
   0 qSerialBufferCharsOut ! 
;

SerialBufferInit

: SerialBuffer# ( -- n )   qSerialBufferCharsIn @ qSerialBufferCharsOut @ -  |SerialBuffer| 1- and ;

: >SerialBuffer ( n -- a )   |SerialBuffer| 1- and  SerialBuffer[] + ;

: SerialBufferPut ( c -- )
   qSerialBufferCharsIn @ >SerialBuffer  c!
   1 qSerialBufferCharsIn +!
;

: SerialBufferGet ( -- c )
   qSerialBufferCharsOut @ >SerialBuffer c@
   1 qSerialBufferCharsOut +!
;

: GetSerialBuffer ( a n -- )   SerialBuffer[] qSerialBufferCharsIn @ ;

: SerialBufferKey ( -- c ) 
   begin  pause  SerialBuffer# until  SerialBufferGet   
;

8 2** constant |SerialReadPad| 
|SerialReadPad| QBuffer: SerialReadPad[]

Qvariable SerialReadNumber

: serialKey? ( -- flag )
   SerialBuffer# ?dup if  exit  then 
\   qvKeyFlag @ dup ?EXIT drop
   qvCOMh @ SerialReadPad[] |SerialReadPad| SerialReadNumber 0 ReadFile 0= if  
      ( [char] Q EMIT(G] #1070 throw ) 
   then 
   
   SerialReadPad[] SerialReadNumber @ over + swap ?do  i c@ SerialBufferPut  loop 
   
   SerialBuffer#
;

: serialKeySimple ( -- char )
   begin  pause  serialKey? until  SerialBufferGet 
;

: serialKey() ( -- char )     \ with timeout
   counter 2000 ( ms ) +
   begin 
      pause
      DUP expired if  drop #1003 throw  then
   serialKey? 0= while
      2 ms  1 Sleep drop  
   repeat
   drop 
   SerialBufferGet
\   cr ." SerialReadNumber = "  SerialReadNumber ?
;

: serialKeyT/O ( -- char | -1 )     \ with timeout
   ['] serialKey() catch if  -1  then 
;

: serialKey ( -- c )
   serialKeyT/O
;


\ *****************************************************************************
\ Serial character 
\ *****************************************************************************

: serialEmit- ( char -- )
   qvEmitChar !  pause
   qvCOMh @ qvEmitChar 1 qvEmitFlag 0 WriteFile 0= throw 
;

: serialEmit ( char -- )
   ['] serialEmit- catch if  drop  then 
;

: serialType- ( addr u -- )
   pause  qvCOMh @  rot rot qvEmitFlag  0 WriteFile 0= throw 
;

: serialType ( addr u -- )
  \ over + swap ?do  i c@ serialEmit  loop
   ['] serialType- catch if  2drop  then 	\ this is necessary!!!
;   

: serialPurge ( -- )   begin  serialKey? while  serialKey drop  repeat ;

: serialShow ( -- )   begin  serialKey? while  serialKey emit  repeat ;
: serialShow+ ( -- )   begin  serialKey? while  serialKey emit+  repeat ;

: serialShowHex ( -- )   begin  serialKey? while  serialKey 2.hex space  repeat ;

\ *****************************************************************************
\ Status display
\ *****************************************************************************

256 CONSTANT |StatusBuffer[]|
|StatusBuffer[]| BUFFER: StatusBuffer[]

variable StatusBarUpdated
variable EnableSerialEcho  \ transmit back received serial characters down the serial line
variable StatusPad

: .StatusOnePort ( -- ) 
   s"  COM" StatusBuffer[] append
   Qselect @ >RegPortNumber @ (.) StatusBuffer[] append  
   s"  "  StatusBuffer[] append
   Qselect @ >RegBaudrate @ (.) StatusBuffer[] append 
      s"  baud "  StatusBuffer[] append
   qvCOMh @ if
      s" +   "
      else
      s" -   "
   then
   StatusBuffer[] append

   RemotePage if  \ if it is not Page A == 0
      s"   Page  " StatusBuffer[] append
      RemotePage $03 and [char] A +  StatusPad c!  StatusPad 1 StatusBuffer[] append
   then

   EnableSerialEcho @ if
      s"  *** Echo Mode ***"  StatusBuffer[] append
   then
;

variable CurrentCOM#

: UpdateStatusBuffer ( -- )
   Qselect @ >r 
   0 StatusBuffer[] c!
\   CurrentCOM# @ (.) StatusBuffer[] append 
   s"    " StatusBuffer[] append 
   COM-A .StatusOnePort
\   COM-B .StatusOnePort
   0 StatusBuffer[] count + c!   \ null terminate the buffer
   r> Qselect !
   -1 StatusBarUpdated !
;

{ ---------------------------------------------------------------------------
Configuration

COMHELP  tells the user a little about the TERM program usage.

?COMHELP  checks for a baudrate and a comport name.

SERCONFIG  runs the CommConfigDialog to let the user configure the port.

serialOpen  opens the current Qselect serial port

realSerialClose closes the current Qselect serial port

--------------------------------------------------------------------------- }

: realSerialClose ( -- )   
   qvCOMh @ if  
\      -RTS  \ disable the RTS handshake output from the PC
\      -DTR  \ disable the DTR handshake output from the PC
      qvCOMh @ CloseHandle drop 
      100 ms 
   0 qvCOMh !
   UpdateStatusBuffer
   then 
;

: SerialClose ( -- )   realSerialClose  ;  \ 


: serialClose-ALL ( -- )   
   qSelect @ >r  
   NUM_QSELECTS 0 ?do
      i Qselect ! realSerialClose  
   loop
   r> qSelect !
;

: InitSerialHandles ( -- )   
   qSelect @ >r  
   NUM_QSELECTS 0 ?do
      i Qselect ! 0 qvCOMh ! 
   loop
   r> qSelect !
;

:PRUNE  ?PRUNE  -EXIT  serialClose-ALL ;

DECIMAL

\ convert the ASCII string a , n to a double number
: MyDnumber ( a n -- d )
   2dup 2>r NUMBER? case
   0 of  ( RED.  ."  Invalid Dnumber string : " 2r@ type  NORMAL. ) 0. #9103 throw  endof
   1 of  s>d  endof
   2 of  endof
   cr ." what?"
   endcase
   2r> 2drop
;

: ComNameTo#old ( c-addr n -- u )   3 /string base @ >r  decimal  MyDnumber drop  r> base ! ;

\ : ComNameTo#New ( c-addr n -- u )   [char] C split 2drop 3 /string base @ >r  decimal  MyDnumber drop  r> base ! ;
: ComNameTo# ( c-addr n -- port# | 0 | -1 )   
   dup 4 < if  2drop  0  exit  then 
   s" COM" search(NC) \ [char] C split 2drop 
   if 
      3 /string 
      0 -rot  over + swap ?do  #10 * i c@ [char] 0 - +  loop
   else
      -1    \ the string "COM" was not found
   then
;

: yyyComNameTo# ( c-addr n -- u )   
     ['] ComNameTo# catch if  2drop  1  then
;

: (ttcnt)  
   2dup cr type  
   ComNameTo# space . 
   \ [char] C split 2drop 2dup cr type 3 /string 2dup cr type 2dup cr type 
;

: ttcnt     
   s" //./COM17" (ttcnt)
   s" //./COM1" (ttcnt)
   s" COM17" (ttcnt)
   s" COM1" (ttcnt)
   pad 0 (ttcnt)
;


\ COMM-TIMEOUTS  is an array of data to tell the open com port to
\ return immediately if no characters are waiting.
CREATE COMM-TIMEOUTS   -1 , 0 , 0 , 1 , 20 ,

$200 Buffer: TempFileName[]

: MakeComFileName ( n -- z$ )
   s" \\.\COM" TempFileName[] place
   (.) 0 max  TempFileName[] append
\   (.) TempFileName[]1 +place
   0 TempFileName[] count + c!   \ null terminate
   TempFileName[] 1+
;

variable ForceFastBaud  

: serialOpen ( -- )
   realSerialClose
   Qselect @ >RegPortname count ComNameTo# MakeComFileName
   dup zcount Qselect @ >RegPortname place
   GENERIC_READ GENERIC_WRITE OR
   0
   0
   OPEN_EXISTING
   0
   0
   CreateFile
   dup INVALID_HANDLE_VALUE = if 
\      RED. ." Cannot open port " Qselect @ >RegPortname count type  ."  !  "  RESTORE.
      #1002 throw
   else
      qvCOMh !
      qvCOMh @ COMM-TIMEOUTS SetCommTimeouts 0= if  16 throw  then 
      qvCOMh @ CommConfig ->DCB GetCommState 0= if  ( ." GetCommState" ) #17 throw  then \ get the system port state
      Qselect @ >RegBaudrate @  
\      ForceFastBaud @ if  drop  115200  then 
      CommConfig ->Baudrate ! \ set our baudrate in the DCB using data from the registry
      Qselect @ >RegPortSettings @ case
      8N1 of  set8N1  endof  \  set 8 data bits, no parity, 1 stop bit
      8E1 of  set8E1  endof  \  set 8 data bits, even parity, 1 stop bit
      7E1 of  set7E1  endof  \  set 7 data bits, Even parity, 1 stop bit
         set8N1
      endcase
      qvCOMh @  CommConfig ->DCB  SetCommState 0= if  ( ." SetCommState" ) #18 throw  then 
\      -RTS  \  the RTS handshake output from the PC
\      -DTR  \  the DTR handshake output from the PC
      +RTS  \  the RTS handshake output from the PC
      +DTR  \  the DTR handshake output from the PC
   then
   UpdateStatusBuffer
   200 ms 
;

: SetBaud ( baud -- )   Qselect @ >RegBaudrate !  SetRegValues ;
: SetPortSettings ( c -- )   Qselect @ >RegPortSettings !  SetRegValues ;

$20 constant |ShortCOMnameBuffer| 
|ShortCOMnameBuffer| Buffer: ShortCOMnameBuffer[] 

: GetzShortCOMname ( -- z$ )   ShortCOMnameBuffer[] |ShortCOMnameBuffer| erase  s" COM" ShortCOMnameBuffer[] place 
   Qselect @ >RegPortname count ComNameTo# (.) ShortCOMnameBuffer[] append
   ShortCOMnameBuffer[] count drop
;


: SELECT-BAUD ( -- )
   GetzShortCOMname  HWND  CommConfig  CommConfigDialog if
      \ qvCOMh @ ->DCB SetCommState 
      CommConfig ->Baudrate @ SetBaud 
      UpdateStatusBuffer
   else
      \ RED. cr ." SELECT-BAUD - Failed to open CommConfigDialog !  "  ShowLastError  RESTORE. 
   then
;

{ 
: SELECT-BAUD-A
   Qselect @ >r 
   COM-A SELECT-BAUD
   r> Qselect !
;

: SELECT-BAUD-B
   Qselect @ >r 
   COM-B SELECT-BAUD
   r> Qselect !
;

: SELECT-BAUD-C
   Qselect @ >r 
   COM-C SELECT-BAUD
   r> Qselect !
;

: SELECT-BAUD-D
   Qselect @ >r 
   COM-D SELECT-BAUD
   r> Qselect !
;
}

{ ---------------------------------------------------------------------------
.COM  displays the ASCII representation of the comport 

The message switch SBAR-MESSAGES is extended to include the detection
of a left button press on the com port status area of the status bar.
If a press is detected, the configuration dialog is displayed.

COM-SETTING holds miscellaneous settings:
   +0  DCBflags low byte
   +1  ByteSize
   +2  Parity
   +3  StopBits
--------------------------------------------------------------------------- }


{ ---------------------------------------------------------------------
Data structures

Qselect @ >RegPortname holds the ASCIIZ string of the COM port name.

COM-SETTING holds miscellaneous settings:
   +0  DCBflags low byte
   +1  ByteSize
   +2  Parity
   +3  StopBits

DCB: defines the fields within the COM port DCB struct.
DCB holds the copy of the DCB we will modify for setting the port.
DCBSAVE holds a saved copy from which we can restore the port settings.
--------------------------------------------------------------------- }

{ ---------------------------------------------------------------------
Registry key enumeration
--------------------------------------------------------------------- }

Function: RegOpenKeyEx ( hkey zkey 0 security 'handle -- ior )
Function: RegEnumValue ( hkey index 'name 'namelen 0 'type 'data 'datalen -- ior )
\ Function: RegCloseKey ( hkey -- ior )
\ 7 IMPORT: RegGetValue ( hkey lpSubKey lpValue dwFlags pdwType pvData zpcbData -- ior )

: HKEY_SERIALCOMM ( -- hkey )   0 >R
   HKEY_LOCAL_MACHINE Z" HARDWARE\DEVICEMAP\SERIALCOMM" 0 KEY_READ RP@
   RegOpenKeyEX 0= R> AND ;

$200 constant REG_STRING_MAX_LENGTH 

REG_STRING_MAX_LENGTH Buffer: LocalComName

: COM#ToName ( u -- z$ )
   s" COM" LocalComName place
   (.) 0 max  LocalComName append
   0 LocalComName count + c!   \ null terminate
   LocalComName 1+ 
   LocalComName Qselect @ >RegPortname  MAX_PATH move
;

{ 
LONG WINAPI RegEnumValue(
  _In_        HKEY    hKey,
  _In_        DWORD   dwIndex,
  _Out_       LPTSTR  lpValueName,
  _Inout_     LPDWORD lpcchValueName,
  _Reserved_  LPDWORD lpReserved,
  _Out_opt_   LPDWORD lpType,
  _Out_opt_   LPBYTE  lpData,
  _Inout_opt_ LPDWORD lpcbData
);
}


REG_STRING_MAX_LENGTH Buffer: EnumValueName[]
variable EnumValueNameLength
REG_STRING_MAX_LENGTH Buffer: EnumValueData[]
variable EnumValueDataLength

\ Returns a valid COM port number, -1 if not valid, or 0 if no more ports
: EnumCommPort ( index -- COM# | 0 | -1  )   
   
   REG_STRING_MAX_LENGTH EnumValueNameLength !
   REG_STRING_MAX_LENGTH EnumValueDataLength !

   HKEY_SERIALCOMM >r r@ \ hKey,          
   swap                  \ dwIndex,       
   EnumValueName[]       \ lpValueName,   
   EnumValueNameLength   \ lpcchValueName,
   0                     \ lpReserved,    
   0                     \ lpType , null means we do not want to know
   EnumValueData[]       \ lpData,        
   EnumValueDataLength   \ lpcbData       
   RegEnumValue  
   0= IF 
      EnumValueData[] zcount ComNameTo# 
   ELSE  
      0  
   THEN
   r> RegCloseKey drop 
;

: serialShowPorts ( -- )   
   256 0 do  
      i EnumCommPort dup if  
         dup -1 = not if  .  else  drop  then 
      else  
         drop  
         leave  
      then  
   loop 
;

{ ---------------------------------------------------------------------
Configuration dialog
--------------------------------------------------------------------- }

DIALOG Configure_Port-Dialog
[MODAL " Select Serial Port" 20 20 84 154 (FONT 10, MS SANS SERIF)]
 [DEFPUSHBUTTON     " OK"            1     46   134   32   12 ]
 [PUSHBUTTON        " Cancel"        2     6    134   32   12 ]
 [LISTBOX                            100   6    6     72   120 (+STYLE LBS_NOINTEGRALHEIGHT ) ]
\ [LTEXT             " Select port"   101  6    4    64   10 ]
END-DIALOG

\  [LISTBOX                                         100  4    14   72   40 (+STYLE LBS_SORT LBS_NOINTEGRALHEIGHT	) ]
variable mycount
variable myPort

128 BUFFER: MYLIST

\ ZSKIP moves to the end of an ASCIIZ string.
\ : ZSKIP ( addr -- addr' )   zcount + 1+ ;

variable MyDlgItem
variable MyList#

: wwwCurrentCOM#@ ( -- c )
   Qselect @ >RegPortNumber @   
;

: wwwCurrentCOM#! ( c -- )
   Qselect @ >RegPortNumber !
;

variable MyHighlight

: FILL-COMMPORTS-GD1 ( -- )
\   InitValidPorts
   Qselect @ >RegPortname count ComNameTo# CurrentCOM# !
   0 myPort !  
   0 mycount !
   0 MyList# !
   0 MyHighlight ! 
   HWND 100 GetDlgItem MyDlgItem !  
   begin
      mycount @ EnumCommPort dup myPort ! 
   while
      1 mycount +!
\      cr dup .
      myPort @ 1 #99 1+ within if  
         1 MyList# +! 
         myPort @ CurrentCOM# @ = if  MyList# @ MyHighlight !  then
\         dup zcount ComName over COMPARE 0= if  mycount @ myPort !  then
\         dup zcount type  
\         MyDlgItem @ LB_ADDSTRING rot COM#ToName 0 swap SendMessage drop
         MyDlgItem @ LB_ADDSTRING 0 EnumValueData[] SendMessage drop
      then
   repeat 
   MyHighlight @ if  \ highlight the currently selected value
      MyDlgItem @ LB_SETCURSEL  MyHighlight @ 1- 0 max 0 SendMessage DROP
      MyDlgItem @ LB_GETITEMDATA 0 0 SendMessage  MYLIST !
   then
;

: typeR ( a c c -- )   >r  dup >r type  r> r> swap - 0 max  spaces ;


: .COMportSettings ( -- )
   Qselect @ >RegPortSettings @ case
      8N1 of  ."  8N1"  endof  \  set 8 data bits, no parity, 1 stop bit
      8E1 of  ."  8E1"  endof  \  set 8 data bits, even parity, 1 stop bit
      7E1 of  ."  7E1"  endof  \  set 7 data bits, Even parity, 1 stop bit
      RED.  ." Unknown COMportsettings !!! = "  Qselect @ >RegPortSettings @ 8 .hex  RESTORE.
   endcase
;

Qvariable SerialPortStatus \ -1 to exit, 1 to change the COM port number, or 2 to change the baudrate 

: .COMstatus ( c -- )  \ c selects the current comm port to highlight
   base @ >r  decimal
   ." COM-"  Qselect @ [char] A + emit 2 spaces 
   BLUE. Qselect @ >RegPortname count 5 typeR   space  Qselect @ >RegBaudrate @ 6 u.r  .COMportSettings  space
   Qselect @ = if  ."  <--- "  then 

   qvCOMh @ 0= not ( SerialPortStatus @ 0= and ) if  ."  + "  then 
   RESTORE.
   r> base ! 
;

: .COMs 
   GetRegValues
   Qselect @ >r 
   cr  ."   A  "  COM-A r@ .COMstatus  
   cr  ."   B  "  COM-B r@ .COMstatus
   cr  ."   C  "  COM-C r@ .COMstatus  
   cr  ."   D  "  COM-D r@ .COMstatus
   r> Qselect !
;

: ttports  cr ." myPort  ComNameA ComNameB "
   cr  myPort @ 6 u.r   4 spaces
   Qselect @ >RegPortname count type 2 spaces ;

: DONE ( --  flag)
   HWND 100 GetDlgItem DUP
   LB_GETCURSEL 0 0 SendMessage  DUP 0< IF  2DROP FALSE EXIT THEN
   LB_GETTEXT SWAP LocalComName SendMessage DROP  TRUE
;

\ *****************************************************************************
\ *****************************************************************************

{  
INT_PTR WINAPI DialogBoxIndirectParam(
  _In_opt_  HINSTANCE hInstance,
  _In_      LPCDLGTEMPLATE hDialogTemplate,
  _In_opt_  HWND hWndParent,
  _In_opt_  DLGPROC lpDialogFunc,
  _In_      LPARAM dwInitParam
);

typedef struct {
  DWORD style;
  DWORD dwExtendedStyle;
  WORD  cdit;
  short x;
  short y;
  short cx;
  short cy;
] DLGTEMPLATE, *LPDLGTEMPLATE;
} 

\ *****************************************************************************

[SWITCH Configure_Port-COMMANDS DROP ( wparam -- )
   IDOK RUN: HWND DONE  EndDialog ;
   IDCANCEL RUN: HWND FALSE  EndDialog ;
\   IDCANCEL RUN: TTT  ;
SWITCH]

[SWITCH Configure_Port-MESSAGES ZERO ( msg -- flag )
   WM_COMMAND RUN: WPARAM LOWORD Configure_Port-COMMANDS 0 ;
   WM_INITDIALOG RUN: FILL-COMMPORTS-GD1 0 ;
SWITCH]

:NONAME ( -- flag )   MSG LOWORD Configure_Port-MESSAGES ;  4 CB: Configure_Port-CALLBACK

\ VARIABLE vStopPort  0 vStopPort !

\ returns true if a port was selected, false if the cancel button was pressed
: SELECT-PORT() ( -- f )   HWND
   HINST Configure_Port-Dialog ROT Configure_Port-CALLBACK
      0 DialogBoxIndirectParam
;

\ defer DoDeInitGatheringTask
\ defer DoInitGatheringTask
\ variable Logging?    \ true when logging 

: SELECT-PORT ( -- )   
\   -1 vStopPort !  
\   Logging? @ >r 
\   DoDeInitGatheringTask
\   100 MS  pause pause pause 

   GetRegValues
   Qselect @ >RegPortNumber @ CurrentCOM# !

   SELECT-PORT() 
   if  
      LocalComName zcount 0 max REG_STRING_MAX_LENGTH 1- min 2dup Qselect @ >RegPortname place  
      COMnameTo# dup 1 #99 1+ within if  Qselect @ >RegPortNumber !  else  drop  then
      SetRegValues
   else
      #9201 throw
   then
\   r> 2 = if  DoInitGatheringTask  then
\   0 vStopPort !
   UpdateStatusBuffer
;


: SetPort ( c -- )   MakeComFileName zcount Qselect @ >RegPortname place  SetRegValues ;

: ShowOurBaudrate ( -- )   BLUE.  -1 .COMstatus  RESTORE. ;

: serialOpen+ ( -- )   
   ['] serialOpen catch if  
   SELECT-PORT  cr ShowOurBaudrate cr  
      ['] serialOpen catch if 
         cr RED.  ." Failed to open the serial port!"  RESTORE.
         2000 ms
         #9123 throw
      then
   then 
;

: serialOpenAll ( -- )   
   Qselect @ >r  
   NUM_QSELECTS 0 ?do
      i Qselect ! serialOpen+
   loop
   r> Qselect ! 
;



\ *****************************************************************************
\ TTY  runs a simple TeleTYpe program
\ *****************************************************************************
\ Host terminal 
: bell2  1000 200 beep drop ;
: bcks2  ( -1 +cp)  ;
: doCR2 ( -- )   ." <cr>" ;
: doLF2 ( -- )   cr ; \ ." <lf>"  cr ;

create ctrlsLF   0 ,  ' bcks2 ,  0 , ' doLF2 , 0 ,  0 ( ' page(t) ,  ' doCR2 ,
: SPemitLF ( c)   dup 7 14 within if  7 - cell * ctrlsLF + @execute
      else  dup 0= if  drop  127  then 127 and emit  then ;

: serialShow1   COM-A serialShow ;
: serialShow2   COM-B serialShow ;

$7E constant HDLC_FLAG_

\ Host terminal 
: bcks  ( -1 +cp)  ;
\ : doCR ( -- )   ." <CR>"  cr ; \ ." <cr>" ;
\ : doLF ( -- )   ." <LF>"  ( $0A emit ) ; \  exit ."  <-lf-> " ; \ ." <lf>"  cr ;
: doCR ( -- )   ." <CR>" cr ; 
: doLF ( -- )   ." <LF>" ( $0A emit ) ; \  exit ."  <-lf-> " ; \ ." <lf>"  cr ;

create ctrls   0 ,  ' bcks ,  0 , ' doLF , 0 ,  0 ( ' page(t) ,  ' doCR ,
: SPemit+ ( c)   dup 7 14 within if  7 - cell * ctrls + @execute
      else  dup 0= if  drop  127  then 127 and emit  then ;

: SPtype+ ( a n -- )   over + swap ?do  i c@ SPemit+  loop ;

: serialShow++ ( -- )   begin  serialKey? while  SerialBufferGet SPemit+  repeat ;
: SPtypeLF ( a n -- )   over + swap ?do  i c@ SPemit+  loop ;

: UUU   serialOpen   BEGIN  128 0 DO [CHAR] U serialEmit LOOP  KEY? UNTIL  realSerialClose ;
: ABC   serialOpen   BEGIN  128 0 DO s" ABC" serialType  LOOP  KEY? UNTIL  realSerialClose ;

\ a serial terminal on the current COM port
\ Usage :  COM-A  1 SetPort  9600 SetBaud  TTY
\ Usage :  COM-A  4 SetPort  7N1 SetPortSettings  2400 SetBaud  TTY  ( 7 bit even parity and 1 stop bit )
: InitAllMyCOMports ( -- )
   GetRegValues
   COM-A  115200 SetBaud 1 SetPort
   COM-B  115200 SetBaud 2 SetPort
   COM-C  115200 SetBaud 3 SetPort
   COM-D  115200 SetBaud 4 SetPort
;


2variable StartPositionTTY
variable v_stringLength

: NewString ( -- )
   cr CYAN. ." 01234567890123456789012345678901234567890123456789012345678901234567890123456789" RESTORE.
   cr
;

: emit+cr ( c -- )
   dup $0A = if  drop  [char] l  then
   dup $0D = if  drop  [char] c  then
   emit
;

: (TTY) ( -- )

   Get-XY StartPositionTTY 2! 
   
   BLUE.

   0 SerialPortStatus ! 

   begin 
      
      GetRegValues
   
      ['] serialOpen catch if  1 SerialPortStatus !  then 
      
      SerialPortStatus @ 1 = if  SELECT-PORT  SetRegValues  ['] serialOpen catch if  1 SerialPortStatus !  -1 StatusBarUpdated !  then  then
      SerialPortStatus @ 2 = if  SELECT-BAUD  SetRegValues  ['] serialOpen catch if  1 SerialPortStatus !  -1 StatusBarUpdated !  then  then

      0 SerialPortStatus ! 

      ['] serialOpen catch if  1 SerialPortStatus !  then 

      StartPositionTTY 2@ At-XY
      MAGENTA.  ." Serial Terminal - F1 Help, F2 Port, F3 Baud, F8 Exit "  -1 .COMstatus  RESTORE.  
      cr
      2 v_stringLength !
      begin
         2 ms  1 Sleep drop  \ pause 2 ms
         begin
         serialKey? while  
            SerialBufferGet  
            \ dup HDLC_FLAG_ = if  cr  then  
  \          2.hex space
            dup $3E = if  
               cr ." Length = " v_stringLength @  .  1 v_stringLength ! 
               NewString
            else
               1 v_stringLength +!
               v_stringLength @ 1- $0F and 0= if  ( cr ) then 
            then 
            emit+cr \  2.hex space  \ emit+ \ 
            \ SPemit+ 
\         emit+
         repeat
         EKEY? if
            EKEY case
            $10070 ( F1 )     of  ( ShowHelp )          endof  \ 
            $10071 ( F2 )     of  1 SerialPortStatus !  endof  \ SELECT-PORT
            $10072 ( F3 )     of  2 SerialPortStatus !  endof  \ SELECT-BAUD
            $10073 ( F4 )     of                        endof  \ 
            $10074 ( F5 )     of                        endof  \ 
            $10075 ( F6 )     of                        endof  \ 
            $10076 ( F7 )     of                        endof  \ 
            $10077 ( F8 )     of -1 SerialPortStatus !  endof  \ exit 
            $10078 ( F9 )     of                        endof  \ 
            $10079 ( F10)     of                        endof  \ 
            $1007A ( F11)     of                        endof  \ 
            $1007B ( F12)     of                        endof  \ 
            $30073 ( ctrlF4 ) of  cr ."  ok " cr  quit  endof
\               MAGENTA. dup $20 < not if  EMIT  else  drop  then  RESTORE.
                DUP INVERSE.  SPemit+  RESTORE.
               dup serialEmit  
            endcase
            then

\         MenuButtonPressed @ if
\            'ProcessMenuActions @EXECUTE
\            0 MenuButtonPressed ! 
\         then

\         StatusBarUpdated @ if
\            UpdateStatusWindow
\            0 StatusBarUpdated !
\         then
      
     SerialPortStatus @ until 
      
   SerialPortStatus @ -1 = until
   
   realSerialClose  
   cr
   MAGENTA.  ." Serial Terminal closed "  -1 .COMstatus  RESTORE.  
   cr
   
   RESTORE.  
;

: TTY-A ( -- )
   GetRegValues
   COM-A  \ 9600 SetBaud  
   (TTY)
;

: TTY-B ( -- )
   GetRegValues 
   COM-B   \ 9600 SetBaud  
   (TTY)
;

: TTY-C ( -- )
   GetRegValues
   COM-C   \ 9600 SetBaud  
   (TTY)
;

: TTY-D ( -- )
   GetRegValues
   COM-D   \ 9600 SetBaud  
   (TTY)
;

\ starter TTY-A  Program TTY-A
\ starter TTY-B  Program TTY-B
\ starter TTY-C  Program TTY-C
\ starter TTY-D  Program TTY-D

decimal
: SetAllPorts ( -- )
   COM-A  17 SetPort  300 Setbaud  
   COM-B  18 SetPort  300 Setbaud  
   COM-C  17 SetPort  4800 Setbaud  
   COM-D  18 SetPort  4800 Setbaud  
   COM-A
;

\ Add n bytes to the structure, using the n hex ASCII words following, then ignore the rest of the line
\ This allows captured packet data to be copied and pasted...
\ : nc, ( -- )  base @ >r  hex
\    ( n -- ) 0 do   bl word  count number  c,  loop
\    r> base !
\    postpone \
\ ;

\ Add up to 128 bytes to a structure, using the 128 hex ASCII words following, then ignore the rest of the line.
: lineC, ( -- )
   base @ >r  hex                                     \ save the current number base on the return stack and change to hexadecimal
   $80 0 do                                           \ loop up to 128 times, or until something makes us "leave"
      bl word  count
      dup 0= if  2drop  leave  then                   \ there are no more words on the line, so leave the loop
      2dup s" \" compare 0= if  2drop  leave  then    \ we have found a "comment until end of line" character, so leave the loop
      2dup NUMBER? case      \ NUMBER? returns 0, 1 or 2 plus 0, 1 or 2 cells on the stack
         0 of
            dup 1 = if
               over c@ [char] " = if  [CHAR] " STRING  then \  " AAAA" will be added as a counted string AAAA with null terminator
               over c@ [char] % = if  [CHAR] % STRING  then \  % BBBB% will be added as a counted string BBBB with null terminator
               2drop
            else
               \ "AAAA" will be added as the string AAAA , %BBBB% will be added as the string BBBB
               over c@ [CHAR] " = if  1 /string  over + swap ?do  i c@ dup [CHAR] " = if  drop  leave  else  c,  then  loop  then
               over c@ [CHAR] % = if  1 /string  over + swap ?do  i c@ dup [CHAR] % = if  drop  leave  else  c,  then  loop  then
            then
         endof
         1 of  c,  2drop  endof
         2 of  ,   2drop  endof
      endcase
   loop
   r> base !                                          \ restore the number base from the return stack
   postpone \                                         \ perform the action of "\", ignore everything until the end of the line
                                                      \ "postpone"  means don't do this now, but when  lineC,  is called
;


create CP7070Packet ( -- a )  $48 c, 
\ 01234567890123456789012345678901234567890123456789012345678901234567890123456789
lineC, 3E 20 32 30 30 30 20 4A 61 6E 20 31 31 20 31 39
lineC, 3A 35 34 3A 32 32 20 20 35 36 38 36 20 31 32 38
lineC, 20 38 36 2E 34 20 38 36 2E 34 20 32 36 2E 37 20
lineC, 38 36 2E 35 20 41 42 43 20 20 20 20 20 20 20 20
lineC, 20 20 20 20 20 20 0D 0A

: ttCP ( -- )   
   cr ." 01234567890123456789012345678901234567890123456789012345678901234567890123456789"
   CP7070Packet count cr 2dup type 
;

\\

: xCOMKEY2   begin  2 ms serialKEY? while  SerialBufferGet dup $0A = if  drop  cr  else 
\   3 .hex  
      SPemit+  
   then   repeat ;

: xxCOMKEY2   
   begin  
      serialKey 
      dup 
   \   3 .hex  
      SPemit+  
   $0A = until
;

: COMKEY2   
      serialKey 
   \   3 .hex  
      SPemit+  
;


decimal
: SetAllPorts
   COM-A  17 SetPort  300 Setbaud  
   COM-B  18 SetPort  300 Setbaud  
   COM-C  17 SetPort  4800 Setbaud  
   COM-D  18 SetPort  4800 Setbaud  
   COM-A
;

\ display characters on both COM-A and COM-B
: BOTH ( -- ) 
   
   GetRegValues
   
   COM-A serialOpen
   COM-B serialOpen

   cr CYAN. ." BOTH Ports  - Press F1 to exit : " COM-A -1 .COMstatus 2 spaces COM-B -1 .COMstatus   RESTORE.  COM-A  cr
   begin
      COM-A  begin  serialKey? while  MAGENTA. COMKEY2 RESTORE. repeat
      COM-B  begin  serialKey? while  BLUE. COMKEY2  RESTORE. repeat
      EKEY? if
         EKEY dup $06 ( ctrlF ) = if  cr ."  ok " cr  QUIT  then
         dup $10000 and if  COM-A serialClose  COM-B realSerialClose  COM-A  cr ." BOTH done."  cr  exit  then
\         dup serialEmit  EMIT
          drop  \ serialEmit
      then
   again ;

{ --------------------------------------------------------------------- 
TTY  runs a simple TeleTYpe program
\TTY  sets up the COM port and baudrate before running TTY
--------------------------------------------------------------------- }

: /BUF ( -- )   8 qvCOMh @ PurgeComm DROP  ;

\ : COM-SHOW   begin  (COM-KEY?) while  (COM-KEY) EMIT+  repeat ;


\ --------------------------------------------------------------------- 
\ Character matching and capture
\ --------------------------------------------------------------------- 

$400 constant MAX_DATA_SIZE

MAX_DATA_SIZE constant |MyBuffer| 

|MyBuffer| Buffer: MyBuffer
variable ptrMyBuffer
variable ShowMatching

: >MyBuffer ( -- a )  ptrMyBuffer @ |MyBuffer| 1- and MyBuffer + ;

: MyBufferPut ( c -- )   
\   ShowMatching @ if  dup EMIT+  then 
   >MyBuffer c!  1 ptrMyBuffer +! 
;

: COM-CAPTURE-LINE   
   0 ptrMyBuffer !
   begin  serialKey? while  SerialBufferGet DUP 10 = IF  DROP EXIT  THEN  MyBufferPut  repeat 
;

: COM-SHOW-LINE      0 ptrMyBuffer !
   begin  serialKey? while  SerialBufferGet DUP 10 = IF  DROP EXIT  THEN  dup SPemit+  MyBufferPut  repeat 
;

variable #ToMatch

: COM-KEYW ( -- c )   begin  PAUSE serialKey? UNTIL  SerialBufferGet ;

\ match one character
: Match1 ( c -- )     
   COM-KEYW    
\   dup emit
   dup >r =    
   if  
      r> MyBufferPut       \ add to the buffer
   else
      r> drop  0 ptrMyBuffer !   \ clear the buffer
   then
;

\ wait until the received characters match string a n
: MATCH ( a n -- )    
   #ToMatch !
   0 ptrMyBuffer !
   serialPurge
   begin
      dup c@ Match1  
      1+             \ next character in string for next time 
      #ToMatch @ ptrMyBuffer @ = until
   drop
;

\ collect n characters 
: COLLECT ( n -- )  
   0 ?do 
      serialKey MyBufferPut \ add to the buffer
   loop
;

\ collect all characters until character c is received
: COLLECT-UNTIL ( c -- )   
   begin  
      serialKey 2dup = not while  
         MyBufferPut        \ add to the buffer
   repeat
   drop
;
 
\ --------------------------------------------------------------------- 
\ Tests
\ --------------------------------------------------------------------- 

DECIMAL

\ --------------------------------------------------------------------- 
\ Character matching and capture
\ --------------------------------------------------------------------- 
{ 
MAX_DATA_SIZE constant |MyBuffer| 

|MyBuffer| Buffer: MyBuffer
variable ptrMyBuffer
variable ShowMatching

: >MyBuffer ( -- a )  ptrMyBuffer @ |MyBuffer| 1- and MyBuffer + ;

: MyBufferPut ( c -- )   
\   ShowMatching @ if  dup SPemit+  then 
   >MyBuffer c!  1 ptrMyBuffer +! 
;

: COM-CAPTURE-LINE
   0 ptrMyBuffer !
   begin  serialKey? while  SerialBufferGet DUP 10 = IF  DROP EXIT  THEN  MyBufferPut  repeat 
;

: COM-SHOW-LINE      0 ptrMyBuffer !
   begin  serialKey? while  SerialBufferGet DUP 10 = IF  DROP EXIT  THEN  dup SPemit+  MyBufferPut  repeat 
;

variable #ToMatch

: COM-KEYW ( -- c )   begin  pause serialKey? UNTIL  SerialBufferGet ;

\ match one character
: Match1 ( c -- )     
   COM-KEYW
\   dup emit
   dup >r =    
   if  
      r> MyBufferPut       \ add to the buffer
   else
      r> drop  0 ptrMyBuffer !   \ clear the buffer
   then
;

\ wait until the received characters match string a n
: MATCH ( a n -- )    
   #ToMatch !
   0 ptrMyBuffer !
   serialPurge
   begin
      dup c@ Match1  
      1+             \ next character in string for next time 
      #ToMatch @ ptrMyBuffer @ = until
   drop
;

\ collect n characters 
: COLLECT ( n -- )  
   0 ?do 
      serialKey MyBufferPut \ add to the buffer
   loop
;

\ collect all characters until character c is received
: COLLECT-UNTIL ( c -- )   
   begin  
      serialKey 2dup = not while  
         MyBufferPut        \ add to the buffer
   repeat
   drop
;
}  
\ --------------------------------------------------------------------- 
\ Tests
\ --------------------------------------------------------------------- 

\\ 

        internal static string GetComPortByDescription(string Description)
        {
            string Result = string.Empty;
            Guid guid = PInvoke.GUID_DEVCLASS_PORTS;
            uint nDevice = 0;
            uint nBytes = 300;
            byte[] retval = new byte[nBytes];
            uint RequiredSize = 0;
            uint PropertyRegDataType = 0;

            PInvoke.SP_DEVINFO_DATA devInfoData = new PInvoke.SP_DEVINFO_DATA();
            devInfoData.cbSize = Marshal.SizeOf(typeof(PInvoke.SP_DEVINFO_DATA));

            IntPtr hDeviceInfo = PInvoke.SetupDiGetClassDevs(
                ref guid, 
                null, 
                IntPtr.Zero, 
                PInvoke.DIGCF.DIGCF_PRESENT);

            while (PInvoke.SetupDiEnumDeviceInfo(hDeviceInfo, nDevice++, ref devInfoData))
            {
                if (PInvoke.SetupDiGetDeviceRegistryProperty(
                        hDeviceInfo, 
                        ref devInfoData, 
                        PInvoke.SPDRP.SPDRP_FRIENDLYNAME,
                        out PropertyRegDataType, 
                        retval, 
                        nBytes, 
                        out RequiredSize))
                {
                    if (System.Text.Encoding.Unicode.GetString(retval).Substring(0, Description.Length).ToLower() ==
                        Description.ToLower())
                    {
                        string tmpstring = System.Text.Encoding.Unicode.GetString(retval);
                        Result = tmpstring.Substring(tmpstring.IndexOf("COM"),tmpstring.IndexOf(')') - tmpstring.IndexOf("COM"));
                    } // if retval == description
                } // if (PInvoke.SetupDiGetDeviceRegistryProperty( ... SPDRP_FRIENDLYNAME ...
            } // while (PInvoke.SetupDiEnumDeviceInfo(hDeviceInfo, nDevice++, ref devInfoData))

            PInvoke.SetupDiDestroyDeviceInfoList(hDeviceInfo);

            return Result;
        }
{ 
LONG WINAPI RegGetValue(
  __in         HKEY     hkey,
  __in_opt     LPCTSTR  lpSubKey,
  __in_opt     LPCTSTR  lpValue,
  __in_opt     DWORD    dwFlags,
  __out_opt    LPDWORD  pdwType,
  __out_opt    PVOID    pvData,
  __inout_opt  LPDWORD  zpcbData
);
}
